Package: tmap.mapgl
Title: Extensions to 'tmap' with Two New Modes: 'mapbox' and 'maplibre'
Type: Package
Authors@R: c(
	person("Martijn", "Tennekes", email = "mtennekes@gmail.com", role = c("aut", "cre"))
	)
Description: The 'tmap' package provides two plotting modes for static and interactive thematic maps. This package extends 'tmap' with two additional modes based on 'Mapbox GL JS' and 'MapLibre GL JS'. These modes feature interactive vector tiles, globe views, and other modern web-mapping capabilities, while maintaining a consistent 'tmap' interface across all plotting modes.
License: GPL-3
Version: 0.3
Encoding: UTF-8
Depends: R (>= 4.1),
Imports: tmap (>= 4.4), mapgl (>= 0.4.6), terra, rlang, cli, stars,
        data.table, grDevices, colorspace, htmltools, htmlwidgets, sf,
        tmaptools, units
Suggests: knitr
Config/Needs/website: bookdown, rmarkdown, r-tmap/tmap, walkerke/mapgl,
        dplyr, osmdata, mapview
URL: https://github.com/r-tmap/tmap.mapgl,
        https://r-tmap.github.io/tmap.mapgl/
BugReports: https://github.com/r-tmap/tmap.mapgl/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-15 09:13:36 UTC; mtes
Author: Martijn Tennekes [aut, cre]
Maintainer: Martijn Tennekes <mtennekes@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-15 10:50:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(tmapMapboxAuxPlot,default)
S3method(tmapMapboxAuxPlot,tm_aux_basemap)
S3method(tmapMapboxAuxPlot,tm_aux_rain)
S3method(tmapMapboxAuxPlot,tm_aux_snow)
S3method(tmapMapboxAuxPlot,tm_aux_tiles)
S3method(tmapMapboxAuxPrepare,default)
S3method(tmapMapboxAuxPrepare,tm_aux_basemap)
S3method(tmapMapboxAuxPrepare,tm_aux_rain)
S3method(tmapMapboxAuxPrepare,tm_aux_snow)
S3method(tmapMapboxAuxPrepare,tm_aux_tiles)
S3method(tmapMapboxCompHeight,tm_geocoder)
S3method(tmapMapboxCompHeight,tm_legend_landscape)
S3method(tmapMapboxCompHeight,tm_legend_portrait)
S3method(tmapMapboxCompHeight,tm_minimap)
S3method(tmapMapboxCompHeight,tm_scalebar)
S3method(tmapMapboxCompHeight,tm_snow)
S3method(tmapMapboxCompHeight,tm_title)
S3method(tmapMapboxCompPlot,tm_geocoder)
S3method(tmapMapboxCompPlot,tm_legend_landscape)
S3method(tmapMapboxCompPlot,tm_legend_portrait)
S3method(tmapMapboxCompPlot,tm_minimap)
S3method(tmapMapboxCompPlot,tm_scalebar)
S3method(tmapMapboxCompPlot,tm_snow)
S3method(tmapMapboxCompPlot,tm_title)
S3method(tmapMapboxCompPrepare,default)
S3method(tmapMapboxCompPrepare,tm_chart_none)
S3method(tmapMapboxCompPrepare,tm_geocoder)
S3method(tmapMapboxCompPrepare,tm_legend_landscape)
S3method(tmapMapboxCompPrepare,tm_legend_portrait)
S3method(tmapMapboxCompPrepare,tm_minimap)
S3method(tmapMapboxCompPrepare,tm_scalebar)
S3method(tmapMapboxCompPrepare,tm_snow)
S3method(tmapMapboxCompPrepare,tm_title)
S3method(tmapMapboxCompWidth,tm_geocoder)
S3method(tmapMapboxCompWidth,tm_legend_landscape)
S3method(tmapMapboxCompWidth,tm_legend_portrait)
S3method(tmapMapboxCompWidth,tm_minimap)
S3method(tmapMapboxCompWidth,tm_scalebar)
S3method(tmapMapboxCompWidth,tm_snow)
S3method(tmapMapboxCompWidth,tm_title)
S3method(tmapMapboxDataPlot,default)
S3method(tmapMapboxDataPlot,tm_data_borders)
S3method(tmapMapboxDataPlot,tm_data_bubbles)
S3method(tmapMapboxDataPlot,tm_data_dots)
S3method(tmapMapboxDataPlot,tm_data_fill)
S3method(tmapMapboxDataPlot,tm_data_lines)
S3method(tmapMapboxDataPlot,tm_data_polygons)
S3method(tmapMapboxDataPlot,tm_data_polygons_3d)
S3method(tmapMapboxDataPlot,tm_data_raster)
S3method(tmapMapboxDataPlot,tm_data_symbols)
S3method(tmapMaplibreAuxPlot,default)
S3method(tmapMaplibreAuxPlot,tm_aux_basemap)
S3method(tmapMaplibreAuxPlot,tm_aux_tiles)
S3method(tmapMaplibreAuxPrepare,default)
S3method(tmapMaplibreAuxPrepare,tm_aux_basemap)
S3method(tmapMaplibreAuxPrepare,tm_aux_tiles)
S3method(tmapMaplibreCompHeight,tm_credits)
S3method(tmapMaplibreCompHeight,tm_geocoder)
S3method(tmapMaplibreCompHeight,tm_legend_landscape)
S3method(tmapMaplibreCompHeight,tm_legend_portrait)
S3method(tmapMaplibreCompHeight,tm_minimap)
S3method(tmapMaplibreCompHeight,tm_mouse_coordinates)
S3method(tmapMaplibreCompHeight,tm_scalebar)
S3method(tmapMaplibreCompHeight,tm_title)
S3method(tmapMaplibreCompPlot,tm_credits)
S3method(tmapMaplibreCompPlot,tm_geocoder)
S3method(tmapMaplibreCompPlot,tm_legend_landscape)
S3method(tmapMaplibreCompPlot,tm_legend_portrait)
S3method(tmapMaplibreCompPlot,tm_minimap)
S3method(tmapMaplibreCompPlot,tm_mouse_coordinates)
S3method(tmapMaplibreCompPlot,tm_scalebar)
S3method(tmapMaplibreCompPlot,tm_title)
S3method(tmapMaplibreCompPrepare,default)
S3method(tmapMaplibreCompPrepare,tm_chart_none)
S3method(tmapMaplibreCompPrepare,tm_credits)
S3method(tmapMaplibreCompPrepare,tm_geocoder)
S3method(tmapMaplibreCompPrepare,tm_legend_landscape)
S3method(tmapMaplibreCompPrepare,tm_legend_portrait)
S3method(tmapMaplibreCompPrepare,tm_minimap)
S3method(tmapMaplibreCompPrepare,tm_mouse_coordinates)
S3method(tmapMaplibreCompPrepare,tm_scalebar)
S3method(tmapMaplibreCompPrepare,tm_title)
S3method(tmapMaplibreCompWidth,tm_credits)
S3method(tmapMaplibreCompWidth,tm_geocoder)
S3method(tmapMaplibreCompWidth,tm_legend_landscape)
S3method(tmapMaplibreCompWidth,tm_legend_portrait)
S3method(tmapMaplibreCompWidth,tm_minimap)
S3method(tmapMaplibreCompWidth,tm_mouse_coordinates)
S3method(tmapMaplibreCompWidth,tm_scalebar)
S3method(tmapMaplibreCompWidth,tm_title)
S3method(tmapMaplibreDataPlot,default)
S3method(tmapMaplibreDataPlot,tm_data_borders)
S3method(tmapMaplibreDataPlot,tm_data_bubbles)
S3method(tmapMaplibreDataPlot,tm_data_dots)
S3method(tmapMaplibreDataPlot,tm_data_fill)
S3method(tmapMaplibreDataPlot,tm_data_lines)
S3method(tmapMaplibreDataPlot,tm_data_polygons)
S3method(tmapMaplibreDataPlot,tm_data_polygons_3d)
S3method(tmapMaplibreDataPlot,tm_data_raster)
S3method(tmapMaplibreDataPlot,tm_data_symbols)
export(opt_tm_polygons_3d)
export(renderTmapMapbox)
export(renderTmapMaplibre)
export(tm_geocoder)
export(tm_mapbox)
export(tm_maplibre)
export(tm_polygons_3d)
export(tm_rain)
export(tm_snow)
export(tmapMapboxArrange)
export(tmapMapboxAux)
export(tmapMapboxAuxPlot)
export(tmapMapboxAuxPrepare)
export(tmapMapboxComp)
export(tmapMapboxCompHeight)
export(tmapMapboxCompPlot)
export(tmapMapboxCompPrepare)
export(tmapMapboxCompWidth)
export(tmapMapboxDataPlot)
export(tmapMapboxGridXLab)
export(tmapMapboxGridYLab)
export(tmapMapboxInit)
export(tmapMapboxOverlay)
export(tmapMapboxProviders)
export(tmapMapboxRun)
export(tmapMapboxShape)
export(tmapMapboxWrap)
export(tmapMapboxXtab)
export(tmapMaplibreArrange)
export(tmapMaplibreAux)
export(tmapMaplibreAuxPlot)
export(tmapMaplibreAuxPrepare)
export(tmapMaplibreComp)
export(tmapMaplibreCompHeight)
export(tmapMaplibreCompPlot)
export(tmapMaplibreCompPrepare)
export(tmapMaplibreCompWidth)
export(tmapMaplibreDataPlot)
export(tmapMaplibreGridXLab)
export(tmapMaplibreGridYLab)
export(tmapMaplibreInit)
export(tmapMaplibreOverlay)
export(tmapMaplibreProviders)
export(tmapMaplibreRun)
export(tmapMaplibreShape)
export(tmapMaplibreWrap)
export(tmapMaplibreXtab)
export(tmapOutputMapbox)
export(tmapOutputMaplibre)
export(tmapProxyMapbox)
export(tmapProxyMaplibre)
export(tmapValuesCVV_height)
export(tmapValuesCheck_height)
export(tmapValuesColorize_height)
export(tmapValuesIsDiv_height)
export(tmapValuesRange_height)
export(tmapValuesScale_height)
export(tmapValuesSubmit_height)
export(tmapValuesVV_height)
export(tmap_mapbox)
export(tmap_maplibre)
import(cli)
import(data.table)
import(sf)
import(stars)
importFrom(colorspace,deutan)
importFrom(colorspace,protan)
importFrom(colorspace,tritan)
importFrom(grDevices,col2rgb)
importFrom(grDevices,rgb)
importFrom(htmltools,tags)
importFrom(htmlwidgets,prependContent)
importFrom(stats,na.omit)
importFrom(terra,crop)
importFrom(terra,ext)
importFrom(terra,is.lonlat)
importFrom(terra,project)
importFrom(terra,rast)
importFrom(tmaptools,bb_poly)
importFrom(units,drop_units)
importFrom(units,set_units)
importFrom(utils,head)
importFrom(utils,tail)

.onLoad = function(...) {
	requireNamespace("tmap", quietly = TRUE)
	requireNamespace("mapgl", quietly = TRUE)
	requireNamespace("data.table", quietly = TRUE)

	tmap::tmapSubmitOptions(
		options = list(
			value.const = list(height = 0),
			value.na = list(height = NA),
			value.blank = list(height = 0),
			values.var = list(height = c(0, 1)),
			values.range = list(height = c(0, 1)),
			scales.var =  list(height = list(fact = "categorical", num = "continuous", datetime = "continuous", date = "continuous")),
			modes =
			list(mapbox =
				list(name = "Mapbox",
					 pitch = 0,
					 bearing = 0,
					   basemap.show = TRUE,
					 control.position = c("left", "top"),
					 control.collapse = TRUE,
					 zoom = NA,
					   scalebar.position = tmap::tm_pos_in("left", "bottom"),
					   minimap.position = tmap::tm_pos_in("left", "bottom"),
					   geocoder.position = tmap::tm_pos_in(pos.h = "right", pos.v = "top"),
					   geocoder.placeholder = "Search",
					   draw.download_button = TRUE,
					   draw.show_measurements = TRUE,
					   fullscreen.position =  tmap::tm_pos_in(pos.h = "right", pos.v = "top"),
					   basemap.server = "ofm.positron",
					   component.frame.color = "gray40",
					   crs_basemap = 4326),
				maplibre =
				list(name = "Maplibre",
					 pitch = 0,
					 bearing = 0,
					 basemap.show = TRUE,
					 control.position = c("left", "top"),
					 control.collapse = TRUE,
					 zoom = NA,
					 scalebar.position = tmap::tm_pos_in("left", "bottom"),
					 minimap.position = tmap::tm_pos_in("left", "bottom"),
					 geocoder.position = tmap::tm_pos_in(pos.h = "right", pos.v = "top"),
					 geocoder.placeholder = "Search",
					 draw.download_button = TRUE,
					 draw.show_measurements = TRUE,
					 fullscreen.position =  tmap::tm_pos_in(pos.h = "right", pos.v = "top"),
					 basemap.server = "ofm.positron",
					 component.frame.color = "gray40",
					 crs_basemap = 4326))),
		styleOptions = list(cobalt = list(modes =
										  	list(mapbox = list(basemap.server = "ofm.dark"),
										  		 maplibre = list(basemap.server = "ofm.dark")))))

	# Cross-mode basemap equivalences: when a basemap valid in one mode is
	# reproduced in another mode that lacks it, the closest provider below is
	# substituted instead of dropping to the mode default. Mappings are
	# approximate; the first provider listed per mode is the representative used
	# when substituting into that mode.
	tmap::tmapSubmitBasemapEquivalents(list(
		gray = list(
			plot     = "Esri.WorldGrayCanvas",
			view     = "Esri.WorldGrayCanvas",
			maplibre = "ofm.positron",
			mapbox   = "mapbox.light"),
		positron = list(
			plot     = "CartoDB.Positron",
			view     = "CartoDB.Positron",
			maplibre = "carto.positron",
			mapbox   = "carto.positron"),
		dark = list(
			plot     = "CartoDB.DarkMatter",
			view     = "CartoDB.DarkMatter",
			maplibre = c("carto.dark_matter", "ofm.dark"),
			mapbox   = c("mapbox.dark", "carto.dark_matter")),
		voyager = list(
			plot     = "CartoDB.Voyager",
			view     = "CartoDB.Voyager",
			maplibre = "carto.voyager",
			mapbox   = "carto.voyager"),
		streets = list(
			plot     = c("OpenStreetMap", "Esri.WorldStreetMap"),
			view     = c("OpenStreetMap", "Esri.WorldStreetMap"),
			maplibre = c("ofm.liberty", "esri.streets"),
			mapbox   = "mapbox.streets"),
		bright = list(
			plot     = "OpenStreetMap",
			view     = "OpenStreetMap",
			maplibre = "ofm.bright",
			mapbox   = "mapbox.standard"),
		topo = list(
			plot     = "Esri.WorldTopoMap",
			view     = "Esri.WorldTopoMap",
			maplibre = c("esri.topographic", "maptiler.topo"),
			mapbox   = "mapbox.outdoors"),
		satellite = list(
			plot     = "Esri.WorldImagery",
			view     = "Esri.WorldImagery",
			maplibre = c("esri.imagery", "maptiler.satellite"),
			mapbox   = c("mapbox.satellite", "mapbox.satellite_streets"))
	))

	.TMAP_MAPBOX$crs_options = c("eqearth" = "equalEarth", "wintri" = "winkelTripel", "aea" = "albers", "3857" = "mercator", "lcc" =  "lambertConformalConic")
	.TMAP_MAPLIBRE$crs_options = c("3857" = "mercator")
}

.TMAP_MAPBOX = new.env(FALSE, parent = globalenv())
.TMAP_MAPLIBRE = new.env(FALSE, parent = globalenv())

# ============================================================
#  Basemap style providers
# ============================================================
# Provider names follow "<family>.<style>[.<variant>]" and are resolved to a
# mapgl style URL/spec by resolve_style() in mapgl_shape.R. Families:
#   ofm      -> OpenFreeMap        (no key)
#   carto    -> mapgl::carto_style()      (no key; works on both engines)
#   mapbox   -> mapgl::mapbox_style()     (Mapbox token; mapbox mode only)
#   maptiler -> mapgl::maptiler_style()   (MAPTILER_API_KEY; both engines)
#   esri     -> mapgl::esri_style()       (ARCGIS_API_KEY; both engines)

.mapgl_styles_ofm = c("ofm.positron", "ofm.liberty", "ofm.bright", "ofm.dark", "ofm.fiord")

.mapgl_styles_carto = c("carto.voyager", "carto.positron", "carto.dark_matter",
						"carto.voyager_nolabels", "carto.positron_nolabels",
						"carto.dark_matter_nolabels")

# MapTiler styles are not a fixed enum (any MapTiler style works); this is a
# curated set. Colour variants (.dark/.light/.pastel) are appended as a final
# segment and parsed back into maptiler_style(variant=). Not all styles support
# all variants.
.mapgl_styles_maptiler = c("maptiler.streets", "maptiler.streets.dark", "maptiler.streets.light",
						   "maptiler.basic", "maptiler.basic.dark", "maptiler.basic.light",
						   "maptiler.bright", "maptiler.outdoor", "maptiler.topo",
						   "maptiler.winter", "maptiler.satellite", "maptiler.hybrid",
						   "maptiler.ocean", "maptiler.dataviz", "maptiler.dataviz.dark",
						   "maptiler.dataviz.light", "maptiler.backdrop", "maptiler.backdrop.dark",
						   "maptiler.backdrop.light", "maptiler.landscape", "maptiler.openstreetmap")

# Esri ArcGIS basemap styles (esri_style()).
.mapgl_styles_esri = paste0("esri.",
						c("navigation", "navigation_night", "streets", "streets_night",
						  "streets_relief", "community", "outdoor", "topographic", "terrain",
						  "imagery", "light_gray", "dark_gray", "oceans", "hillshade",
						  "human_geography", "human_geography_dark", "charted_territory",
						  "colored_pencil", "nova", "modern_antique", "midcentury", "newspaper"))

.mapgl_styles_mapbox = c("mapbox.standard", "mapbox.streets", "mapbox.outdoors",
						 "mapbox.light", "mapbox.dark", "mapbox.satellite",
						 "mapbox.satellite_streets", "mapbox.navigation_day",
						 "mapbox.navigation_night", "mapbox.standard_satellite")

#' @param credits credits
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxProviders = function(credits) {
	p = c(.mapgl_styles_mapbox, .mapgl_styles_ofm, .mapgl_styles_carto,
		  .mapgl_styles_maptiler, .mapgl_styles_esri)
	structure(as.list(p), names = p)
}

#' @param credits credits
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreProviders = function(credits) {
	p = c(.mapgl_styles_ofm, .mapgl_styles_carto, .mapgl_styles_maptiler,
		  .mapgl_styles_esri)
	structure(as.list(p), names = p)
}

#' @param tms tmap objects
#' @param nx number of facets
#' @param ncol, nrow number of rows and columns
#' @param opts options
#' @param knit knit
#' @param show show
#' @param args args
#' @param options options
#' @export
#' @keywords internal
#' @name tmapMapboxArrange
#' @rdname tmapMapbox
tmapMapboxArrange = function(tms, nx, ncol, nrow, opts, knit, show, args, options) {
	stop("arrange not implemented yet for mapbox")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreArrange = function(tms, nx, ncol, nrow, opts, knit, show, args, options) {
	stop("arrange not implemented yet for maplibre")
}

#' Extensions to 'tmap' with Two New Modes: 'mapbox' and 'maplibre'
#'
#' The 'tmap' package provides two plotting modes for static and interactive
#' thematic maps. This package extends 'tmap' with two additional modes based
#' on 'mapbox' <https://mapbox.com> and 'maplibre' <https://maplibre.org>.
#' These modes feature interactive vector tiles, globe views, and other modern
#' web-mapping capabilities, while using the same 'tmap' syntax for creating
#' maps (Tennekes 2018, <doi:10.32614/RJ-2018-027>).
#'
#' @author Martijn Tennekes \email{mtennekes@gmail.com}
#' @import data.table
#' @concept GIS
#' @concept thematic maps
#' @concept statistical maps
#' @concept choropleth
#' @concept bubble map
#' @example examples/mapbox.R
"_PACKAGE"

#' Map layer: polygons in 3d (experimental)
#'
#' Map layer that draws 3d (extruded) polygons. Supported visual variables are: `height` (the ), `fill` (the fill color),
#' `col` (the border color), `lwd` (line width), `lty` (line type),
#' `fill_alpha` (fill color alpha transparency) and `col_alpha` (border color alpha transparency).
#'
#' The visual variable arguments (e.g. `col`) can be specified with either a data
#' variable name (e.g., a spatial vector attribute or a raster layer of the object
#' specified in [tm_shape()]), or with a visual value (for `col`, a color is expected).
#' See \href{https://r-tmap.github.io/tmap/articles/basics_vv}{vignette about visual variables}.
#'
#' Multiple values can be specified: in that case facets are created.
#' These facets can be combined with other faceting data variables, specified with [tm_facets()].
#' See \href{https://r-tmap.github.io/tmap/articles/basics_facets}{vignette about facets}.
#'
#' - The `*.scale` arguments determine the used scale to map the data values to
#' visual variable values. These can be specified with one of the available
#' `tm_scale_*()` functions. The default is specified by the tmap option ([tm_options()]) `scales.var`.
#' See \href{https://r-tmap.github.io/tmap/articles/basics_scales}{vignette about scales}.
#'
#' - The `*.legend` arguments determine the used legend, specified with [tm_legend()].
#' The default legend and its settings are determined by the tmap options ([tm_options()]) `legend.` .
#' See \href{https://r-tmap.github.io/tmap/articles/basics_legends}{vignette about legends}.
#'
#' - The `*.chart` arguments specify additional charts, specified with `tm_chart_`, e.g. [tm_chart_histogram()].
#' See \href{https://r-tmap.github.io/tmap/articles/basics_charts}{vignette about charts}.
#'
#' - The `*.free` arguments determine whether scales are applied freely across facets, or shared.
#' A logical value is required. They can also be specified with a vector of three
#' logical values; these determine whether scales are applied freely per facet dimension.
#' This is only useful when facets are applied (see [tm_facets()]).
#' There are maximally three facet dimensions: rows, columns, and pages. This only
#' applies for a facet grid ([tm_facets_grid()]). For instance, `col.free = c(TRUE, FALSE, FALSE)`
#' means that for the visual variable `col`, each row of facets will have its own
#' scale, and therefore its own legend. For facet wraps and stacks
#' ([tm_facets_wrap()] and [tm_facets_stack()]) there is only one facet dimension,
#' so the `*.free` argument requires only one logical value.
#'
#' @param height,height.scale,height.legend,height.chart,height.free `r .doc_vv("height")`
#' @param fill,fill.scale,fill.legend,fill.chart,fill.free `r .doc_vv("fill")`
#' @param col,col.scale,col.legend,col.chart,col.free `r .doc_vv("col")`
#' @param lwd,lwd.scale,lwd.legend,lwd.chart,lwd.free `r .doc_vv("lwd")`
#' @param lty,lty.scale,lty.legend,lty.chart,lty.free `r .doc_vv("lty")`
#' @param fill_alpha,fill_alpha.scale,fill_alpha.chart,fill_alpha.legend,fill_alpha.free `r .doc_vv("fill_alpha")`
#' @param col_alpha,col_alpha.scale,col_alpha.legend,col_alpha.chart,col_alpha.free `r .doc_vv("col_alpha")`
#' @param linejoin,lineend Line join and line end. See [gpar()][grid::gpar()] for details.
#' @param plot.order Specification in which order the spatial features are drawn.
#'   See [tm_plot_order()] for details.
#' @param zindex Map layers are drawn on top of each other. The `zindex` numbers
#'   (one for each map layer) determines the stacking order.
#'   By default the map layers are drawn in the order they are called.
#' @param group Name of the group to which this layer belongs. This is only
#'   relevant in view mode, where layer groups can be switched (see `group.control`)
#' @param group.control In view mode, the group control determines how layer groups
#'   can be switched on and off. Options: `"radio"` for radio buttons
#'   (meaning only one group can be shown), `"check"` for check boxes
#'   (so multiple groups can be shown), and `"none"` for no control
#'   (the group cannot be (de)selected).
#' @param options options passed on to the corresponding `opt_<layer_function>` function
#' @param popup.vars names of data variables that are shown in the popups
#'   in `"view"` mode. Set popup.vars to `TRUE` to show all variables in the
#'   shape object. Set popup.vars to `FALSE` to disable popups. Set `popup.vars`
#'   to a character vector of variable names to those those variables in the popups.
#'   The default (`NA`) depends on whether visual variables (e.g.`fill`) are used.
#'   If so, only those are shown. If not all variables in the shape object are shown.
#' @param popup.format list of formatting options for the popup values.
#'   See the argument `legend.format` for options. Only applicable for
#'   numeric data variables. If one list of formatting options is provided,
#'   it is applied to all numeric variables of `popup.vars`. Also, a (named)
#'   list of lists can be provided. In that case, each list of formatting options
#'   is applied to the named variable.
#' @param hover name of the data variable that specifies the hover labels (view mode only). Set to `FALSE` to disable hover labels. By default `FALSE`, unless `id` is specified. In that case, it is set to `id`,
#' @param id name of the data variable that specifies the indices of the spatial
#'   features. Only used for `"view"` mode.
#' @example ./examples/tm_polygons_3d.R
#' @return a [tmap::tmap-element], supposed to be stacked after [tmap::tm_shape()] using the `+` operator. The `opt_<layer_function>` function returns a list that should be passed on to the `options` argument.
#' @seealso \href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth example (1)} and \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{choropleth example (2)}
#' @export
tm_polygons_3d = function(height = tmap::tm_const(),
						 height.scale = tmap::tm_scale(),
						 height.legend = tmap::tm_legend_hide(),
						 height.chart = tmap::tm_chart_none(),
						 height.free = NA,
					   fill = tmap::tm_const(),
					   fill.scale = tmap::tm_scale(),
					   fill.legend = tmap::tm_legend(),
					   fill.chart = tmap::tm_chart_none(),
					   fill.free = NA,
					   col = tmap::tm_const(),
					   col.scale = tmap::tm_scale(),
					   col.legend = tmap::tm_legend(),
					   col.chart = tmap::tm_chart_none(),
					   col.free = NA,
					   lwd = tmap::tm_const(),
					   lwd.scale = tmap::tm_scale(),
					   lwd.legend = tmap::tm_legend(),
					   lwd.chart = tmap::tm_chart_none(),
					   lwd.free = NA,
					   lty = tmap::tm_const(),
					   lty.scale = tmap::tm_scale(),
					   lty.legend = tmap::tm_legend(),
					   lty.chart = tmap::tm_chart_none(),
					   lty.free = NA,
					   fill_alpha = tmap::tm_const(),
					   fill_alpha.scale = tmap::tm_scale(),
					   fill_alpha.legend = tmap::tm_legend(),
					   fill_alpha.chart = tmap::tm_chart_none(),
					   fill_alpha.free = NA,
					   col_alpha = tmap::tm_const(),
					   col_alpha.scale = tmap::tm_scale(),
					   col_alpha.legend = tmap::tm_legend(),
					   col_alpha.chart = tmap::tm_chart_none(),
					   col_alpha.free = NA,
					   linejoin = "round",
					   lineend = "round",
					   plot.order = tmap::tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
					   zindex = NA,
					   group = NA,
					   group.control = "check",
					   popup.vars = NA,
					   popup.format = list(),
					   hover = NA,
					   id = "",
					   options = opt_tm_polygons_3d()) {

	args_called = names(rlang::call_match()[-1])


	tmap::tm_element_list(tmap::tm_element(
		layer = c("polygons_3d", "polygons"),
		trans.fun = tmap::tmapTransPolygons,
		trans.args = options$trans.args,
		trans.aes = list(),
		trans.apply_to = "this",
		mapping.aes = list(height = tmap::tmapScale(aes = "height",
										   value = height,
										   scale = height.scale,
										   legend = height.legend,
										   chart = height.chart,
										   free = height.free),
						   fill = tmap::tmapScale(aes = "fill",
											value = fill,
											scale = fill.scale,
											legend = fill.legend,
											chart = fill.chart,
											free = fill.free),
						   col = tmap::tmapScale(aes = "col",
						   				value = col,
						   				scale = col.scale,
						   				legend = col.legend,
						   				chart = col.chart,
						   				free = col.free),
						   lwd = tmap::tmapScale(aes = "lwd",
						   				value = lwd,
						   				scale = lwd.scale,
						   				legend = lwd.legend,
						   				chart = lwd.chart,
						   				free = lwd.free),
						   lty = tmap::tmapScale(aes = "lty",
						   				value = lty,
						   				scale = lty.scale,
						   				legend = lty.legend,
						   				chart = lty.chart,
						   				free = lty.free),
						   fill_alpha = tmap::tmapScale(aes = "fill_alpha",
						   					   value = fill_alpha,
						   					   scale = fill_alpha.scale,
						   					   legend = fill_alpha.legend,
						   					   chart = fill_alpha.chart,
						   					   free = fill_alpha.free),
						   col_alpha = tmap::tmapScale(aes = "col_alpha",
						   					  value = col_alpha,
						   					  scale = col_alpha.scale,
						   					  legend = col_alpha.legend,
						   					  chart = col_alpha.chart,
						   					  free = col_alpha.free)),

		gpar = tmap::tmapGpar(height = "__height",
						fill = "__fill",
						col = "__col",
						shape = NA,
						size = NA,
						fill_alpha = "__fill_alpha",
						col_alpha = "__col_alpha",
						pattern = "fill",
						lty = "__lty",
						lwd = "__lwd",
						linejoin = linejoin,
						lineend = lineend),
		tpar = tmap::tmapTpar(area = "AREA"),
		plot.order = plot.order,
		mapping.fun = "Polygons3d",
		mapping.args = options$mapping.args,
		zindex = zindex,
		group = group,
		group.control = group.control,
		popup.vars = popup.vars,
		popup.format = popup.format,
		hover = hover,
		id = id,
		subclass = c("tm_aes_layer", "tm_layer")))
}

#' @param polygons.only should only polygon geometries of the shape object (defined in [tm_shape()]) be plotted? By default `"ifany"`, which means `TRUE` in case a geometry collection is specified.
#' @param height.max,height.min Maximum and minimum height. The data variable values assigned to `height` are are scaled between these two values. Each can be an absolute number (in meters) or relative to the bounding box area, more specifically, the percentage of the square root of the bounding box area. The default values are `"10\%"` for `height.max` and `0.1\%` for `height.min`. The latter is not set to 0 because of artefacts with maplibre in globe view.
#' @param vertical.gradient Should a vertical gradient be applied to the sides of the extrusions? When `TRUE`, the side walls are shaded from darker at the base to lighter toward the top, which gives a sense of depth. Supported in **both** `"maplibre"` and `"mapbox"` modes. Passed to `fill_extrusion_vertical_gradient`. The default `NULL` keeps the engine default, which is `TRUE` - so the visible effect is mainly when set to `FALSE` (flat, evenly-coloured side walls).
#' @param cast.shadows Should the extruded polygons cast shadows? **`"mapbox"` mode only** (silently ignored with a message in `"maplibre"` mode, whose fill-extrusion implementation does not support shadows). Passed to `fill_extrusion_cast_shadows`. The default `NULL` keeps the engine default.
#' @param ambient.occlusion.intensity,ambient.occlusion.radius Ambient-occlusion intensity (0-1) and radius for the extruded polygons, adding soft contact shading in corners and crevices. **`"mapbox"` mode only** (ignored with a message in `"maplibre"` mode). Passed to `fill_extrusion_ambient_occlusion_intensity` / `fill_extrusion_ambient_occlusion_radius`. The default `NULL` keeps the engine default.
#' @param emissive.strength Emissive strength of the extruded polygons, controlling how much they appear self-lit under the map's 3D lighting. **`"mapbox"` mode only** (ignored with a message in `"maplibre"` mode). Passed to `fill_extrusion_emissive_strength`. The default `NULL` keeps the engine default.
#' @export
#' @rdname tm_polygons_3d
opt_tm_polygons_3d = function(polygons.only = "ifany", height.max = "10%", height.min = "0.1%",
							  cast.shadows = NULL,
							  ambient.occlusion.intensity = NULL,
							  ambient.occlusion.radius = NULL,
							  vertical.gradient = NULL,
							  emissive.strength = NULL) {
	list(trans.args = list(polygons.only = polygons.only),
		 mapping.args = list(height.max = height.max,
		 					 height.min = height.min,
		 					 cast.shadows = cast.shadows,
		 					 ambient.occlusion.intensity = ambient.occlusion.intensity,
		 					 ambient.occlusion.radius = ambient.occlusion.radius,
		 					 vertical.gradient = vertical.gradient,
		 					 emissive.strength = emissive.strength))
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
renderTmapMapbox = function(expr, env, quoted, execOnResize) {
	mapgl::renderMapboxgl(expr, env, quoted)
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapOutputMapbox = function(outputId, width, height) {
	mapgl::mapboxglOutput(outputId, width = width, height = height)
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapProxyMapbox = function(mapId, session, x) {
	print(x, dg = mapgl::mapboxgl_proxy(mapId, session), show = FALSE, in.shiny = TRUE, proxy = TRUE)
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
renderTmapMaplibre = function(expr, env, quoted, execOnResize) {
	mapgl::renderMaplibre(expr, env, quoted)
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapOutputMaplibre = function(outputId, width, height) {
	mapgl::maplibreOutput(outputId, width = width, height = height)
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapProxyMaplibre = function(mapId, session, x) {
	print(x, dg = mapgl::maplibre_proxy(mapId, session), show = FALSE, in.shiny = TRUE, proxy = TRUE)
}

tmapMapboxCompPrepare = function(comp, o) {
	UseMethod("tmapMapboxCompPrepare")
}

tmapMapboxCompHeight = function(comp, o) {
	UseMethod("tmapMapboxCompHeight")
}

tmapMapboxCompWidth = function(comp, o) {
	UseMethod("tmapMapboxCompWidth")
}

tmapMapboxCompPlot = function(comp, m, o) {
	UseMethod("tmapMapboxCompPlot")
}


#' @export
tmapMapboxCompPrepare.tm_legend_portrait = function(comp, o) {
	mapgl_legend_comp(comp, o, mode = "mapbox")
}


#' @export
tmapMapboxCompHeight.tm_legend_portrait = function(comp, o) {
	mapgl_leg_height(comp, o)
}



#' @export
tmapMapboxCompWidth.tm_legend_portrait = function(comp, o) {
	mapgl_leg_width(comp, o)

}


#' @export
tmapMapboxCompPrepare.tm_legend_landscape = function(comp, o) {
	mapgl_legend_comp(comp, o, mode = "mapbox")
}

#' @export
tmapMapboxCompHeight.tm_legend_landscape = function(comp, o) {
	mapgl_leg_height(comp, o)
}



#' @export
tmapMapboxCompWidth.tm_legend_landscape = function(comp, o) {
	mapgl_leg_width(comp, o)

}










tmapMaplibreCompPrepare = function(comp, o) {
	UseMethod("tmapMaplibreCompPrepare")
}

tmapMaplibreCompHeight = function(comp, o) {
	UseMethod("tmapMaplibreCompHeight")
}

tmapMaplibreCompWidth = function(comp, o) {
	UseMethod("tmapMaplibreCompWidth")
}

tmapMaplibreCompPlot = function(comp, m, o) {
	UseMethod("tmapMaplibreCompPlot")
}


#' @export
tmapMaplibreCompPrepare.tm_legend_portrait = function(comp, o) {
	mapgl_legend_comp(comp, o, mode = "maplibre")
}


#' @export
tmapMaplibreCompHeight.tm_legend_portrait = function(comp, o) {
	mapgl_leg_height(comp, o)
}



#' @export
tmapMaplibreCompWidth.tm_legend_portrait = function(comp, o) {
	mapgl_leg_width(comp, o)
}


#' @export
tmapMaplibreCompPrepare.tm_legend_landscape = function(comp, o) {
	mapgl_legend_comp(comp, o, mode = "maplibre")
}

#' @export
tmapMaplibreCompHeight.tm_legend_landscape = function(comp, o) {
	mapgl_leg_height(comp, o)
}



#' @export
tmapMaplibreCompWidth.tm_legend_landscape = function(comp, o) {
	mapgl_leg_width(comp, o)

}

mapgl_leg_height = function (comp, o) {
	if (is.na(comp$height)) {
		if (comp$type == "gradient") {
			comp$height = 5.5 * 20
		} else {
			comp$height = (number_text_lines(comp$title) + 2 + comp$nitems * 1.25) * 20
		}
	}
	comp
}


mapgl_leg_width = function (comp, o) {
	if (is.na(comp$width)) {
		if (comp$type == "gradient") {
			comp$width = 250
		} else {
			comp$width = 75 + max(nchar(c(comp$title, comp$labels))) * 6
		}
	}
	comp
}


mapgl_legend_comp = function(comp, o, mode) {
	within(comp, {
		if ("biv" %in% names(attributes(gp$fill))) {
			warning("Bivariate legend not implemented for mapbox mode", call. = FALSE)
			show = FALSE
		}

		nuq = vapply(comp$gp, length, FUN.VALUE = integer(1))
		varying = names(nuq)[which(nuq>1)]

		if (all(c("col", "fill") %in% varying)) {
			cli::cli_inform("Legend in mode {.val {mode}} doesn't support both fill and col varying. Setting col to the first value")
			nuq["col"] = 1
			varying = setdiff(varying, "col")
			gp$col = gp$col[1]
		}


		type = if (!any(c("col", "fill", "lwd", "size") %in% varying)) {
			cli::cli_inform("No legends available in mode {.val {mode}} for map variables {.val {varying}}")
			"none"
		} else if ((!is.na(gp$fill[1]) && any(nchar(gp$fill) > 50)) || (!is.na(gp$fill_alpha[1]) && any(nchar(gp$fill_alpha) > 50)) ||
				   (!is.na(gp$col[1]) && any(nchar(gp$col) > 50)) || (!is.na(gp$col_alpha[1]) && any(nchar(gp$col_alpha) > 50))) {
			#message("Continuous legend not implemented in mapbox mode")
			"gradient"
		} else if (any(c("bgcol", "bgcol_alpha") %in% varying)) {
			"none"
		} else if ("lwd" %in% varying || ("tm_data_lines" %in% mfun)) {
			"lines"
		} else {
			"symbols"
		}

		if (type == "none") show = FALSE

		bg.color = do.call("process_color", c(list(col=bg.color), o$pc))
		title.color = do.call("process_color", c(list(col=title.color), o$pc))
		text.color = do.call("process_color", c(list(col=text.color), o$pc))

		gp2 = gp_to_lpar(gp, mfun = comp$mfun, shape = comp$item.shape)
	})
}

# ============================================================
#  Shared helpers
# ============================================================

# Build a MapLibre 'match' expression from a categorical scale mapping.
# Returns a properly nested list that serialises to
#   ["match", ["get", var], level1, value1, ..., fallback]
build_match_expr = function(var, mapping) {
	pairs = as.list(rbind(mapping$levels_orig, mapping$values_orig))  # interleaved pairs
	c(list("match", list("get", var)), pairs, list(mapping$value_na))
}

# Factory: returns a get_pmt_aes() closure over dt and .TMAP.
# For scale-driven aesthetics it resolves the MapLibre expression;
# for fixed values it returns the raw scalar from dt[[a]][1].
make_get_pmt_aes = function(dt) {
	function(a) {
		v      = dt[[a]][1]
		is_var = substr(v, 1, 5) == "scale"
		if (!is_var) return(v)

		snr     = as.numeric(substr(v, 6, 8))
		leg     = .TMAP$legs[[snr]]
		var     = substr(v, 10, nchar(v))
		mapping = leg$layer_args$mapping

		if (!is.null(mapping$levels_orig)) {
			build_match_expr(var, mapping)
		} else {
			mapgl::get_column(var)
		}
	}
}

# ------------------------------------------------------------
#  Interactive-legend categorical columns
# ------------------------------------------------------------
# A data layer can carry up to two data-driven colour aesthetics: the body
# (fill) and the outline (col). Each gets its own legend, and mapgl filters a
# layer by matching a *data column* against the clicked category. We therefore
# materialise a dedicated category column per aesthetic on the source:
#   __tmap_cat_fill__  – for the fill legend
#   __tmap_cat_col__   – for the border/line (col) legend
# Using one shared column (the old "__tmap_cat__") made the two legends collide:
# both keyed off the same column and both targeted the fill layer, so the col
# legend filtered on the fill aesthetic (and clobbered the fill legend's filter,
# since mapgl lets only one interactive legend own a layer).
#
# Which legend backs an aesthetic is found by glid + colour matching: among the
# legends carrying this layer's glid, the backing one is whichever's swatch
# colours (vvalues) actually appear in the resolved per-feature colours. This is
# the original approach (it does NOT parse dt — dt's "scaleNNN_var" encoding only
# exists in the PMTiles path, not the sf path, where colours are already
# resolved). NULL is returned for a constant or continuous aesthetic.

mapgl_cat_info = function(glid, vals) {
	non_na = !is.na(vals)
	if (!any(non_na)) return(NULL)

	# Normalise colours to upper-case #RRGGBB before matching, so an alpha suffix
	# (#RRGGBBAA) or case difference between the resolved fills and the swatch
	# colours doesn't break the match. (Categorical/interval palettes have
	# distinct RGB per class, so dropping alpha never collapses two classes.)
	norm = function(x) toupper(substr(as.character(x), 1L, 7L))
	vn_all = norm(vals)
	vn = vn_all[non_na]

	best = NULL
	best_frac = 0
	best_cc = NULL
	for (leg in .TMAP$legs) {
		if (!("glid" %in% names(leg)) || !isTRUE(leg$glid == glid)) next
		cc = leg$vvalues
		cv = leg$labels
		if (is.null(cc) || is.null(cv)) next
		ccn = norm(cc)
		frac = mean(vn %in% ccn)
		if (frac > best_frac) { best_frac = frac; best = leg; best_cc = ccn }
	}

	# A discrete scale (categorical OR binned numeric) paints (almost) every
	# feature with a swatch colour; a continuous scale interpolates off-swatch
	# colours, so its match fraction is tiny. Require a majority to treat the
	# aesthetic as categorical/interactive (and leave continuous to the
	# non-interactive gradient legend).
	if (is.null(best) || best_frac < 0.5) return(NULL)

	list(
		labels = best$labels[match(vn_all, best_cc)],
		values = best$labels,
		colors = best$vvalues
	)
}

# Attach the categorical columns for the fill and/or col aesthetics to the
# source `shp2`, and record in .TMAP$mapgl_cat[[glid]] which aesthetics are
# categorical plus the concrete fill/border layer names. The legend renderer
# (mapgl_legend_target()) reads this registry to pick the right target layer(s)
# and filter column. `border_layer = NA` marks geometries with no separate
# outline layer (tm_symbols' single circle layer); `fill_layer = NA` marks
# outline-only geometries (lines).
mapgl_attach_cat = function(shp2, glid, fill_layer = NA_character_, border_layer = NA_character_,
							fill_col = "fill", col_col = "col") {
	if (is.null(.TMAP$mapgl_cat)) .TMAP$mapgl_cat = list()
	reg = list(fill = NULL, col = NULL, fill_layer = fill_layer, border_layer = border_layer)

	if (!is.null(fill_col) && fill_col %in% names(shp2)) {
		fi = mapgl_cat_info(glid, shp2[[fill_col]])
		if (!is.null(fi)) {
			shp2[["__tmap_cat_fill__"]] = fi$labels
			reg$fill = list(column = "__tmap_cat_fill__", values = fi$values, colors = fi$colors)
		}
	}
	if (!is.null(col_col) && col_col %in% names(shp2)) {
		ci = mapgl_cat_info(glid, shp2[[col_col]])
		if (!is.null(ci)) {
			shp2[["__tmap_cat_col__"]] = ci$labels
			reg$col = list(column = "__tmap_cat_col__", values = ci$values, colors = ci$colors)
		}
	}
	.TMAP$mapgl_cat[[glid]] = reg
	shp2
}

# Shared PMTiles guard: emits cli messages and returns TRUE when the
# pointer cannot be rendered, so callers can do:
#   if (pmtiles_unsupported(shpTM, mode)) return(NULL)
pmtiles_unsupported = function(shpTM, mode) {
	smeta = shpTM$smeta
	if (mode == "mapbox") {
		cli::cli_inform("Source shapes are not supported in {.str mapbox} mode yet, only in {.str maplibre}")
		return(TRUE)
	}
	if (smeta$type != "pmtiles") {
		cli::cli_inform("Source shapes other than PMTiles not supported yet")
		return(TRUE)
	}
	if (smeta$tile_type != "mvt") {
		cli::cli_inform("Source shape is not a vector format")
		return(TRUE)
	}
	FALSE
}

# ============================================================
#  Popup helper
# ============================================================

# Popup styling (width, max.height, the per-region align/color, and css) lives
# on the tm_popup() `layout` list built in tmap core and is threaded here as
# `popup.layout`. Core is the single source of truth: it owns the defaults
# (popup_layout_default()) and the resolution of bare numbers to CSS lengths
# (complete_popup_layout(): width -> px, max.height -> em, NA/Inf -> "none").
# We reuse those directly via getFromNamespace() — the same pattern the package
# uses elsewhere (e.g. pane_name) — so there is no duplicated default table or
# number-resolution logic to keep in sync with view mode.

view_format_popups_mapgl = function(id = NULL, titles, format, values, layout = NULL) {
	complete_popup_layout = utils::getFromNamespace("complete_popup_layout", "tmap")
	style_attr            = utils::getFromNamespace(".popup_style_attr", "tmap")
	layout = complete_popup_layout(layout)

	h = lapply(format, function(f) {
		if (f$html.escape) {
			htmltools::htmlEscape
		} else {
			function(x) x
		}
	})

	if (!is.null(id)) {
		labels = paste0("<b>", h[[1]](id), "</b>")
	} else {
		labels = ""
	}

	titles_format = mapply(function(ti, hi) {
		hi(ti)
	}, titles, h, SIMPLIFY = FALSE)

	values_format = mapply(function(v, f, hi) {
		if (inherits(v, "units")) {
			popup_append = paste0(" ", as.character(attr(v, "units")))
		} else {
			popup_append = ""
		}
		numbers = hi(if (is.numeric(v)) do.call("fancy_breaks", c(list(vec = as.numeric(v), intervals = FALSE), f)) else v)
		paste0(numbers, popup_append)
	}, values, format, h, SIMPLIFY = FALSE)

	# Per-region styling (alignment + color), shared with view mode.
	td_label = paste0("<td class=\"tmap-popup-label\"", style_attr(layout$label.align, layout$label.color), "><nobr>")
	td_value = paste0("<td class=\"tmap-popup-value\"", style_attr(layout$value.align, layout$value.color), "><nobr>")
	th_attr  = style_attr(layout$title.align, layout$title.color)

	labels2 = mapply(function(l, v) {
		paste0("<tr>", td_label, l, "</nobr></td>", td_value, v, "</nobr></td>")
	}, titles_format, values_format, SIMPLIFY = FALSE)

	labels3 = paste0(do.call("paste", c(labels2, list(sep = "</tr>"))), "</tr>")

	# max.height = "none" removes the cap (popup grows to fit, never scrolls).
	# When capped, reserve scrollbar space with padding-right so the scroll bar
	# doesn't overlap the values (notably Safari's overlay scrollbar). Mirrors
	# view mode: pad once the table is long enough to scroll (~>13 lines).
	padding_right = if (length(titles_format) > 13) 15 else 0
	mh = layout$max.height
	no_cap = is.null(mh) || identical(mh, "none") || (length(mh) == 1L && is.na(mh))
	if (no_cap) {
		div_style = paste0("width:", layout$width, "; overflow-x:hidden;")
	} else {
		div_style = paste0("width:", layout$width, "; max-height:", mh, "; overflow-y:auto; overflow-x:hidden; padding-right:", padding_right, "px;")
	}

	# Mirror view mode: with a fixed width let the table fill it (extra space
	# falls between the label and value columns); with "auto" the table shrinks
	# to fit. This plain table replaced the width:100%/table-layout:fixed version
	# that made popups ~150% too wide and split the columns evenly.
	table_style = if (identical(layout$width, "auto")) "" else " style=\"width:100%;\""

	# Free-form user CSS (tm_popup(css=)) injected verbatim. Target classes:
	# .tmap-popup, .tmap-popup-table, .tmap-popup-title, .tmap-popup-label,
	# .tmap-popup-value. To resize the popup box itself, target MapLibre's own
	# .maplibregl-popup-content.
	css_block = if (!is.null(layout$css) && length(layout$css) == 1L && !is.na(layout$css) && nzchar(layout$css)) {
		paste0("<style>", layout$css, "</style>")
	} else {
		""
	}

	paste0(
		css_block,
		"<div class=\"tmap-popup\" style='", div_style, "'>",
		"<table class=\"tmap-popup-table\"", table_style, ">",
		"<thead><tr><th class=\"tmap-popup-title\" colspan=\"2\"", th_attr, ">", labels, "</th></tr></thead>",
		labels3,
		"</table></div>"
	)
}

# ============================================================
#  Shared hover / popup attachment for sf-backed layers
# ============================================================

# Attaches hover text and popup HTML columns to shp2 and returns
# the resolved tooltip/popup arguments ready for add_*_layer().
#
# Returns a list:
#   $shp2    – the sf data frame, possibly with $hover and/or $popup columns added
#   $hdt_arg – value to pass as tooltip= (NULL or mapgl::get_column("hover"))
#   $pdt_arg – value to pass as popup=   (NULL or mapgl::get_column("popup"))
attach_hover_popup = function(shp2, dt, hdt, pdt, idt, popup.format, popup.layout = NULL, ptdt = NULL) {

	hdt_arg = NULL
	pdt_arg = NULL

	if (!is.null(hdt)) {
		shp2$hover = hdt$hover[match(dt$tmapID__, hdt$tmapID__)]
		shp2$hover = vapply(shp2$hover, htmltools::HTML, FUN.VALUE = character(1))
		hdt_arg    = mapgl::get_column("hover")
	}

	if (!is.null(pdt)) {
		mtch    = match(dt$tmapID__, pdt$tmapID__)
		pdt_sub = pdt[mtch][, tmapID__ := NULL]

		# Popup header: an explicit popup title (tm_popup(title=), carried in
		# ptdt) takes precedence; otherwise fall back to the hover text, then to
		# no header. Mirrors the Leaflet backend's title resolution.
		id_arg = if (!is.null(ptdt)) {
			ptdt$title[match(dt$tmapID__, ptdt$tmapID__)]
		} else if (!is.null(hdt)) {
			shp2$hover
		} else {
			NULL
		}

		shp2$popup = view_format_popups_mapgl(
			id     = id_arg,
			titles = names(pdt_sub),
			values = pdt_sub,
			format = popup.format,
			layout = popup.layout
		)
		pdt_arg = mapgl::get_column("popup")
	}

	list(shp2 = shp2, hdt_arg = hdt_arg, pdt_arg = pdt_arg)
}

# ============================================================
#  split_alpha_channel
# ============================================================

split_alpha_channel = function(x, alpha) {
	if (is.null(x)) {
		list(col = NULL, opacity = 0)
	} else {
		RGBA    = col2rgb(x, alpha = TRUE)
		col     = rgb(RGBA[1, ], RGBA[2, ], RGBA[3, ], maxColorValue = 255)
		opacity = unname(RGBA[4, ] / 255 * alpha)
		list(col = col, opacity = opacity)
	}
}

# ============================================================
#  lty2dash
# ============================================================

lty2dash = function(lty) {
	tab = c(solid = "", dashed = "4 4", dotted = "1 3", dotdash = "1 3 4 3",
			 longdash = "7 3", twodash = "2 2 6 2")
	are_words   = (lty %in% names(tab))
	if (all(are_words)) {
		unname(tab[lty])
	} else {
		are_letters = (suppressWarnings(!is.na(as.numeric(lty))))
		if (!all(are_letters | are_words)) {
			stop("Incorrect lty specification: ", lty[which(!are_letters & !are_words)[1]])
		} else {
			lty[are_words]   = unname(tab[lty[are_words]])
			lty[are_letters] = vapply(
				strsplit(lty[are_letters], ""),
				FUN = function(x) paste(x, collapse = " "),
				FUN.VALUE = character(1)
			)
		}
		lty
	}
}

# ============================================================
#  Source id helper
# ============================================================

# Source ids must be unique per render. clear_layer() (used by tm_remove_layer
# in proxy mode) removes *layers* but not their *sources*, so reusing a source
# id on a redraw collides with the still-present stale source: MapLibre keeps
# the old data and the re-added layers recolour from it. .TMAP$stamp is bumped
# by print.tmap() on every render, giving a fresh source id each time. Layer
# ids stay stable (glid-based) so the tm_remove_layer registry keeps matching.
mapgl_srcid = function(base) sprintf("%s_%.0f", base, .TMAP$stamp * 1e6)

# ============================================================
#  Layer blending guard
# ============================================================

# MapLibre / Mapbox GL composite all layers in a single WebGL context with
# standard (source-over) alpha blending; the GL style spec exposes no
# per-layer compositing operator (unlike grid::groupGrob() in plot mode or CSS
# mix-blend-mode in Leaflet view mode). So a non-trivial `blend` value cannot
# be honoured here. We inform once per layer and render without blending,
# mirroring how the package already reports other unsupported features.
mapgl_blend_unsupported = function(a, mode) {
	b = a$blend
	if (!is.null(b) && length(b) == 1L && !is.na(b) && !identical(b, "over")) {
		cli::cli_inform(c(
			"!" = "Layer blending ({.code blend = {.str {b}}}) is not supported in {.str {mode}} mode.",
			"i" = "The layer is rendered without blending. Use {.str plot} or {.str view} mode to enable layer blending."
		))
	}
	invisible(NULL)
}

# ============================================================
#  mapgl_polygons
# ============================================================

mapgl_polygons = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp,
						   bbx, facet_row, facet_col, facet_page,
						   id, pane, group, glid, o, ..., mode, popup.layout = NULL, ptdt = NULL) {

	m = get_mapgl(facet_row, facet_col, facet_page, mode)
	mapgl_blend_unsupported(a, mode)
	rc_text    = frc(facet_row, facet_col)
	shp_is_pointer = inherits(shpTM$shp, "character")

	# ----------------------------------------------------------
	#  PMTiles branch
	# ----------------------------------------------------------
	if (shp_is_pointer) {
		if (pmtiles_unsupported(shpTM, mode)) return(NULL)

		smeta      = shpTM$smeta
		srcname    = mapgl_srcid(paste0("layer", pane))
		layername1 = paste0(glid, "polygons_fill")
		layername2 = paste0(glid, "polygons_border")
		url        = smeta$url

		get_pmt_aes = make_get_pmt_aes(dt)

		aes_f   = get_pmt_aes("fill")
		aes_fo  = get_pmt_aes("fill_alpha")
		aes_c   = get_pmt_aes("col")
		aes_co  = get_pmt_aes("col_alpha")
		aes_lwd = get_pmt_aes("lwd")

		# Note: hover / popup over PMTiles are intentionally not supported here —
		# the popup data lives in dt (R memory) while tiles are fetched remotely,
		# so there is no row-level join available at render time.

		m |>
			mapgl::add_pmtiles_source(id = srcname, url = url) |>
			mapgl::add_fill_layer(layername1, source = srcname,
								  source_layer  = smeta$layer,
								  fill_color    = aes_f,
								  fill_opacity  = aes_fo) |>
			mapgl::add_line_layer(layername2, source = srcname,
								  source_layer  = smeta$layer,
								  line_color    = aes_c,
								  line_opacity  = aes_co,
								  line_width    = aes_lwd) |>
			assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

		mapgl_submit_group(group, c(layername1, layername2), mode, pane)
		return(NULL)
	}

	# ----------------------------------------------------------
	#  sf branch
	# ----------------------------------------------------------
	res = select_sf(shpTM, dt)
	shp = res$shp
	dt  = res$dt

	x = sf::st_sfc(list(sf::st_point(c(mean(bbx[c(1, 3)]), bbx[2]))), crs = sf::st_crs(bbx))
	y = sf::st_sfc(list(sf::st_point(c(mean(bbx[c(1, 3)]), bbx[4]))), crs = sf::st_crs(bbx))

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	if (any(nchar(gp$fill) == 9)) {
		fa        = split_alpha_channel(gp$fill, alpha = gp$fill_alpha)
		gp$fill   = fa$col
		gp$fill_alpha = gp$fill_alpha * fa$opacity
	}
	if (any(nchar(gp$col) == 9)) {
		fa        = split_alpha_channel(gp$col, alpha = gp$col_alpha)
		gp$col    = fa$col
		gp$col_alpha = gp$fill_alpha * fa$opacity
	}

	shp2 = sf::st_sf(
		unclass(gp[c("fill", "col", "lwd", "fill_alpha", "col_alpha")]),
		id       = 1:length(shp),
		geometry = shp
	)

	srcname    = mapgl_srcid(paste0("layer", pane))
	layername1 = paste0(glid, "polygons_fill")
	layername2 = paste0(glid, "polygons_border")

	# Per-aesthetic categorical columns for the interactive legend(s).
	shp2 = mapgl_attach_cat(shp2, glid, fill_layer = layername1, border_layer = layername2)

	ahp  = attach_hover_popup(shp2, dt, hdt, pdt, idt, popup.format, popup.layout, ptdt)
	shp2 = ahp$shp2

	m |>
		mapgl::add_source(srcname, data = shp2) |>
		mapgl::add_fill_layer(layername1, source = srcname,
							  fill_color   = mapgl::get_column("fill"),
							  fill_opacity = mapgl::get_column("fill_alpha"),
							  tooltip      = ahp$hdt_arg,
							  popup        = ahp$pdt_arg) |>
		mapgl::add_line_layer(layername2, source = srcname,
							  line_color   = mapgl::get_column("col"),
							  line_opacity = mapgl::get_column("col_alpha"),
							  line_width   = mapgl::get_column("lwd")) |>
		assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

	mapgl_submit_group(group, c(layername1, layername2), mode, pane)
	NULL
}

# ============================================================
#  mapgl_polygons_3d
# ============================================================
# fill-extrusion paint props added in mapgl >= 0.4.6. Forward ONLY the props
# the user set AND that the active engine supports: MapLibre's style spec only
# implements fill-extrusion-vertical-gradient, whereas ambient occlusion, cast
# shadows and emissive strength are Mapbox GL JS 3D-lighting features. Sending
# an unknown paint property makes MapLibre reject the entire layer (the polygons
# then render with no fill), so Mapbox-only props are dropped in maplibre mode
# with an informative message. (If a future MapLibre adds e.g. emissive strength
# to its spec, flip that entry's `mapbox_only` to FALSE.)
mapgl_fe_props = function(a, mode) {
	specs = list(
		list(arg = "cast.shadows",                param = "fill_extrusion_cast_shadows",                mapbox_only = TRUE),
		list(arg = "ambient.occlusion.intensity", param = "fill_extrusion_ambient_occlusion_intensity", mapbox_only = TRUE),
		list(arg = "ambient.occlusion.radius",    param = "fill_extrusion_ambient_occlusion_radius",    mapbox_only = TRUE),
		list(arg = "vertical.gradient",           param = "fill_extrusion_vertical_gradient",           mapbox_only = FALSE),
		list(arg = "emissive.strength",           param = "fill_extrusion_emissive_strength",           mapbox_only = TRUE)
	)
	is_mapbox = identical(mode, "mapbox")

	out     = list()
	dropped = character(0)
	for (s in specs) {
		val = a[[s$arg]]
		if (is.null(val)) next
		if (s$mapbox_only && !is_mapbox) {
			dropped = c(dropped, s$arg)
		} else {
			out[[s$param]] = val
		}
	}
	if (length(dropped)) {
		cli::cli_inform(c(
			"!" = "These {.fn tm_polygons_3d} options are Mapbox-only and were ignored in {.str {mode}} mode: {.field {dropped}}.",
			"i" = "Use {.code tmap_mode(\"mapbox\")} for ambient occlusion, cast shadows, and emissive strength."
		))
	}
	out
}

mapgl_polygons_3d = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp,
							  bbx, facet_row, facet_col, facet_page,
							  id, pane, group, glid, o, ..., mode, popup.layout = NULL, ptdt = NULL) {

	m       = get_mapgl(facet_row, facet_col, facet_page, mode)
	mapgl_blend_unsupported(a, mode)
	rc_text = frc(facet_row, facet_col)
	shp_is_pointer = inherits(shpTM$shp, "character")

	# ----------------------------------------------------------
	#  PMTiles branch
	# ----------------------------------------------------------
	if (shp_is_pointer) {
		if (pmtiles_unsupported(shpTM, mode)) return(NULL)

		smeta      = shpTM$smeta
		srcname    = mapgl_srcid(paste0("layer", pane))
		layername1 = paste0(glid, "polygons_fill")
		layername2 = paste0(glid, "polygons_border")
		url        = smeta$url

		get_pmt_aes = make_get_pmt_aes(dt)

		aes_f   = get_pmt_aes("fill")
		aes_fo  = get_pmt_aes("fill_alpha")
		aes_c   = get_pmt_aes("col")
		aes_co  = get_pmt_aes("col_alpha")
		aes_lwd = get_pmt_aes("lwd")
		aes_h_raw = get_pmt_aes("height")

		# --- read limits from tm_scale_continuous -------------------------------
		h_var = dt[["height"]][1]
		if (substr(h_var, 1, 5) == "scale") {
			snr = as.numeric(substr(h_var, 6, 8))
			leg = .TMAP$legs[[snr]]
			limits = leg$layer_args$limits
			if (is.null(limits)) {
				cli::cli_abort(c(
					"!" = "Cannot determine height range for PMTiles source.",
					"i" = "Specify {.code height.scale = tm_scale_continuous(limits = c(min, max))}.",
					"i" = "Example: {.code tm_scale_continuous(limits = c(0, 28672))}"
				))
			}
			h_min_raw = limits[1]
			h_max_raw = limits[2]
		} else {
			cli::cli_abort(c(
				"!" = "Height aesthetic for PMTiles must use {.fn tm_scale_continuous}.",
				"i" = "Specify {.code height.scale = tm_scale_continuous(limits = c(min, max))}."
			))
		}

		# --- resolve height.max / height.min to metres -------------------------
		if (is.character(a$height.max)) {
			is_perc_max = grepl("%$", a$height.max)
			height.max  = as.numeric(sub("%$", "", a$height.max))
			if (is_perc_max) height.max = height.max / 100
		} else {
			is_perc_max = FALSE
			height.max  = as.numeric(a$height.max)
		}

		if (is.character(a$height.min)) {
			is_perc_min = grepl("%$", a$height.min)
			height.min  = as.numeric(sub("%$", "", a$height.min))
			if (is_perc_min) height.min = height.min / 100
		} else {
			is_perc_min = FALSE
			height.min  = as.numeric(a$height.min)
		}

		if (is_perc_max || is_perc_min) {
			sqrt_area_m = if (consider_global(bbx)) {
				sqrt(5.1e+14)
			} else {
				bbx |>
					tmaptools::bb_poly() |>
					sf::st_area() |>
					sqrt() |>
					units::set_units("m") |>
					units::drop_units()
			}
			if (is_perc_max) height.max = sqrt_area_m * height.max
			if (is_perc_min) height.min = sqrt_area_m * height.min
		}

		# --- build MapLibre scaling expression ---------------------------------
		# Replicates: height.min + (raw - h_min_raw) / (h_max_raw - h_min_raw)
		#                        * (height.max - height.min)
		h_range   = h_max_raw - h_min_raw
		out_range = height.max - height.min

		aes_h_scaled = list(
			"+",
			height.min,
			list("*",
				 list("/",
				 	 list("-", aes_h_raw, h_min_raw),
				 	 h_range),
				 out_range)
		)

		m = m |>
			mapgl::add_pmtiles_source(id = srcname, url = url) |>
			mapgl::add_line_layer(layername2, source = srcname,
								  source_layer          = smeta$layer,
								  line_color            = aes_c,
								  line_opacity          = aes_co,
								  line_width            = aes_lwd)

		m = do.call(mapgl::add_fill_extrusion_layer, c(
			list(m, layername1,
				 source                 = srcname,
				 source_layer           = smeta$layer,
				 fill_extrusion_color   = aes_f,
				 fill_extrusion_base    = 0,
				 fill_extrusion_height  = aes_h_scaled),
			mapgl_fe_props(a, mode)))

		m |> assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

		mapgl_submit_group(group, c(layername1, layername2), mode, pane)
		return(NULL)
	}

	# ----------------------------------------------------------
	#  sf branch (unchanged)
	# ----------------------------------------------------------
	res = select_sf(shpTM, dt)
	shp = res$shp
	dt  = res$dt

	x = sf::st_sfc(list(sf::st_point(c(mean(bbx[c(1, 3)]), bbx[2]))), crs = sf::st_crs(bbx))
	y = sf::st_sfc(list(sf::st_point(c(mean(bbx[c(1, 3)]), bbx[4]))), crs = sf::st_crs(bbx))

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	if (any(nchar(gp$fill) == 9)) {
		fa            = split_alpha_channel(gp$fill, alpha = gp$fill_alpha)
		gp$fill       = fa$col
		gp$fill_alpha = gp$fill_alpha * fa$opacity
	}
	if (any(nchar(gp$col) == 9)) {
		fa           = split_alpha_channel(gp$col, alpha = gp$col_alpha)
		gp$col       = fa$col
		gp$col_alpha = gp$fill_alpha * fa$opacity
	}

	shp2 = sf::st_sf(
		unclass(gp[c("height", "fill", "col", "lwd", "fill_alpha", "col_alpha")]),
		id       = 1:length(shp),
		geometry = shp
	)

	# NOTE on transparency: fill-extrusion layers do NOT support per-feature alpha.
	# The GL spec ignores the alpha channel of fill-extrusion-color (and a long-
	# standing engine bug renders alpha=0 as solid black), and fill-extrusion-opacity
	# is per-LAYER only, not data-driven. So we honour transparency as a single
	# layer-wide opacity, taken as the mean of the per-feature fill_alpha values.
	fill_ext_opacity = {
		fa = suppressWarnings(as.numeric(shp2$fill_alpha))
		fa = fa[is.finite(fa)]
		if (!length(fa)) 1 else min(max(mean(fa), 0), 1)
	}

	ahp  = attach_hover_popup(shp2, dt, hdt, pdt, idt, popup.format, popup.layout, ptdt)
	shp2 = ahp$shp2

	srcname    = mapgl_srcid(paste0("layer", pane))
	layername1 = paste0(glid, "polygons_fill")
	layername2 = paste0(glid, "polygons_border")

	if (is.character(a$height.max)) {
		is_perc_max = grepl("%$", a$height.max)
		height.max  = as.numeric(sub("%$", "", a$height.max))
		if (is_perc_max) height.max = height.max / 100
	} else {
		is_perc_max = FALSE
		height.max  = as.numeric(a$height.max)
	}

	if (is.character(a$height.min)) {
		is_perc_min = grepl("%$", a$height.min)
		height.min  = as.numeric(sub("%$", "", a$height.min))
		if (is_perc_min) height.min = height.min / 100
	} else {
		is_perc_min = FALSE
		height.min  = as.numeric(a$height.min)
	}

	if (is_perc_max || is_perc_min) {
		sqrt_area_m = if (consider_global(shp)) {
			sqrt(5.1e+14)
		} else {
			bbx |>
				tmaptools::bb_poly() |>
				sf::st_area() |>
				sqrt() |>
				units::set_units("m") |>
				units::drop_units()
		}
		if (is_perc_max) height.max = sqrt_area_m * height.max
		if (is_perc_min) height.min = sqrt_area_m * height.min
	}

	shp2$height = height.min + shp2$height * (height.max - height.min)
	shp2_naomit = shp2[!is.na(shp2$height), ]

	m = m |>
		mapgl::add_source(srcname, data = shp2_naomit) |>
		mapgl::add_line_layer(layername2, source = srcname,
							  line_color   = mapgl::get_column("col"),
							  line_opacity = mapgl::get_column("col_alpha"),
							  line_width   = mapgl::get_column("lwd"))

	m = do.call(mapgl::add_fill_extrusion_layer, c(
		list(m, layername1,
			 source                 = srcname,
			 fill_extrusion_color   = mapgl::get_column("fill"),
			 fill_extrusion_opacity = fill_ext_opacity,
			 fill_extrusion_base    = 0,
			 fill_extrusion_height  = mapgl::get_column("height"),
			 tooltip                = ahp$hdt_arg,
			 popup                  = ahp$pdt_arg),
		mapgl_fe_props(a, mode)))

	m |> assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

	mapgl_submit_group(group, c(layername1, layername2), mode, pane)
	NULL
}

# ============================================================
#  mapgl_lines
# ============================================================
mapgl_lines = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp,
						bbx, facet_row, facet_col, facet_page,
						id, pane, group, glid, o, ..., mode, popup.layout = NULL, ptdt = NULL) {
	m          = get_mapgl(facet_row, facet_col, facet_page, mode = mode)
	mapgl_blend_unsupported(a, mode)
	rc_text    = frc(facet_row, facet_col)
	shp_is_pointer = inherits(shpTM$shp, "character")
	# ----------------------------------------------------------
	#  PMTiles branch
	# ----------------------------------------------------------
	if (shp_is_pointer) {
		if (pmtiles_unsupported(shpTM, mode)) return(NULL)
		smeta      = shpTM$smeta
		srcname    = mapgl_srcid(paste0("layer", pane))
		layername1 = paste0(glid, "lines")
		url        = smeta$url
		get_pmt_aes = make_get_pmt_aes(dt)
		aes_c   = get_pmt_aes("col")
		aes_co  = get_pmt_aes("col_alpha")
		aes_lwd = get_pmt_aes("lwd")
		m |>
			mapgl::add_pmtiles_source(id = srcname, url = url) |>
			mapgl::add_line_layer(layername1, source = srcname,
								  source_layer  = smeta$layer,
								  line_color    = aes_c,
								  line_opacity  = aes_co,
								  line_width    = aes_lwd) |>
			assign_mapgl(facet_row, facet_col, facet_page, mode = mode)
		mapgl_submit_group(group, layername1, mode, pane)
		return(NULL)
	}
	# ----------------------------------------------------------
	#  sf branch
	# ----------------------------------------------------------
	res = select_sf(shpTM, dt)
	shp = res$shp
	dt  = res$dt
	x = sf::st_sfc(list(sf::st_point(c(mean(bbx[c(1, 3)]), bbx[2]))), crs = sf::st_crs(bbx))
	y = sf::st_sfc(list(sf::st_point(c(mean(bbx[c(1, 3)]), bbx[4]))), crs = sf::st_crs(bbx))
	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)
	shp2 = sf::st_sf(
		unclass(gp[c("col", "lwd", "col_alpha")]),
		id       = 1:length(shp),
		geometry = shp
	)

	srcname    = mapgl_srcid(paste0("layer", pane))
	layername1 = paste0(glid, "lines")  # was paste0(srcname, "lines") — fixed to match legend

	# Lines have only a col (line-colour) aesthetic on a single line layer.
	shp2 = mapgl_attach_cat(shp2, glid, fill_layer = NA_character_, border_layer = layername1,
							fill_col = NULL, col_col = "col")

	ahp  = attach_hover_popup(shp2, dt, hdt, pdt, idt, popup.format, popup.layout, ptdt)
	shp2 = ahp$shp2

	m |>
		mapgl::add_source(srcname, data = shp2) |>
		mapgl::add_line_layer(layername1, source = srcname,
							  line_color   = mapgl::get_column("col"),
							  line_opacity = mapgl::get_column("col_alpha"),
							  line_width   = mapgl::get_column("lwd"),
							  tooltip      = ahp$hdt_arg,
							  popup        = ahp$pdt_arg) |>
		assign_mapgl(facet_row, facet_col, facet_page, mode = mode)
	mapgl_submit_group(group, layername1, mode, pane)
	NULL
}

# ============================================================
#  mapgl_symbols
# ============================================================

mapgl_symbols = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp,
						  bbx, facet_row, facet_col, facet_page,
						  id, pane, group, glid, o, ..., mode, popup.layout = NULL, ptdt = NULL) {

	m          = get_mapgl(facet_row, facet_col, facet_page, mode)
	mapgl_blend_unsupported(a, mode)
	rc_text    = frc(facet_row, facet_col)
	shp_is_pointer = inherits(shpTM$shp, "character")

	# ----------------------------------------------------------
	#  PMTiles branch
	# ----------------------------------------------------------
	if (shp_is_pointer) {
		if (pmtiles_unsupported(shpTM, mode)) return(NULL)

		smeta      = shpTM$smeta
		srcname    = mapgl_srcid(paste0("layer", pane))
		layername1 = paste0(glid, "symbols_fill")
		url        = smeta$url

		get_pmt_aes = make_get_pmt_aes(dt)

		aes_f    = get_pmt_aes("fill")
		aes_fo   = get_pmt_aes("fill_alpha")
		aes_c    = get_pmt_aes("col")
		aes_co   = get_pmt_aes("col_alpha")
		aes_lwd  = get_pmt_aes("lwd")
		aes_size = get_pmt_aes("size")

		# Mirror the *10 scaling applied in the sf branch for numeric constants
		if (is.numeric(aes_size)) aes_size = aes_size * 10

		m |>
			mapgl::add_pmtiles_source(id = srcname, url = url) |>
			mapgl::add_circle_layer(layername1, source = srcname,
									source_layer          = smeta$layer,
									circle_color          = aes_f,
									circle_opacity        = aes_fo,
									circle_stroke_color   = aes_c,
									circle_stroke_opacity = aes_co,
									circle_stroke_width   = aes_lwd,
									circle_radius         = aes_size) |>
			assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

		mapgl_submit_group(group, layername1, mode, pane)
		return(NULL)
	}

	# ----------------------------------------------------------
	#  sf branch
	# ----------------------------------------------------------
	res = select_sf(shpTM, dt)
	shp = res$shp
	dt  = res$dt

	x = sf::st_sfc(list(sf::st_point(c(mean(bbx[c(1, 3)]), bbx[2]))), crs = sf::st_crs(bbx))
	y = sf::st_sfc(list(sf::st_point(c(mean(bbx[c(1, 3)]), bbx[4]))), crs = sf::st_crs(bbx))

	gp = impute_gp(gp, dt)
	gp = rescale_gp(gp, o$scale_down)

	# ----------------------------------------------------------
	#  Grob-shape branch (custom glyphs)
	# ----------------------------------------------------------
	shape_codes = suppressWarnings(as.integer(gp$shape))
	is_grob     = !is.na(shape_codes) & shape_codes > 999L
	if (any(is_grob)) {
		# In non-Grid modes tmapValuesSubmit_shape() already rasterised each grob
		# via grob2icon(); shapeLib holds leaflet icon objects (iconUrl -> a PNG),
		# not grobs. Reuse those PNGs directly, exactly like the Leaflet backend.
		merge_icons = utils::getFromNamespace("merge_icons", "tmap")
		iconset     = merge_icons(get("shapeLib", envir = .TMAP)[shape_codes[is_grob] - 999L])

		gd = a$grob.dim
		rw = if (!is.null(gd)) gd[["render.width"]] else 256   # PNG native px; icon-size is relative to this

		# glyph_scale: per-backend calibration knob (mapgl analogue of the
		# plot/view donut factors) — tune against the legend / view mode.
		glyph_scale = a$icon.scale
		iconsize    = (gp$size * 10 * glyph_scale) / rw

		ids          = rep("", length(shape_codes))
		ids[is_grob] = paste0(glid, "_glyph_", which(is_grob))
		m            = mapgl_add_glyph_images(m, iconset$iconUrl, ids[is_grob])

		shp2 = sf::st_sf(icon = ids, iconsize = iconsize,
						 id = seq_along(shape_codes), geometry = shp)

		srcname    = mapgl_srcid(paste0("layer", pane))
		layername1 = paste0(glid, "symbols_fill")

		shp2 = mapgl_attach_cat(shp2, glid, fill_layer = layername1,
								border_layer = NA_character_, fill_col = NULL, col_col = NULL)
		ahp  = attach_hover_popup(shp2, dt, hdt, pdt, idt, popup.format, popup.layout, ptdt)
		shp2 = ahp$shp2

		m |>
			mapgl::add_source(srcname, data = shp2) |>
			mapgl::add_symbol_layer(layername1, source = srcname,
									icon_image            = mapgl::get_column("icon"),
									icon_size             = mapgl::get_column("iconsize"),
									icon_allow_overlap    = TRUE,
									icon_ignore_placement = TRUE,
									icon_anchor           = "center",
									tooltip               = ahp$hdt_arg,
									popup                 = ahp$pdt_arg) |>
			assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

		mapgl_submit_group(group, layername1, mode, pane)
		return(NULL)
	}

	if (any(nchar(gp$fill) == 9)) {
		fa        = split_alpha_channel(gp$fill, alpha = gp$fill_alpha)
		gp$fill   = fa$col
		gp$fill_alpha = gp$fill_alpha * fa$opacity
	}
	if (any(nchar(gp$col) == 9)) {
		fa        = split_alpha_channel(gp$col, alpha = gp$col_alpha)
		gp$col    = fa$col
		gp$col_alpha = gp$fill_alpha * fa$opacity
	}

	shp2 = sf::st_sf(
		unclass(gp[c("fill", "col", "lwd", "fill_alpha", "col_alpha", "size")]),
		id       = 1:length(shp),
		geometry = shp
	)

	shp2$size = shp2$size * 10

	srcname    = mapgl_srcid(paste0("layer", pane))
	layername1 = paste0(glid, "symbols_fill")  # was paste0(srcname, ...) — fixed

	# tm_symbols renders as a single circle layer carrying both circle-color
	# (fill) and circle-stroke-color (col); there is no separate outline layer,
	# so border_layer = NA. Both legends therefore key off this one layer (and,
	# per mapgl, only one can own its interactive filter at a time).
	shp2 = mapgl_attach_cat(shp2, glid, fill_layer = layername1, border_layer = NA_character_)

	ahp  = attach_hover_popup(shp2, dt, hdt, pdt, idt, popup.format, popup.layout, ptdt)
	shp2 = ahp$shp2

	m |>
		mapgl::add_source(srcname, data = shp2) |>
		mapgl::add_circle_layer(layername1, source = srcname,
								circle_color          = mapgl::get_column("fill"),
								circle_opacity        = mapgl::get_column("fill_alpha"),
								circle_stroke_color   = mapgl::get_column("col"),
								circle_stroke_opacity = mapgl::get_column("col_alpha"),
								circle_stroke_width   = mapgl::get_column("lwd"),
								circle_radius         = mapgl::get_column("size"),
								tooltip               = ahp$hdt_arg,
								popup                 = ahp$pdt_arg) |>
		assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

	mapgl_submit_group(group, layername1, mode, pane)
	NULL
}

mapgl_add_glyph_images = function(m, urls, ids) {
	for (i in seq_along(urls)) m = mapgl::add_image(m, id = ids[i], url = urls[i])
	m
}

# Build smooth great-circle polygons (one per point) in EPSG:4326. Each circle
# is an `npoints`-vertex polygon generated from the spherical geodesic
# destination formula, so it is ground-accurate (radius in metres) and renders
# smoothly at any zoom — unlike an s2 cell-buffer, whose boundary is faceted.
# Longitudes are left continuous around the centre (not wrapped to [-180, 180])
# so circles near the antimeridian stay contiguous rather than splitting.
mapgl_circle_polys = function(pts, radius_m, npoints = 90L) {
	R   = 6378137                       # WGS84 mean Earth radius (m)
	cc  = sf::st_coordinates(pts)       # columns X (lon), Y (lat) in degrees
	n   = nrow(cc)
	rad = rep_len(radius_m / R, n)      # angular radius (radians)
	ang = seq(0, 2 * pi, length.out = npoints + 1L)

	polys = lapply(seq_len(n), function(i) {
		lon0 = cc[i, 1] * pi / 180
		lat0 = cc[i, 2] * pi / 180
		d    = rad[i]
		lat2 = asin(sin(lat0) * cos(d) + cos(lat0) * sin(d) * cos(ang))
		lon2 = lon0 + atan2(sin(ang) * sin(d) * cos(lat0),
							 cos(d) - sin(lat0) * sin(lat2))
		ring = cbind(lon2 * 180 / pi, lat2 * 180 / pi)
		ring[npoints + 1L, ] = ring[1L, ]   # close the ring exactly
		sf::st_polygon(list(ring))
	})
	sf::st_sfc(polys, crs = 4326)
}

# ============================================================
#  mapgl_circles
# ============================================================
#
# Geographically-fixed circles: the radius is expressed in metres (per the
# tm_circles() contract, gp$size is already a metre radius — the scale output
# is used directly and rescale_gp is NOT applied to size), so the circles scale
# with zoom and stay anchored to the map, unlike tm_symbols()/tm_bubbles()
# whose circle-radius is a fixed number of screen pixels.
#
# MapLibre's circle-radius paint property is in PIXELS, so a circle layer would
# be screen-fixed. To get true metre-based, zoom-scaling circles we materialise
# each point as a geodesic buffer polygon (in EPSG:4326, using s2 so the radius
# is in ground metres regardless of the data CRS — this also sidesteps the Web
# Mercator cos(lat) distortion that bit the grid backend) and render it exactly
# like a polygon layer: a fill layer for the body and a line layer for the
# border. The fill layer is named "<glid>symbols_fill" so the interactive
# categorical legend (which keys off cmp$layer containing "symbols") keeps
# matching it.

mapgl_circles = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp,
						  bbx, facet_row, facet_col, facet_page,
						  id, pane, group, glid, o, ..., mode, popup.layout = NULL, ptdt = NULL) {

	m          = get_mapgl(facet_row, facet_col, facet_page, mode)
	mapgl_blend_unsupported(a, mode)
	rc_text    = frc(facet_row, facet_col)
	shp_is_pointer = inherits(shpTM$shp, "character")

	# ----------------------------------------------------------
	#  PMTiles branch — unsupported
	# ----------------------------------------------------------
	# Metre-radius circles require buffering each point geometry at render time,
	# but for a remote tile source the per-feature geometry/data join is not
	# available in R memory (same limitation as hover/popup over PMTiles). Fall
	# back gracefully.
	if (shp_is_pointer) {
		cli::cli_inform(c(
			"!" = "{.fn tm_circles} (metre-based radii) is not supported for source shapes (e.g. PMTiles) in {.str {mode}} mode.",
			"i" = "Use {.fn tm_symbols} / {.fn tm_bubbles} for pixel-based circles over tiled sources."
		))
		return(NULL)
	}

	# ----------------------------------------------------------
	#  sf branch
	# ----------------------------------------------------------
	res = select_sf(shpTM, dt)
	shp = res$shp
	dt  = res$dt

	gp = impute_gp(gp, dt)

	# Preserve the metre radius before rescale_gp (which would convert size to
	# "lines" / apply scale_down — meaningless for a geographic radius). lwd
	# (border width, in px) is left to rescale like a normal polygon border.
	size_raw = gp$size
	gp = rescale_gp(gp, o$scale_down)
	gp$size = size_raw

	if (any(nchar(gp$fill) == 9)) {
		fa            = split_alpha_channel(gp$fill, alpha = gp$fill_alpha)
		gp$fill       = fa$col
		gp$fill_alpha = gp$fill_alpha * fa$opacity
	}
	if (any(nchar(gp$col) == 9)) {
		fa           = split_alpha_channel(gp$col, alpha = gp$col_alpha)
		gp$col       = fa$col
		gp$col_alpha = gp$fill_alpha * fa$opacity
	}

	# --- materialise circles as smooth geodesic polygons in EPSG:4326 --------
	# st_buffer() with s2 approximates the disc with S2 cells, whose boundary
	# follows cell edges and looks faceted ("pixelated"). Instead we generate a
	# proper great-circle polygon with `npoints` vertices per circle: smooth,
	# tunable, ground-accurate, and with no global s2-flag side effects.
	pts4326  = sf::st_transform(sf::st_geometry(shp), 4326)
	radius_m = as.numeric(gp$size)   # metres

	circ = mapgl_circle_polys(pts4326, radius_m, npoints = 90L)

	shp2 = sf::st_sf(
		unclass(gp[c("fill", "col", "lwd", "fill_alpha", "col_alpha", "size")]),
		id       = seq_along(shp),
		geometry = circ
	)

	# drop features without a finite radius (NA-sized) — they have no geometry
	keep = is.finite(radius_m) & radius_m > 0
	shp2 = shp2[keep, ]

	# --- Add categorical columns for interactive legend(s) -------------------
	# tm_circles is materialised as fill + line polygon layers, so the fill and
	# col legends can each own their own layer (like tm_polygons).
	srcname    = mapgl_srcid(paste0("layer", pane))
	layername1 = paste0(glid, "symbols_fill")    # matches the categorical legend
	layername2 = paste0(glid, "symbols_border")

	shp2 = mapgl_attach_cat(shp2, glid, fill_layer = layername1, border_layer = layername2)
	# -------------------------------------------------------------------------

	ahp  = attach_hover_popup(shp2, dt[keep, ], hdt, pdt, idt, popup.format, popup.layout, ptdt)
	shp2 = ahp$shp2

	m |>
		mapgl::add_source(srcname, data = shp2) |>
		mapgl::add_fill_layer(layername1, source = srcname,
							  fill_color   = mapgl::get_column("fill"),
							  fill_opacity = mapgl::get_column("fill_alpha"),
							  tooltip      = ahp$hdt_arg,
							  popup        = ahp$pdt_arg) |>
		mapgl::add_line_layer(layername2, source = srcname,
							  line_color   = mapgl::get_column("col"),
							  line_opacity = mapgl::get_column("col_alpha"),
							  line_width   = mapgl::get_column("lwd")) |>
		assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

	mapgl_submit_group(group, c(layername1, layername2), mode, pane)
	NULL
}

# ============================================================
#  mapgl_raster  (unchanged — reproduced for completeness)
# ============================================================

mapgl_raster = function(a, shpTM, dt, gp, pdt, popup.format, hdt, idt,
						 bbx, facet_row, facet_col, facet_page,
						 id, pane, group, glid, o, ..., mode) {

	rc_text  = frc(facet_row, facet_col)
	shp      = shpTM$shp
	tmapID   = shpTM$tmapID
	shp_is_pointer = inherits(shp, "character")

	if (shp_is_pointer) {
		if (pmtiles_unsupported(shpTM, mode)) return(NULL)

		smeta      = shpTM$smeta
		srcname    = mapgl_srcid(paste0("layer", pane))
		layername1 = paste0(glid, "raster")
		url        = smeta$url

		m = get_mapgl(facet_row, facet_col, facet_page, mode = mode)

		m |>
			mapgl::add_pmtiles_source(id = srcname, url = url, source_type = "raster") |>
			mapgl::add_raster_layer(layername1, source = srcname) |>
			assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

		return(NULL)

	} else if (is_regular_grid(shp)) {

		tid   = intersect(tmapID, dt$tmapID__)
		color = rep(NA, length(tmapID))
		sel   = which(tmapID %in% tid)
		tid2  = tmapID[sel]
		color[sel] = dt$col[match(tid2, dt$tmapID__)]

		pal = na.omit(unique(color))
		pal = pal[substr(pal, 8, 10) != "00"]

		if (!length(pal)) return(NULL)

		res         = split_alpha_channel(pal, alpha = 1)
		pal_col     = res$col
		pal_opacity = if (length(res$opacity) == 0L) 0 else max(res$opacity)

		if ("col_alpha" %in% names(dt)) pal_opacity = max(dt$col_alpha)

		col_ids = match(color, pal)
		m_mat   = matrix(col_ids, ncol = ncol(shp))
		shp2    = stars::st_as_stars(m_mat, dimensions = shp)
		rst     = terra::rast(shp2)

		if (!terra::is.lonlat(rst)) rst = terra::project(rst, "epsg:4326")

		ext = terra::ext(rst)
		if (ext$ymin < -89.9) ext$ymin = -89
		if (ext$ymax >  89.9) ext$ymax =  89
		rst2 = terra::crop(rst, ext)

		srcname    = mapgl_srcid(paste0("layer", pane))
		layername1 = paste0(glid, "raster")

		m = get_mapgl(facet_row, facet_col, facet_page, mode = mode)

		m |>
			mapgl::add_image_source(srcname, data = rst2, colors = pal) |>
			mapgl::add_raster_layer(layername1, source = srcname,
									raster_opacity    = pal_opacity,
									raster_resampling = "nearest") |>
			assign_mapgl(facet_row, facet_col, facet_page, mode = mode)

		mapgl_submit_group(group, layername1, mode, pane)

	} else {

		m_mat = matrix(tmapID, nrow = nrow(shp), ncol = ncol(shp))
		shp2  = structure(list(tmapID = m_mat), class = "stars", dimensions = shp)
		shp3  = sf::st_geometry(sf::st_as_sf(shp2))
		crs   = get_option_class(o$crs_step4, "sf")
		shpTM = tmap::shapeTM(sf::st_transform(shp3, crs), tmapID)

		gp$lty = "solid"

		mapgl_polygons(a, shpTM, dt, pdt,
					   popup.format = NULL, hdt = NULL, idt = NULL,
					   gp, bbx, facet_row, facet_col, facet_page,
					   id, pane, group, glid, o, mode = mode)
	}
	NULL
}

#' @param bbx bbx
#' @export
#' @keywords internal
#' @name tmapMapboxShape
#' @rdname tmapMapbox
tmapMapboxShape = function(bbx, facet_row, facet_col, facet_page, o) {
	mapgl_shape(bbx = bbx,
				facet_row = facet_row,
				facet_col = facet_col,
				facet_page = facet_page,
				o = o,
				mode = "mapbox")
}

#' @export
#' @rdname tmapMapbox
tmapMaplibreShape = function(bbx, facet_row, facet_col, facet_page, o) {
	mapgl_shape(bbx = bbx,
				facet_row = facet_row,
				facet_col = facet_col,
				facet_page = facet_page,
				o = o,
				mode = "maplibre")
}

get_style = function(name) {
	x = sub("^[A-Za-z]+[._]([A-Za-z]+(?:[._][A-Za-z]+)*)$", "\\1", name)
	chartr("._", "--", x = x)
}

# Resolve a tmap basemap provider name to a mapgl style URL/spec. Provider names
# are "<family>.<style>[.<variant>]" (e.g. "carto.positron",
# "mapbox.satellite_streets", "maptiler.dataviz.dark", "esri.human_geography",
# "ofm.liberty"). Unknown or prefixless names are returned verbatim, so a raw
# style URL passed by the user still works. esri/maptiler require an API key
# (ARCGIS_API_KEY / MAPTILER_API_KEY); mapbox styles require a Mapbox token and
# are only offered in mapbox mode.
# Resolve a tmap basemap provider name to a mapgl style URL/spec. Provider names
# are "<family>.<style>[.<variant>]" (e.g. "carto.positron",
# "mapbox.satellite_streets", "maptiler.dataviz.dark", "esri.human_geography",
# "ofm.liberty"). Unknown or prefixless names are returned verbatim, so a raw
# style URL passed by the user still works.
#
# esri/maptiler need an API key (ARCGIS_API_KEY / MAPTILER_API_KEY env var, or
# `api_key` passed through from tm_basemap). If the key is missing, we warn and
# fall back to `default` - the caller passes the mode's basemap option
# (o$basemap.server[1]); the literal here is only a last-resort safety net.
#
# `bg` is the layout background colour (o$bg.color). When no basemap is applied
# (e.g. tm_basemap(NULL)) it is used to paint a background layer, so the blank
# canvas matches tm_layout(bg.color = ...), the way view mode colours the
# leaflet container.
resolve_style = function(name, api_key = NULL, default = "ofm.positron", bg = NULL) {
	fallback = function(msg) {
		cli::cli_warn(c("!" = msg,
						"i" = "Falling back to the default basemap {.str {default}}."))
		resolve_style(default, default = default, bg = bg)
	}

	# An empty / NULL style means no basemap was applied (e.g. tm_basemap(NULL)):
	# return a minimal blank GL style so the engine renders no basemap (like view
	# mode), instead of erroring in regexpr() below. If a background colour is
	# available, add a background layer so the canvas honours tm_layout(bg.color).
	if (is.null(name) || length(name) != 1L || is.na(name) || !nzchar(name)) {
		layers = list()
		if (!is.null(bg) && length(bg) == 1L && !is.na(bg) && nzchar(bg)) {
			# Normalise to #RRGGBB (handles R colour names like "grey90" and hex
			# with alpha); fall back to the raw value if it can't be parsed.
			bg_css = tryCatch(
				grDevices::rgb(t(grDevices::col2rgb(bg)), maxColorValue = 255),
				error = function(e) bg)
			layers = list(list(id = "background", type = "background",
							   paint = list(`background-color` = bg_css)))
		}
		return(list(version = 8L,
					sources = structure(list(), names = character(0)),
					layers  = layers))
	}

	pos = regexpr("[._]", name)
	if (pos < 1L) return(name)
	family = substr(name, 1L, pos - 1L)
	rest   = substr(name, pos + 1L, nchar(name))
	dash   = function(x) chartr("._", "--", x)

	have_key = function(env) (!is.null(api_key) && nzchar(api_key)) || nzchar(Sys.getenv(env))

	switch(family,
		   ofm      = paste0("https://tiles.openfreemap.org/styles/", dash(rest)),
		   carto    = mapgl::carto_style(dash(rest)),
		   mapbox   = mapgl::mapbox_style(dash(rest)),
		   esri     = if (have_key("ARCGIS_API_KEY")) {
		   	mapgl::esri_style(dash(rest), token = api_key)
		   } else {
		   	fallback("Esri basemap {.str {name}} needs an ArcGIS API key (set env var {.envvar ARCGIS_API_KEY}, or pass {.arg api_key} to {.fn tm_basemap}).")
		   },
		   maptiler = if (!have_key("MAPTILER_API_KEY")) {
		   	fallback("MapTiler basemap {.str {name}} needs a MapTiler API key (set env var {.envvar MAPTILER_API_KEY}, or pass {.arg api_key} to {.fn tm_basemap}).")
		   } else {
		   	parts   = strsplit(rest, "[._]")[[1]]
		   	variant = NULL
		   	if (length(parts) > 1L && parts[length(parts)] %in% c("dark", "light", "pastel")) {
		   		variant = parts[length(parts)]
		   		parts   = parts[-length(parts)]
		   	}
		   	mapgl::maptiler_style(paste(parts, collapse = "-"), variant = variant, api_key = api_key)
		   },
		   name)
}

mapgl_shape = function(bbx, facet_row, facet_col, facet_page, o, mode) {
	# In proxy mode the canvas is the proxy object that mapgl_init() placed in
	# e$ms. Building a fresh base map here would discard the proxy and render a
	# throwaway widget (the one leaking into the RStudio viewer). Skip it.
	if (isTRUE(.TMAP$proxy)) return(NULL)

	m = get_mapgl(facet_row, facet_col, facet_page, mode)
	bbx = sf::st_bbox(sf::st_transform(tmaptools::bb_poly(bbx), crs = 4326))

	ll = unname(c(mean(bbx[c(1,3)]), mean(bbx[c(2,4)])))
	zoom = if (is.na(o$zoom)) findZoom(bbx) else o$zoom


	e = if (mode == "mapbox") {
		.TMAP_MAPBOX
	} else {
		.TMAP_MAPLIBRE
	}

	# set projection
	crs_o = if (!is.na(o$crs)) o$crs else "auto"

	crs_str = if (inherits(crs_o, "crs")) {
		sf::st_crs(crs_o)$input
	} else {
		crs_o
	}

	crs = "globe"
	for (i in 1L:length(e$crs_options)) {
		if (length(grep(names(e$crs_options[i]), crs_str, fixed = TRUE)) > 0) {
			crs = unname(e$crs_options[i])
			break
		}
	}

	style = resolve_style(e$style, api_key = e$api_key, default = o$basemap.server[1], bg = o$bg.color)

	if (mode == "mapbox") {
		# quick & dirty
		if (zoom < 3) {
			# ignore center for global view (otherwise it will be (0, -3) due to Antarctica)
			m = mapgl::mapboxgl(center = c(0,0), zoom = zoom, pitch = o$pitch, bearing = o$bearing, style = style) |>
				mapgl::add_navigation_control(visualize_pitch = TRUE) |>
				mapgl::set_projection(crs)
		} else {
			m = mapgl::mapboxgl(center = ll, zoom = zoom, pitch = o$pitch, bearing = o$bearing, style = style) |>
				mapgl::add_navigation_control(visualize_pitch = TRUE) |>
				mapgl::set_projection(crs)
		}
	} else {
		# quick & dirty
		if (zoom < 3) {
			# ignore center for global view (otherwise it will be (0, -3) due to Antarctica)
			m = mapgl::maplibre(center = c(0,0), zoom = zoom, pitch = o$pitch, bearing = o$bearing, style = style) |>
				mapgl::add_navigation_control(visualize_pitch = TRUE) |>
				mapgl::set_projection(crs) |>
				mapgl::add_globe_control()
		} else {
			m = mapgl::maplibre(center = ll, zoom = zoom, pitch = o$pitch, bearing = o$bearing, style = style) |>
				mapgl::add_navigation_control(visualize_pitch = TRUE) |>
				mapgl::set_projection(crs) |>
				mapgl::add_globe_control()
		}
	}
	assign_mapgl(m, facet_row, facet_col, facet_page, mode = mode)

	NULL
}



#' @export
#' @keywords internal
#' @name tmapMapboxOverlay
#' @rdname tmapMapbox
tmapMapboxOverlay = function(bbx, facet_row, facet_col, facet_page, o) {
	NULL
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreOverlay = function(bbx, facet_row, facet_col, facet_page, o) {
	NULL
}

# internal helpers -----------------------------------------------------------

.mapgl_output_fun <- function(gs) switch(gs,
										 Maplibre = mapgl::maplibreOutput,
										 Mapbox   = mapgl::mapboxglOutput)

.mapgl_proxy_fun <- function(gs) switch(gs,
										Maplibre = mapgl::maplibre_proxy,
										Mapbox   = mapgl::mapboxgl_proxy)

# renderTmap -----------------------------------------------------------------


.renderTmapGS_mapgl <- function(x, expr, env, ...) {
	expr <- bquote(utils::getFromNamespace("print.tmap", "tmap")(.(expr), in.shiny = TRUE))
	htmlwidgets::shinyRenderWidget(expr, .mapgl_output_fun(x$gs), env, quoted = TRUE)
}

#' @exportS3Method tmap::renderTmapGS
renderTmapGS.Maplibre <- .renderTmapGS_mapgl

#' @exportS3Method tmap::renderTmapGS
renderTmapGS.Mapbox <- .renderTmapGS_mapgl

# tmapOutput -----------------------------------------------------------------

.tmapOutputGS_mapgl <- function(x, outputId, width, height, ...) {
	.mapgl_output_fun(x$gs)(outputId = outputId, width = width, height = height)
}

#' @exportS3Method tmap::tmapOutputGS
tmapOutputGS.Maplibre <- .tmapOutputGS_mapgl

#' @exportS3Method tmap::tmapOutputGS
tmapOutputGS.Mapbox <- .tmapOutputGS_mapgl

# tmapProxy ------------------------------------------------------------------

.tmapProxyGS_mapgl <- function(x, mapId, session, tmobj, ...) {
	print(tmobj,
			   m = .mapgl_proxy_fun(x$gs)(mapId, session),  # maplibre_proxy / mapboxgl_proxy
			   show = FALSE, in.shiny = TRUE, proxy = TRUE)
}


#' @exportS3Method tmap::tmapProxyGS
tmapProxyGS.Maplibre <- .tmapProxyGS_mapgl

#' @exportS3Method tmap::tmapProxyGS
tmapProxyGS.Mapbox <- .tmapProxyGS_mapgl


mapgl_layer_ids = function(e, zindex) {
	reg = e$layer_zindex
	if (is.null(reg) || !length(reg)) return(character(0))
	unlist(reg[as.character(zindex)], use.names = FALSE)
}

#' @param show show
#' @param knit knit
#' @param knit_opts knit options
#' @param args args
#' @export
#' @keywords internal
#' @name tmapMapboxRun
#' @importFrom grDevices col2rgb rgb
#' @importFrom colorspace deutan protan tritan
#' @importFrom htmltools tags
#' @importFrom htmlwidgets prependContent
#' @import sf
#' @importFrom tmaptools bb_poly
#' @importFrom terra rast is.lonlat project ext crop
#' @importFrom stats na.omit
#' @importFrom utils head tail getFromNamespace
#' @import stars
#' @import cli
#' @importFrom units drop_units set_units
#' @rdname tmapMapbox
tmapMapboxRun = function(o, q, show, knit, knit_opts, args) {
	mapgl_run(o, q, show, knit, args, mode = "mapbox")
}


#' @export
#' @rdname tmapMapbox
tmapMaplibreRun = function(o, q, show, knit, knit_opts, args) {
	mapgl_run(o, q, show, knit, args, mode = "maplibre")

}

# Per-widget CSS injected into mapgl maps. Always hides the interactive-legend
# "Reset Filter" button (mapgl injects it via JS and toggles display:none/block,
# which otherwise shifts/grows the bottom-anchored legend). When a frame colour
# is given, it also draws a matching border on the control/globe button groups
# and on popups, so they read consistently with the legend frame. `frame_hex`
# must be a CSS-valid colour (hex); pass NULL to skip the borders.
mapgl_extra_css = function(frame_hex = NULL) {
	css = ".mapboxgl-legend .legend-reset-btn { display: none !important; }"
	if (!is.null(frame_hex)) {
		css = paste0(
			css, "\n",
			# Give control/globe buttons the same flat frame as the legend: add the
			# gray border and drop maplibre's default 2px "halo" (box-shadow), so
			# the two read consistently. (To instead keep the halo and add it to the
			# legend, drop the box-shadow:none here and shadow .mapboxgl-legend.)
			".maplibregl-ctrl-group, .mapboxgl-ctrl-group { border: 1px solid ", frame_hex, " !important; box-shadow: none !important; }\n",
			".maplibregl-popup-content, .mapboxgl-popup-content { border: 1px solid ", frame_hex, " !important; }"
		)
	}
	htmltools::HTML(css)
}

mapgl_prepend_css = function(w, css) {
	if (inherits(w, "htmlwidget")) {
		htmlwidgets::prependContent(w, htmltools::tags$style(css))
	} else {
		w
	}
}

# Convert an R colour to CSS-safe #RRGGBB (R's "gray40" etc. are not valid CSS);
# NULL/NA -> NULL so callers can skip.
mapgl_col_to_hex = function(col) {
	if (is.null(col) || (length(col) == 1L && is.na(col))) return(NULL)
	tryCatch({
		m = grDevices::col2rgb(col)
		grDevices::rgb(m[1, ], m[2, ], m[3, ], maxColorValue = 255)
	}, error = function(e) NULL)
}

mapgl_run = function(o, q, show, knit, args, mode) {
	e = if (mode == "mapbox") {
		.TMAP_MAPBOX
	} else {
		.TMAP_MAPLIBRE
	}

	if (show && o$show_gif_ani) {
		cli::cli_alert("{.field {mode} mode} Animations are not implemented in {mode} mode, so they are shown as facets")
	}


	ms = get("ms", envir = e)
	grps = get("grps", envir = e)

	# Prepend per-widget CSS: hide the reset button, and (from the legend frame
	# colour) border the control/globe buttons and popups to match.
	extra_css = mapgl_extra_css(mapgl_col_to_hex(o[["component.frame.color"]]))
	ms = lapply(ms, function(msi) lapply(msi, mapgl_prepend_css, css = extra_css))

	ctrl = split(q$group.control, f = q$group)
	ctrl = sapply(ctrl, tail, 1)

	no_grp = names(ctrl)[ctrl == "none"]

	if (length(no_grp)) {
		grps[no_grp] = NULL
	}


	ms2 = lapply(ms, function(msi) {
		x = if (o$nrows == 1 && o$ncols == 1) {
			if (length(grps)) {
				msi[[1]] |>
					mapgl::add_layers_control(layers = grps, collapsible = o$control.collapse, position = mapgl_pos(o$control.position))
			} else {
				msi[[1]]
			}
		} else {
			if (length(msi) > 2) cli::cli_warn("more than 2 facets not supported for the mode {.str mapbox}")
			orientation = ifelse(o$ncols >= o$nrows, "vertical", "horizontal")

			fc = o$free.coords
			sync = if (identical(o$sync, FALSE)) {
				"none"
			} else if (identical(o$sync, TRUE) || all(!fc)) {
				"sync"
			} else if (all(fc)) {
				"none"
			}
			if (o$swipe) {
				mode = "swipe"
			} else if (all(!fc)) {
				mode = sync
			} else {
				mode = "none"
			}

			if (mode == "none") {
				map_layout(msi[[1]], msi[[2]], orientation = orientation)
			} else {
				mapgl::compare(msi[[1]], msi[[2]], mode = mode, orientation = orientation)
			}
		}
		x
	})

	if (length(ms2) == 1) ms2 = ms2[[1]]
	if (show && !knit && !isTRUE(.TMAP$in.shiny)) {
		print(ms2)
	}
	ms2
}

map_layout = function(..., orientation = c("vertical", "horizontal")) {
	orientation = match.arg(orientation)

	flex_direction = if (orientation == "horizontal") "column" else "row"

	flex_style = sprintf(
		"display: flex; flex-direction: %s; gap: 10px; width: 100vw; height: 100vh; overflow: hidden; box-sizing: border-box;",
		flex_direction
	)

	maps = list(...)
	item_style = "flex: 1; min-width: 0; min-height: 0; overflow: hidden;"

	htmltools::browsable(
		htmltools::tagList(
			htmltools::tags$style("body { margin: 0; padding: 0; overflow: hidden; }"),
			htmltools::div(
				style = flex_style,
				lapply(maps, function(m) htmltools::div(style = item_style, m))
			)
		)
	)
}


# ============================================================
#  tmap_arrange support
# ============================================================
# Called from tmap's print_tmap_arrange() via do.call(paste0("tmap", gs,
# "Arrange"), ...) where gs is "Maplibre" / "Mapbox". Each tmap object is
# rendered to its own mapgl widget and the widgets are placed in an
# nrow x ncol CSS grid.

#' @param tms list of tmap objects
#' @param nx number of facets
#' @param nrow number of grid rows
#' @param opts arrange options (widths, heights, height, sync, ...)
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxArrange = function(tms, nx, ncol, nrow, opts, knit, show, args, options) {
	mapgl_arrange(tms, nx, ncol, nrow, opts, knit, show, args, options, mode = "mapbox")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreArrange = function(tms, nx, ncol, nrow, opts, knit, show, args, options) {
	mapgl_arrange(tms, nx, ncol, nrow, opts, knit, show, args, options, mode = "maplibre")
}

mapgl_arrange = function(tms, nx, ncol, nrow, opts, knit, show, args, options, mode) {
	if (isTRUE(opts$sync)) {
		cli::cli_inform(c(
			"!" = "{.arg sync} is not supported for arranged maps in {.str {mode}} mode; the maps pan/zoom independently.",
			"i" = "Synchronised navigation is available for two facets within a single map (via {.fn tm_facets})."
		))
	}

	# Render each tmap object to its own mapgl widget. show = FALSE so nothing is
	# printed here; print.tmap() returns the widget invisibly. Force each widget
	# to fill its grid cell - otherwise it renders at the htmlwidget default
	# height and the grid's height has no visible effect.
	widgets = lapply(tms, function(tm) {
		w = print(tm, show = FALSE, knit = FALSE)
		if (inherits(w, "htmlwidget")) {
			w$width  = "100%"
			w$height = "100%"
		}
		w
	})

	out = mapgl_grid_layout(widgets, ncol = ncol, nrow = nrow,
							widths = opts$widths, heights = opts$heights,
							height = opts$height)

	if (show && !knit && !isTRUE(.TMAP$in.shiny)) {
		print(out)
	}
	out
}

# Place a list of htmlwidgets in an nrow x ncol CSS grid (row-major, matching
# tmap's facet order). `widths`/`heights` are optional proportions (each summing
# to 1, length ncol / nrow); when absent the tracks are equal-sized.
mapgl_grid_layout = function(widgets, ncol, nrow, widths = NA, heights = NA, height = NULL) {
	tracks = function(n, sizes) {
		if (length(sizes) == n && !anyNA(sizes)) {
			paste(sprintf("%ffr", sizes), collapse = " ")
		} else {
			paste(rep("1fr", n), collapse = " ")
		}
	}

	# Overall height of the arranged grid. NULL -> fill the viewport (100vh); a
	# bare number is interpreted as pixels; a string is used as-is (e.g. "80vh").
	h_css = if (is.null(height) || (length(height) == 1L && is.na(height))) {
		"100vh"
	} else if (is.numeric(height)) {
		paste0(height, "px")
	} else {
		as.character(height)
	}

	grid_style = sprintf(
		paste0("display: grid; grid-template-columns: %s; grid-template-rows: %s; ",
			   "gap: 10px; width: 100%%; height: %s; overflow: hidden; box-sizing: border-box;"),
		tracks(ncol, widths), tracks(nrow, heights), h_css
	)
	item_style = "width: 100%; height: 100%; min-width: 0; min-height: 0; overflow: hidden;"

	htmltools::browsable(
		htmltools::tagList(
			htmltools::tags$style("body { margin: 0; padding: 0; }"),
			htmltools::div(
				style = grid_style,
				lapply(widgets, function(w) htmltools::div(style = item_style, w))
			)
		)
	)
}

#' Export tmap to mapbox and maplibre
#'
#' * `tmap_mapbox()` returns a [`mapgl`][mapgl::mapboxgl()] object (`"mapbox" mode`)
#' * `tmap_maplibre()` a [`mapgl`][mapgl::maplibregl()] object (`"maplibre"` mode).
#'
#' @param x a tmap object.
#' @param show show the map?
#' @param ... passed on to [`tmap`][tmap::print.tmap()]
#' @return  a [`mapgl`][mapgl::mapboxgl()] object (`"mapbox" mode`) or a [`mapgl`][mapgl::maplibregl()] object (`"maplibre"` mode). In case small multiples are shown, a list is returned.
#' @export
#' @examples
#' library(tmap)
#' library(tmap.mapgl)
#' map = tm_shape(World) + tm_polygons()
#' tmap_maplibre(map, show = TRUE)
tmap_mapbox = function(x,
						show = FALSE,
						...) {
	current_mode = getOption("tmap.mode")
	on.exit({
		options(tmap.mode = current_mode)
	})
	options(tmap.mode = "mapbox")
	print(x, show = show, ...)
}

#' @rdname tmap_mapbox
#' @export
tmap_maplibre = function(x,
						 show = FALSE,
						 ...) {
	current_mode = getOption("tmap.mode")
	on.exit({
		options(tmap.mode = current_mode)
	})
	options(tmap.mode = "maplibre")
	print(x, show = show, ...)
}

#' Map layer: let it snow!
#'
#' Map layer that generates snow.
#'
#' Arguments and their default values have been taken from [mapbox::set_snow()] (package version 0.2.2)
#'
#' @param density A number between 0 and 1 controlling the snow particles density. Default is 0.85.
#' @param intensity A number between 0 and 1 controlling the snow particles movement speed. Default is 1.0.
#' @param color A string specifying the color of the snow particles. Default is "#ffffff".
#' @param opacity A number between 0 and 1 controlling the snow particles opacity. Default is 1.0.
#' @param center_thinning A number between 0 and 1 controlling the thinning factor of snow particles from center. Default is 0.4.
#' @param direction A numeric vector of length 2 defining the azimuth and polar angles of the snow direction. Default is c(0, 50).
#' @param flake_size A number between 0 and 5 controlling the snow flake particle size. Default is 0.71.
#' @param vignette A number between 0 and 1 controlling the snow vignette screen-space effect. Default is 0.3.
#' @param vignette_color A string specifying the snow vignette screen-space corners tint color. Default is "#ffffff".
#' @return tmap layer
#' @seealso [tm_rain()]
#' @export
tm_snow = function(density = 0.85, intensity = 1.0, color = "#ffffff",
				   opacity = 1.0, center_thinning = 0.4, direction = c(0, 50),
				   flake_size = 0.71, vignette = 0.3, vignette_color = "#ffffff") {
	tmap::tm_element_list(tmap::tm_element(
		args = list(density = density, intensity = intensity, color = color,
					opacity = opacity, center_thinning = center_thinning, direction = direction,
					flake_size = flake_size, vignette = vignette, vignette_color = vignette_color),
		mapping.fun = "tm_aux_snow",
		zindex = 0,
		group = "",
		group.control = "none",
		subclass = c("tm_snow", "tm_aux_layer")))
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAuxPrepare.tm_aux_snow = function(a, bs, id, o) {
	""
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAuxPlot.tm_aux_snow = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	get_mapgl(facet_row, facet_col, facet_page, mode = "mapbox") |>
		(\(map) do.call(mapgl::set_snow, c(list(map = map), unclass(a))))() |>
		assign_mapgl(facet_row, facet_col, facet_page, mode = "mapbox")
	NULL
}

#' Map layer: let it rain!
#'
#' Map layer that generates rain. Only available in "mapbox" mode.
#'
#' Arguments and their default values have been taken from [mapbox::set_rain()] (package version 0.2.2)
#'
#' @param density A number between 0 and 1 controlling the rain particles density. Default is 0.5.
#' @param intensity A number between 0 and 1 controlling the rain particles movement speed. Default is 1.
#' @param color A string specifying the color of the rain droplets. Default is "#a8adbc".
#' @param opacity A number between 0 and 1 controlling the rain particles opacity. Default is 0.7.
#' @param center_thinning A number between 0 and 1 controlling the thinning factor of rain particles from center. Default is 0.57.
#' @param direction A numeric vector of length 2 defining the azimuth and polar angles of the rain direction. Default is c(0, 80).
#' @param droplet_size A numeric vector of length 2 controlling the rain droplet size (x - normal to direction, y - along direction). Default is c(2.6, 18.2).
#' @param distortion_strength A number between 0 and 1 controlling the rain particles screen-space distortion strength. Default is 0.7.
#' @param vignette A number between 0 and 1 controlling the screen-space vignette rain tinting effect intensity. Default is 1.0.
#' @param vignette_color A string specifying the rain vignette screen-space corners tint color. Default is "#464646".
#' @return tmap layer
#' @seealso [tm_show()]
#' @export
tm_rain = function(density = 0.5, intensity = 1.0, color = "#a8adbc",
				   opacity = 0.7, center_thinning = 0.57, direction = c(0, 80),
				   droplet_size = c(2.6, 18.2), distortion_strength = 0.7,
				   vignette = 1.0, vignette_color = "#464646") {
	tmap::tm_element_list(tmap::tm_element(
		args = list(density = density, intensity = intensity, color = color,
					opacity = opacity, center_thinning = center_thinning, direction = direction,
					droplet_size = droplet_size, distortion_strength = distortion_strength,
					vignette = vignette, vignette_color = vignette_color),
		mapping.fun = "tm_aux_rain",
		zindex = 0,
		group = "",
		group.control = "none",
		subclass = c("tm_rain", "tm_aux_layer")))
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAuxPrepare.tm_aux_rain = function(a, bs, id, o) {
	""
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAuxPlot.tm_aux_rain = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	get_mapgl(facet_row, facet_col, facet_page, mode = "mapbox") |>
		(\(map) do.call(mapgl::set_rain, c(list(map = map), unclass(a))))() |>
		assign_mapgl(facet_row, facet_col, facet_page, mode = "mapbox")
	NULL
}

mapgl_pos = function(pos) {
	if (is.character(pos)) pos = tmap::tm_pos_in(pos[1], pos[2])
	if (pos$type %in% c("out", "autoout")) {
		sel = c("cell.v", "cell.h")
	} else {
		sel = c("pos.v", "pos.h")
	}
	x = tolower(unlist(pos[sel]))

	if (x[1] %in% c("center", "centre")) x[1] = "top"
	if (x[2] %in% c("center", "centre")) x[2] = "left"

  paste(x, collapse = "-")
}

cont_split = function(x) strsplit(x, split = "_", fixed=TRUE)

gp_to_lpar = function(gp, mfun, shape = 20, pick_middle = TRUE) {
	# create a list of gp elements

	lst = c(list(fillColor = {if (!all(is.na(gp$fill))) gp$fill else "#000000"},
				 color = {if (!all(is.na(gp$col))) gp$col else "#000000"},
				 fillOpacity = {if (!all(is.na(gp$fill_alpha))) gp$fill_alpha else 0},
				 opacity = {if (!all(is.na(gp$col_alpha))) gp$col_alpha else 0},
				 'stroke-width' = {if (!all(is.na(gp$lwd))) gp$lwd else 0},
				 'stroke-dasharray' = {if (!all(is.na(gp$lty))) lty2dash(gp$lty) else "none"},
				 size = {if (!all(is.na(gp$size))) gp$size else 1},
				 shape = {if (!all(is.na(gp$shape))) gp$shape else shape}))

	lst_isnum = c(fillColor = FALSE,
				  color = FALSE,
				  fillOpacity = TRUE,
				  opacity = TRUE,
				  'stroke-width' = TRUE,
				  'stroke-dash' = FALSE,
				  size = TRUE,
				  shape = TRUE)

	lst = mapply(function(lsti, isnum) {
		if (!is.character(lsti)) return(lsti)

		if (nchar(lsti[1]) > 50) {
			x = cont_split(lsti)
			x = lapply(x, function(i) {
				i[i=="NA"] = NA
				i
			})
			if (isnum) x = lapply(x, as.numeric)
			if (pick_middle) {
				x = sapply(x, function(i) {
					if (all(is.na(i))) NA else {
						sq = c(5,6,4,7,3,8,2,9,1,10) # priority for middle values
						i[sq[which(!is.na(i)[sq])[1]]]
					}
				})
			}
			return(x)

		} else {
			return(lsti)
		}
	}, lst, lst_isnum[names(lst)], SIMPLIFY = FALSE)

	pch2shp = c("rect", "circle", "triangle", "plus", "cross", "diamond", "triangle",
				"cross", "star", "diamond", "circle", "polygon", "plus", "cross",
				"triangle", "rect", "circle", "triangle", "diamond", "circle",
				"circle", "circle", "rect", "diamond", "triangle", "polygon", "stadium") # shapes for pch 0:25 + 26 for stadium (NOTE: last one is a triangle upside-down. Since 21:25 are the defaults, and a polygon is chosen to differentiate from the other triangle)
	lst$shape = get_pch_names(lst$shape)

	if ("tm_data_lines" %in% mfun) lst$shape = "line"

	lst$width = lst$size * 20
	lst$height = lst$size * 20
	#lst$width[]
	lst$size = NULL
	lst
}



make_equal_list = function(x) {
	cls = class(x)
	n = max(vapply(x, length, integer(1)))
	structure(lapply(x, rep, length.out = n), class = cls)
}




# Map tmap's legend styling onto a mapgl::legend_style() object, to be passed as
# `style=` to the add_*_legend functions.
#
# Background and frame/border are always taken from `grp` (tmap core supplies the
# resolved defaults), so the legend honours tmap's styling regardless of whether
# the user named those arguments explicitly. The two exceptions, kept from the
# explicit-call handling, are bg = FALSE -> fully transparent background, and
# frame = FALSE -> no border. Title/text typography (colour, font family, weight)
# is mapped whenever a concrete value is present. Text/title SIZE is deliberately
# NOT mapped: tmap sizes are relative line-heights whereas legend_style() expects
# pixels, so there is no faithful 1:1 conversion; mapgl's own sizing is kept.
# points -> CSS pixels bridge for the legend frame corner radius. In plot mode
# frame.r is interpreted in points and multiplied by the component scale
# (rndrectGrob -> grid::unit(frame.r * sc, "pt"), sc = min(1/clipT) * o$scale);
# in mapgl border_radius is CSS pixels. The points->pixels part is 96/72; the
# remaining ~2.25 is an empirical density match between the static device and the
# on-screen CSS legend (mapgl can't reproduce tmap's legend auto-scaling). Folded
# together this is ~3. frame.lwd needs no unit factor: grid's lwd unit (1/96")
# already equals one CSS pixel, which is why borders match at x1.
mapgl_frame_r_px = (96 / 72) * 2.25   # ~= 3

# CSS-pixel height of one tmap "line-height" in mapgl mode, used to convert the
# line-height-based stack_margin into the pixel gap between stacked legends. It's
# a different unit bridge from mapgl_frame_r_px (which converts *points*), so the
# two constants are intentionally separate. o$scale is applied at the call site.
mapgl_line_px = 28

mapgl_legend_style = function(grp, grp_called, cmp, o) {
	# NA / NULL / "" -> NULL, so the arg is omitted and mapgl's default applies.
	nn = function(x) if (is.null(x) || (length(x) == 1L && (is.na(x) || identical(x, "")))) NULL else x

	# mapgl formats border width/radius with sprintf("%d"), so they must be whole
	# integers (a fractional frame.lwd like 1.5 would otherwise error).
	as_int = function(x) { x = nn(x); if (is.null(x)) NULL else as.integer(round(x)) }

	# Convert an R colour to #RRGGBB for CSS. CSS understands hex and standard
	# named colours (e.g. "pink"), but NOT R's numbered greys ("gray50"/"grey50")
	# or other R-only names, which would silently void the whole CSS declaration
	# (border/background) and make the frame disappear. Unknown strings are left
	# as-is.
	to_hex = function(col) {
		col = nn(col)
		if (is.null(col)) return(NULL)
		tryCatch({
			m = grDevices::col2rgb(col)
			grDevices::rgb(m[1, ], m[2, ], m[3, ], maxColorValue = 255)
		}, error = function(e) col)
	}

	face_to_weight = function(ff) {
		ff = nn(ff)
		if (is.null(ff)) NULL else if (grepl("bold", ff, fixed = TRUE)) "bold" else "normal"
	}

	# Carry tmap's global scale through, mirroring plot mode's `* sc`. The
	# legend-autoscale part (min(1/clipT)) has no mapgl equivalent, so only o$scale
	# is portable; at the default scale = 1 this is a no-op.
	sc = if (!is.null(o$scale)) o$scale else 1

	# --- background ---
	background_color = NULL
	background_opacity = NULL
	if (("bg" %in% grp_called) && isFALSE(grp$bg)) {
		background_opacity = 0
	} else {
		background_color = to_hex(grp$bg.color)
		background_opacity = nn(grp$bg.alpha)
	}

	# --- frame / border ---
	border_color = NULL
	border_width = NULL
	border_radius = NULL
	if (("frame" %in% grp_called) && isFALSE(grp$frame)) {
		border_width = 0
	} else {
		border_color = to_hex(if (is.character(grp$frame)) grp$frame else grp$frame.color)
		border_width = as_int({ w = nn(grp$frame.lwd); if (is.null(w)) NULL else w * sc })
		border_radius = as_int({ r = nn(grp$frame.r); if (is.null(r)) NULL else r * sc * mapgl_frame_r_px })
	}

	# --- patch (element) border: the symbol / polygon outline ---
	element_border_color = NULL
	element_border_width = NULL
	has_outline = !is.null(cmp$gp$col) && !all(is.na(cmp$gp$col)) &&
		!is.null(cmp$gp$lwd) && any(cmp$gp$lwd > 0, na.rm = TRUE)
	if (has_outline) {
		element_border_color = to_hex(cmp$gp$col[which(!is.na(cmp$gp$col))[1]])
		element_border_width = as_int(max(cmp$gp$lwd, na.rm = TRUE) * sc)
		if (!is.null(element_border_width) && element_border_width < 1L) element_border_width = 1L
	}

	mapgl::legend_style(
		background_color   = background_color,
		background_opacity = background_opacity,
		border_color       = border_color,
		border_width       = border_width,
		border_radius      = border_radius,
		text_color         = nn(cmp$text.color),
		title_color        = nn(cmp$title.color),
		font_family        = nn(cmp$text.fontfamily),
		title_font_family  = nn(cmp$title.fontfamily),
		font_weight        = face_to_weight(cmp$text.fontface),
		title_font_weight  = face_to_weight(cmp$title.fontface),
		element_border_color = element_border_color,
		element_border_width = element_border_width
	)
}

# Resolve the interactive target layer(s) + filter column for a categorical
# legend from the registry written by the layer renderers (.TMAP$mapgl_cat).
# `is_col` marks a border/line-colour legend (vs a fill legend); the caller
# detects it via colVary.
#
# Routing rules (mapgl allows only one interactive legend to own a layer):
#  - both aesthetics categorical -> fill legend owns the fill layer, col legend
#    owns the border layer (each filters its own visual channel; no clash).
#  - only one aesthetic categorical -> that legend owns BOTH layers, so toggling
#    a class hides the whole feature (fill + outline).
#  - no separate outline layer (tm_symbols' single circle layer, or lines) ->
#    fall back to the one layer that exists.
# Returns NULL when there is nothing categorical to filter (caller then keeps a
# legacy single-layer fallback).
mapgl_legend_target = function(glid, is_col) {
	reg = .TMAP$mapgl_cat[[glid]]
	if (is.null(reg)) return(NULL)
	both_cat = !is.null(reg$fill) && !is.null(reg$col)
	fl = reg$fill_layer
	bl = reg$border_layer

	if (is_col) {
		info   = reg$col
		layers = if (is.na(bl)) fl else if (both_cat) bl else c(fl, bl)
	} else {
		info   = reg$fill
		layers = if (both_cat || is.na(bl)) fl else c(fl, bl)
	}
	layers = layers[!is.na(layers)]
	if (!length(layers) || is.null(info)) return(NULL)
	list(layer_id = layers, filter_column = info$column)
}

mapgl_legend = function(cmp, m, o, orientation, mode) {

	legpos = mapgl_pos(cmp$position)

	m2 = if (cmp$type == "none") {
		#message("Text based legends not supported in view mode")
		m
	} else if (cmp$type == "gradient") {

		colVary = length(cmp$gp2$color) > 1L
		if (colVary) cmp$gp2$fillColor = cmp$gp2$color

		# remove na
		if (cmp$na.show) {
			labs = head(cmp$labels, -1)
			cols = head(cmp$gp2$fillColor, -1)
		} else {
			labs = cmp$labels
			cols = cmp$gp2$fillColor
		}
		colsNA = is.na(cols)
		if (any(colsNA)) {
			labs = labs[!colsNA]
			cols = cols[!colsNA]
		}

		m |> mapgl::add_continuous_legend(legend_title = cmp$title, values = labs, colors = cols, add = TRUE, position = legpos,
										  margin_top = cmp$margin_top,
										  margin_bottom = cmp$margin_bottom,
										  margin_left = cmp$margin_left,
										  margin_right = cmp$margin_right, draggable = TRUE,
										  style = cmp$mapgl_style)
	} else if (cmp$type == "lines") {
		cat_col    = "__tmap_cat__"
		colVary    = length(cmp$gp2$color) > 1L
		gp2        = make_equal_list(cmp$gp2)
		if (colVary) gp2$fillColor = gp2$color
		circular_patches = !any(is.na(cmp$gp$shape)) && all(cmp$gp$shape %in% c(1, 10, 16, 19:21))

		# A lines legend always represents the line-colour (col) aesthetic.
		tgt = mapgl_legend_target(cmp$glid, is_col = TRUE)
		if (is.null(tgt)) tgt = list(layer_id = paste0(cmp$glid, "lines"), filter_column = NULL)

		m |> mapgl::add_categorical_legend(
			colors           = gp2$fillColor,
			values           = cmp$labels,
			position         = legpos,
			legend_title     = cmp$title,
			circular_patches = circular_patches,
			add              = TRUE,
			patch_shape      = "line",
			sizes            = cmp$gp$lwd,
			interactive      = TRUE,
			margin_top       = cmp$margin_top,
			margin_bottom    = cmp$margin_bottom,
			margin_left      = cmp$margin_left,
			margin_right     = cmp$margin_right,
			layer_id         = tgt$layer_id,
			draggable        = TRUE,
			filter_column    = tgt$filter_column,
			style            = cmp$mapgl_style
		)
	} else { # "symbols"
		colVary = length(cmp$gp2$color) > 1L   # TRUE => this legend is the border/col aesthetic
		gp2     = make_equal_list(cmp$gp2)
		if (colVary) gp2$fillColor = gp2$color

		patches = if (!any(is.na(cmp$gp$shape)) && all(cmp$gp$shape %in% c(1, 10, 16, 19:21))) {
			"circle"
		} else {
			"square"
		}
		sizes = if (!is.na(cmp$gp$size[1])) cmp$gp$size * 20 else NULL

		# Pick the target layer(s) + per-aesthetic filter column from the registry
		# the layer renderer wrote. Fall back to the legacy single fill-layer
		# target (auto-detected column) if no registry entry exists.
		tgt = mapgl_legend_target(cmp$glid, is_col = colVary)
		if (is.null(tgt)) {
			geom = if ("polygons" %in% cmp$layer) "polygons" else if ("symbols" %in% cmp$layer) "symbols" else cmp$layer[1]
			tgt = list(layer_id = paste0(cmp$glid, geom, "_fill"), filter_column = NULL)
		}

		m |> mapgl::add_categorical_legend(
			colors         = gp2$fillColor,
			values         = cmp$labels,
			position       = legpos,
			legend_title   = cmp$title,
			patch_shape    = patches,
			add            = TRUE,
			interactive    = TRUE,
			sizes          = sizes,
			margin_top     = cmp$margin_top,
			margin_bottom  = cmp$margin_bottom,
			margin_left    = cmp$margin_left,
			margin_right   = cmp$margin_right,
			layer_id       = tgt$layer_id, draggable = TRUE,
			filter_column  = tgt$filter_column,
			style          = cmp$mapgl_style
		)
	}
	m2

}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPlot.tm_legend_portrait = function(comp, m, o) {
	mapgl_legend(comp, m, o, orientation = "vertical", mode = "mapbox")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPlot.tm_legend_landscape = function(comp, m, o) {
	mapgl_legend(comp, m, o, orientation = "horizontal", mode = "mapbox")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPlot.tm_legend_portrait = function(comp, m, o) {
	mapgl_legend(comp, m, o, orientation = "vertical", mode = "mapbox")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPlot.tm_legend_landscape = function(comp, m, o) {
	mapgl_legend(comp, m, o, orientation = "horizontal", mode = "mapbox")
}

#' @param facet_row,facet_col,facet_page row column and page id
#' @param class class
#' @param stack stack
#' @param stack_auto stack_auto
#' @param pos.h pos.h
#' @param pos.v pos.v
#' @param bbox bbox
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxComp = function(comp, o, facet_row = NULL, facet_col = NULL, facet_page, class, stack, stack_auto, pos.h, pos.v, bbox) {
	mapgl_comp(comp = comp,
			   o = o,
			   facet_row = facet_row,
			   facet_col = facet_col,
			   facet_page = facet_page,
			   class = class,
			   stack = stack,
			   stack_auto = stack_auto,
			   pos.h = pos.h,
			   pos.v = pos.v,
			   bbox = bbox,
			   mode = "mapbox")}

#' @export
#' @keywords internal
#' @name tmapMapboxLegend
#' @rdname tmapMapbox
tmapMaplibreComp = function(comp, o, facet_row = NULL, facet_col = NULL, facet_page, class, stack, stack_auto, pos.h, pos.v, bbox) {
	mapgl_comp(comp = comp,
			   o = o,
			   facet_row = facet_row,
			   facet_col = facet_col,
			   facet_page = facet_page,
			   class = class,
			   stack = stack,
			   stack_auto = stack_auto,
			   pos.h = pos.h,
			   pos.v = pos.v,
			   bbox = bbox,
			   mode = "maplibre")
}


mapgl_comp = function(comp, o, facet_row = NULL, facet_col = NULL, facet_page, class, stack, stack_auto, pos.h, pos.v, bbox, mode) {

		### from tmapGridComp, migrate to tmap generic
	    # get component group settings
		grp = comp[[1]][c("position",
						  "stack",
						  "frame_combine",
						  "equalize",
						  "resize_as_group",
						  "stack_margin",
						  "offset",
						  "frame" ,
						  "frame.color",
						  "frame.lwd",
						  "frame.r",
						  "bg",
						  "bg.color",
						  "bg.alpha")]

		any_legend_chart_inset = any(vapply(comp, inherits, FUN.VALUE = logical(1), c("tm_legend", "tm_chart", "tm_inset")))
		grp_called = setdiff(unique(do.call(c, lapply(comp, FUN = "[[", "called_via_comp_group"))), "group_id")

		if (!("frame" %in% grp_called)) grp$frame = any_legend_chart_inset
		if (!("bg" %in%grp_called)) grp$bg = any_legend_chart_inset
	    ###

	m = get_mapgl(facet_row, facet_col, facet_page, mode = mode)
	rc_text = frc(facet_row, facet_col)

	stack = stack[1]

	legpos = mapgl_pos(comp[[1]]$position) # should be identical over components

	# stack_margin defaults to a named vector c(combined, apart); frame_combine is
	# always FALSE in mapgl mode, so use the "apart" value (fall back to the first).
	sc = if (!is.null(o$scale)) o$scale else 1
	sm = grp$stack_margin
	sm = if (!is.null(names(sm)) && all(c("combined", "apart") %in% names(sm))) sm[["apart"]] else sm[1]
	stack_margin_px = if (is.null(sm) || length(sm) != 1L || is.na(sm)) 0 else sm * mapgl_line_px * sc

	# Each legend is positioned by an offset (os) from the anchor. When the anchor
	# is the far side (bottom for a vertical stack, right for a horizontal one) the
	# stack grows away from it, so the first component (lowest z) must sit at the
	# far end of the stack. Iterate in reverse for those anchors so the on-screen
	# z-order matches plot mode regardless of position.
	reverse = if (stack == "vertical") {
		legpos %in% c("bottom-left", "bottom-right")
	} else {
		legpos %in% c("top-right", "bottom-right")
	}

	os = 0
	cred = 35
	for (cmp in (if (reverse) rev(comp) else comp)) {
		if (is.null(cmp$height) || is.na(cmp$height)) cmp$height = 100
		if (is.null(cmp$width) || is.na(cmp$width)) cmp$width = 500

		if (stack == "vertical") {
			if (legpos %in% c("bottom-right", "bottom-left")) {
				cmp$margin_bottom = os + ifelse(legpos == "bottom-right", cred, 0)
				cmp$margin_top = 0
			} else {
				cmp$margin_top = os
				cmp$margin_bottom =  ifelse(legpos == "bottom-right", cred, 0)
			}
			cmp$margin_left = 0
			cmp$margin_right = 0
			os = os + cmp$height + stack_margin_px
		} else {
			if (legpos %in% c("top-left", "bottom-left"))  {
				cmp$margin_left = os
				cmp$margin_right = 0
			} else {
				cmp$margin_right = os
				cmp$margin_left = 0
			}
			cmp$margin_top = 0
			cmp$margin_bottom = ifelse(legpos == "bottom-right", cred, 0)
			os = os + cmp$width + stack_margin_px
		}


		cmp$mapgl_style = mapgl_legend_style(grp, grp_called, cmp, o)

		if (mode == "maplibre") {
			m = tmapMaplibreCompPlot(cmp, m, o)
		} else {
			m = tmapMapboxCompPlot(cmp, m, o)
		}
	}

	assign_mapgl(m, facet_row, facet_col, facet_page, mode = mode)
	NULL
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPrepare.tm_geocoder = function(comp, o) {
	mapgl_comp_prepare(comp, o)
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompHeight.tm_geocoder = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompWidth.tm_geocoder = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPlot.tm_geocoder = function(comp, m, o) {
	m |> mapgl::add_geocoder_control(position = comp$legpos, placeholder = comp$placeholder)
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPrepare.tm_geocoder = function(comp, o) {
	mapgl_comp_prepare(comp, o)
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompHeight.tm_geocoder = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompWidth.tm_geocoder = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPlot.tm_geocoder = function(comp, m, o) {
	m |> mapgl::add_geocoder_control(position = comp$legpos, placeholder = comp$placeholder)
}

#' Map component: geocoder search bar
#'
#' Map component that adds a geocoder search bar
#'
#' @param placeholder placeholder text. Default is `"Search"`.
#' @param stack stack with other map components, either `"vertical"` or `"horizontal"`.
#' @param position The position specification of the component: an object created with `tm_pos_in()` or `tm_pos_out()`. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is `"left"`, `"center"` or `"right"` (or upper case, meaning tighter to the map frame), the second `"top"`, `"center"` or `"bottom"`. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use `group_id` and specify `component` in [tm_comp_group()].
#' @param group_id Component group id name. All components (e.g. legends, titles, etc) with the same `group_id` will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in [tm_comp_group()] where its argument `id` should correspond to `group_id`.
#' @param z z index, e.g. the place of the component relative to the other componets
#' @seealso \href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
#' @return a [tmap::tmap-element]
#' @export
tm_geocoder = function (placeholder, stack, position, group_id, z) {
	args = lapply(as.list(rlang::call_match()[-1]), eval, envir = parent.frame())
	args$called = names(args)
	args$group_id = args$group_id %||% NA_character_
	args$z = args$z %||% NA_integer_
	tmap::tm_element_list(do.call(tmap::tm_element, c(args, list(subclass = c("tm_geocoder",
																  "tm_component")))))
}

frc = function(row, col) paste0(sprintf("%02d", row), "_", sprintf("%02d", col))

select_sf = function(shpTM, dt) {
	shp = shpTM$shp
	stid = shpTM$tmapID

	dtid = dt$tmapID__

	tid = intersect(stid, dtid)

	d = data.table::data.table(sord = seq_along(tid), ord = dt$ord__[match(tid, dtid)], tid = tid)
	if ("ord" %in% names(d)) {
		data.table::setkeyv(d, cols = c("ord", "sord"))
	} else {
		data.table::setkeyv(d, cols = "sord")
	}
	sid = match(d$tid, stid)

	shpSel = shp[sid] #st_cast(shp[match(tid, tmapID)], "MULTIPOLYGON")

	# assign prop_ vectors to data dt (to be used in plotting) e.g. prop_angle is determined in tmapTransCentroid when along.lines = TRUE
	prop_vars = names(shpTM)[substr(names(shpTM), 1, 5) == "prop_"]
	if (length(prop_vars)) {
		for (p in prop_vars) {
			pname = substr(p, 6, nchar(p))
			dt[[pname]] = shpTM[[p]][sid]
		}
	}

	dt = dt[match(d$tid, dtid), ]
	list(shp = shpSel, dt = dt)
}

impute_gp = function(gp, dt) {
	dtn = setdiff(names(dt), c("tmapID__", paste0("by", 1L:3L, "__")))

	cols = paste0("__", dtn)
	gp1 = sapply(gp, "[[", 1)
	gpids = which(gp1 %in% cols)
	#gp[gpids] = as.list(dt[, dtn, with = FALSE])

	for (i in gpids) gp[i] = as.list(dt[, dtn[match(gp1[i], cols)], with = FALSE])
	gp
}

rescale_gp = function(gp, scale, skip = character()) {
	if ("lwd" %in% names(gp) && (!"lwd" %in% skip)) gp$lwd = gp$lwd * scale
	if ("size" %in% names(gp) && (!"size" %in% skip)) gp$size = gp$size * sqrt(scale)
	if ("cex" %in% names(gp) && (!"cex" %in% skip)) gp$cex = gp$cex * sqrt(scale)
	gp
}

findZoom = function(b) {
	## calculate zoom level
	# borrowed from https://github.com/dkahle/ggmap/blob/master/R/calc_zoom.r
	lon_diff = b[3] - b[1]
	lat_diff = b[4] - b[2]

	zoomlon = ceiling(log2(360 * 2/lon_diff))
	zoomlat = ceiling(log2(180 * 2/lat_diff))
	zoom = as.integer(min(zoomlon, zoomlat))
}


process_color <- function(col, alpha=NA, sepia_intensity=0, saturation=1, color_vision_deficiency_sim="none") {
	#if (length(col)>100) browser()
	isFactor <- is.factor(col)

	if (isFactor) {
		x <- as.integer(col)
		col <- levels(col)
	}

	res <- t(col2rgb(col, alpha=TRUE))

	# set alpha values
	if (!is.na(alpha)) res[res[,4] != 0, 4] <- alpha * 255

	# convert to sepia
	if (sepia_intensity!=0) {
		conv_matrix <- matrix(c(.393, .769, .189,
								.349, .686, .168,
								.272, .534, .131), ncol=3, byrow=FALSE)
		res[,1:3] <-  (res[,1:3] %*% conv_matrix) * sepia_intensity + res[,1:3] * (1-sepia_intensity)
		res[res>255] <- 255
		res[res<0] <- 0
	}

	# convert to black&white
	if (saturation!=1) {
		res[,1:3] <- (res[,1:3] %*% matrix(c(.299, .587, .114), nrow=3, ncol=3))  * (1-saturation) + res[,1:3] * saturation
		res[res>255] <- 255
		res[res<0] <- 0
	}
	if (all(res[,4]==255)) res <- res[,-4, drop=FALSE]

	new_cols <- do.call("rgb", c(unname(as.data.frame(res)), list(maxColorValue=255)))

	rlang::check_installed("colorspace")
	# color blind sim
	sim_colors = switch(color_vision_deficiency_sim,
						deutan = colorspace::deutan,
						protan = colorspace::protan,
						tritan = colorspace::tritan,
						function(x) x)

	new_cols2 = sim_colors(new_cols)

	if (isFactor) {
		new_cols2[x]
	} else {
		new_cols2
	}
}


pchs = stats::setNames(c(seq(0L, 25L, 1L), seq(100L, 109L, 1L)),
					   c(c('open-rect', 'open-circle', 'open-triangle', 'simple-plus',
					   	'simple-cross', 'open-diamond', 'open-down-triangle', 'cross-rect',
					   	'simple-star', 'plus-diamond', 'plus-circle', 'hexagram', 'plus-rect',
					   	'cross-circle', 'triangle-rect', 'solid-rect', 'solid-circle-md',
					   	'solid-triangle', 'solid-diamond', 'solid-circle-bg', 'solid-circle-sm', 'circle',
					   	'rect', 'diamond', 'triangle', 'down-triangle'
					   ),
					   c('rect', 'circle', 'triangle', 'plus', 'cross', 'diamond', 'star', 'stadium', 'line', 'polygon')
					   ))


get_pch_names = function(x) {
	if (is.numeric(x)) {
		if (!(all(x %in% pchs | x > 999))) stop("Unknown symbol values", call. = FALSE)
		y = names(pchs)[match(x, pchs)]
		y[x > 999] = x[x>999]
		y
	} else {
		if (!all(x %in% names(pchs))) stop("Unknown symbol values", call. = FALSE)
		x
	}
}


get_option_class = function(o, class = NULL, spatial_class = TRUE) {
	is_spatial = !spatial_class || (any(names(o) %in% c("stars", "sf", "sfc", "raster", "terra", "sp", "dimensions")))
	if (!is.null(class) && is_spatial) { # && is.list(o)
		mtch = which(names(o) %in% class)
		if (!length(mtch)) mtch = which(names(o) == "")[1]
		o = o[[mtch]]
	}
	o
}

# from stars package
is_regular_grid = function (x) {
	has_raster(x) && !(has_rotate_or_shear(x) || is_rectilinear(x) ||
					   	is_curvilinear(x))
}

has_raster = function (x) {
	if (inherits(x, "stars"))
		x = stars::st_dimensions(x)
	!is.null(r <- attr(x, "raster")) && all(r$dimensions %in%
												names(x))
}

has_rotate_or_shear = function (x) {
	dimensions = stars::st_dimensions(x)
	if (has_raster(x)) {
		r = attr(dimensions, "raster")
		!anyNA(r$affine) && any(r$affine != 0)
	}
	else FALSE
}

is_curvilinear = function (x) {
	d = stars::st_dimensions(x)
	has_raster(x) && isTRUE(attr(d, "raster")$curvilinear)
}

is_rectilinear = function (x) {
	d = stars::st_dimensions(x)
	if (has_raster(x) && !is_curvilinear(x)) {
		xy = attr(d, "raster")$dimensions
		dimx = d[[xy[1]]]
		dimy = d[[xy[2]]]
		(is.na(dimx$delta) || is.na(dimy$delta)) && (!regular_intervals(dimx$values) ||
													 	!regular_intervals(dimy$values))
	}
	else FALSE
}

regular_intervals = function (x, epsilon = 1e-10) {
	if (length(x) <= 1) {
		FALSE
	} else {
		ud = if (is.atomic(x))
			unique(diff(x))
		else {
			if (identical(tail(x$end, -1), head(x$start, -1)))
				x$end - x$start
			else return(FALSE)
		}
		abs(diff(range(ud))/mean.default(ud)) < epsilon
	}
}

consider_global = function (x, th = 0.6)
{
	b = sf::st_bbox(x)
	# in case margins are applied
	if (sf::st_is_longlat(b)) {
		b["xmin"] = max(b["xmin"], -180)
		b["xmax"] = min(b["xmax"], 180)
		b["ymin"] = max(b["ymin"], -90)
		b["ymax"] = min(b["ymax"], 90)
	}
	if (b$xmin == b$xmax || b$ymin == b$ymax) return(FALSE)
	earth_surface = 5.1e+14
	area = as.numeric(sf::st_area(sf::st_as_sfc(b)))
	area > (earth_surface * 0.6)
}


number_text_lines = function (txt)
{
	if (is.character(txt)) {
		length(strsplit(txt, "\n")[[1]])
	}
	else 1
}



fancy_breaks <- function(vec, as.count = FALSE, interval.disjoint = FALSE, intervals=FALSE, interval.closure="left", fun=NULL, scientific=FALSE, big.num.abbr = c("mln" = 6, "bln" = 9), prefix = "", suffix = "", text.separator="to", text.less.than=c("less", "than"), text.less.than_as.prefix = TRUE, text.or.more=c("or", "more"), text.or.more_as.prefix = FALSE, text.align="left", text.to.columns=FALSE, digits=NA, html.escape = TRUE, ...) {
	args <- list(...)
	n <- length(vec)

	args$called = NULL

	digits_defined = !is.na(digits)

	if (!intervals) {
		as.count = FALSE
		interval.disjoint = FALSE
	}

	if (inherits(vec, c("POSIXct", "POSIXlt", "Date"))) {
		x = format(vec)
	} else if (!is.null(fun) && !interval.disjoint) {
		if (as.count) {
			steps <- (vec[-1] - vec[-n])
			vec <- c(vec, vec - 1L, vec + 1L) # needed for: {1, 2, ... 9}
			digits <- 0
		}
		x <- do.call(fun, list(vec))
		if (as.count) {
			x1 <- x[1:(n-1)]
			x2 <- x[(n+2):(2*n)]
			x1p1 <- x[(2*n+1):(3*n-1)]
		}
	} else if (all(is.infinite(vec))) {
		x <- as.character(vec)
	} else {
		# calculate magnitude, needed to determine digits and big number abbreviations
		vec_fin <- unique(vec[!is.infinite(vec)])
		frm <- gsub(" ", "", sprintf("%20.10f", abs(vec_fin)))
		mag <- max(nchar(frm)-11)

		if (as.count) {
			steps <- (vec[-1] - vec[-n])
			vec <- c(vec, vec - 1L, vec + 1L) # needed for: {1, 2, ... 9}
			digits <- 0
		} else {
			# get number of decimals (which is number of decimals in vec, which is reduced when mag is large)
			ndec <- max(10 - nchar(frm) + nchar(sub("0+$","",frm)))
			if (!digits_defined) {
				digits <- max(min(ndec, 4-mag), 0)

				# add sign to frm
				frm_sign <- unique(paste0(ifelse(vec_fin<0, "-", "+"), frm))

				# test if number of digits is sufficient for unique labels
				if (!scientific) {
					while (anyDuplicated(substr(frm_sign, 1, nchar(frm_sign)-10 + digits)) && (digits < 10)) {
						digits <- digits + 1
					}

					# if (interval.disjoint) {
					# 	## NEW
					# 	#update vec_fin
					# 	# add 'to' numbers and see if digits should be increased
					# 	vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					# 	while (anyDuplicated(na.omit(vec_fin2))) {
					# 		digits = digits + 1
					# 		vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					# 	}
					# 	vec = c(vec, vec - 10^-digits)
					# }

				}

				# } else {
				# 	if (interval.disjoint) {
				# 		vec = c(vec, vec - 10^-digits)
				# 	}
			}
		}


		if ((!is.null(fun) && interval.disjoint)) {
			x <- do.call(fun, list(vec))
		} else if (!scientific || as.count) {

			# check whether big number abbrevations should be used
			ext <- ""
			if (!is.na(big.num.abbr[1])) {
				big.num.abbr <- sort(big.num.abbr, decreasing = TRUE)
				for (i in 1:length(big.num.abbr)) {
					o <- unname(big.num.abbr[i])
					if ((mag > o && all(vec - floor(vec/(10^o))*(10^o) < 1))) { #mag>(o+2) || removed
						vec <- vec / (10^o)
						ext <- paste0(" ", names(big.num.abbr)[i])
						break
					}
				}
			}

			if (interval.disjoint) {

				if (!digits_defined) {
					# update digits
					vec_fin <- unique(vec[!is.infinite(vec)])
					vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					while (anyDuplicated(na.omit(vec_fin2))) {
						digits = digits + 1
						vec_fin2 = c(vec_fin, vec_fin[-1] - 10^-digits)
					}
				}
				# add to points
				vec = c(vec, head(vec, -1) - 10^-digits, tail(vec, 1))
			}

			# set default values
			if (!("big.mark" %in% names(args))) args$big.mark <- ","
			if (!("format" %in% names(args))) args$format <- "f"
			if (!("preserve.width" %in% names(args))) args$preserve.width <- "none"

			x = do.call("formatC", c(list(x=vec, digits=digits), args))
			if (as.count || interval.disjoint) {
				x[(n+2):(2*n)] = paste0(x[(n+2):(2*n)], ext)
			} else {
				x = paste0(x, ext)
			}
			x <- paste0(prefix, x, suffix)
		} else {
			if (!("format" %in% names(args))) args$format <- "g"
			x <- do.call("formatC", c(list(x=vec, digits=digits), args))
		}

		if (as.count) {
			x1 <- x[1:(n-1)]
			x2 <- x[(n+2):(2*n)]
			x1p1 <- x[(2*n+1):(3*n-1)]
		} else if (interval.disjoint) {
			x1 = x[1:(n-1)]
			x2 = x[(n+2):(2*n)]


		}
		# x <- formatC(vec, format = "f", digits = 0)
		# x1 <- x[-n]
		# x2 <- formatC(vec[-1] - 1L, format = "f", digits = 0)
		# xs <- (vec[-1] - vec[-n])
		# x1p1 <- formatC(vec[-n] + 1L, format = "f", digits = 0)
	}

	if (intervals) {
		if (scientific) {
			if (as.count) {
				# discrete
				lbls <- paste("{", x1, "}", sep = "")
				lbls[steps == 2] <- paste("{", x1[steps == 2], ", ", x2[steps == 2], "}", sep="")
				lbls[steps > 2] <- paste("{", x1[steps > 2], ", ", x1p1[steps > 2], ", ..., ", x2[steps > 2], "}", sep="")
			} else {
				# continuous
				if (interval.closure=="left") {
					lbls <- paste("[", x[-n], ", ", x[-1], ")", sep="")
					lbls[n-1] <- paste(substr(lbls[n-1], 1, nchar(lbls[n-1])-1), "]", sep="")
				} else {
					lbls <- paste("(", x[-n], ", ", x[-1], "]", sep="")
					lbls[1] <- paste("[", substr(lbls[1], 2, nchar(lbls[1])), sep="")
				}
			}


		} else {
			if (as.count) {
				lbls <- x1
				lbls[steps>1] <- paste(x1[steps>1], x2[steps>1], sep = paste0(" ", text.separator, " "))
				if (vec[n]==Inf) lbls[n-1] <- paste2(x1[n-1], paste(text.or.more, collapse = " "), flipped = text.or.more_as.prefix)
			} else if (interval.disjoint) {
				lbls <- paste(x1, x2, sep = paste0(" ", text.separator, " "))
				if (vec[1]==-Inf) lbls[1] <- paste2(x1[2], paste(text.less.than, collapse = " "), flipped = text.less.than_as.prefix)
				if (vec[n]==Inf) lbls[n-1] <- paste2(x1[n-1], paste(text.or.more, collapse = " "), flipped = text.or.more_as.prefix)
			} else {
				x[vec==-Inf] <- ""

				lbls <- paste(x[-n], x[-1], sep = paste0(" ", text.separator, " "))
				if (vec[1]==-Inf) lbls[1] <- paste2(x[2], paste(text.less.than, collapse = " "), flipped = text.less.than_as.prefix)
				if (vec[n]==Inf) lbls[n-1] <- paste2(x[n-1], paste(text.or.more, collapse = " "), flipped = text.or.more_as.prefix)
			}

			if (text.to.columns) {
				#xtra <- as.numeric(!is.na(text.align) && text.align=="right")


				nc1 <- nchar(paste(x[-n], " ", sep = "")) + 1
				nc2 <- rep(nchar(paste(text.separator, " ", sep = "")), n-1)

				lbls_breaks <- matrix(c(nc1, nc1+nc2), ncol=2)

				if (vec[1]==-Inf) {
					if (length(text.less.than)==1) {
						lbls_breaks[1,] <- rep(nchar(paste(text.less.than[1], " ", sep = "")) + 1, 2)
					} else {
						lbls_breaks[1,] <- cumsum(c(nchar(paste(text.less.than[1], " ", sep = "")) + 1, nchar(text.less.than[2])+1))
					}
				}
				if (vec[n]==Inf) {
					if (length(text.or.more)==1) {
						lbls_breaks[n-1,] <- rep(nchar(paste(x[n-1], " ", sep = "")) + 1, 2)
					} else {
						lbls_breaks[n-1,] <- cumsum(c(nchar(paste(x[n-1], " ", sep = "")) + 1, nchar(text.or.more[1])+1))
					}

				}
				attr(lbls, "brks") <- lbls_breaks
			}




		}
	}

	y <- if (intervals) lbls else x
	attr(y, "align") <- text.align
	y
}

paste2 = function(x, y, flipped) {
	if (flipped) {
		paste(y, x, sep = " ")
	} else {
		paste(x, y, sep = " ")
	}
}

get_facet_id = function(row, col, nrow, ncol) {
	col + (row - 1L) * ncol
}


get_mapgl = function(facet_row, facet_col, facet_page, mode) {
	e = if (mode == "mapbox") {
		.TMAP_MAPBOX
	} else {
		.TMAP_MAPLIBRE
	}

	ms = get("ms", envir = e)
	nrow = get("nrow", envir = e)
	ncol = get("ncol", envir = e)

	m = ms[[facet_page]]

	fr = max(1, facet_row) # facet_row can be -1 or -2
	fc = max(1, facet_col) # facet_row can be -1 or -2

	mbid = get_facet_id(fr, fc, nrow, ncol)

	m[[mbid]]
}

assign_mapgl = function(m, facet_row, facet_col, facet_page, mode) {
	e = if (mode == "mapbox") {
		.TMAP_MAPBOX
	} else {
		.TMAP_MAPLIBRE
	}

	ms = get("ms", envir = e)
	nrow = get("nrow", envir = e)
	ncol = get("ncol", envir = e)

	fr = max(1, facet_row) # facet_row can be -1 or -2
	fc = max(1, facet_col) # facet_row can be -1 or -2

	mid = get_facet_id(fr, fc, nrow, ncol)

	ms[[facet_page]][[mid]] = m
	assign("ms", ms, envir = e)
	NULL
}


mapgl_submit_group = function(group, layers, mode, pane = NULL) {
	e = if (mode == "mapbox") .TMAP_MAPBOX else .TMAP_MAPLIBRE

	grps = get("grps", envir = e)
	if (group %in% names(grps)) {
		grps[[group]] = unique(c(grps[[group]], layers))
	} else {
		grps[[group]] = layers
	}
	assign("grps", grps, envir = e)

	# register layer ids against their z-index (= pane) so tm_remove_layer(zindex)
	# can find and clear them in proxy mode
	if (!is.null(pane)) {
		lz = if (exists("layer_zindex", envir = e, inherits = FALSE)) {
			get("layer_zindex", envir = e)
		} else list()
		zkey = as.character(pane)
		lz[[zkey]] = unique(c(lz[[zkey]], layers))
		assign("layer_zindex", lz, envir = e)
	}
	NULL
}

#' @param id id
#' @param pane pane
#' @param group group
#' @param bs bs
#' @export
#' @keywords internal
#' @return internal tmap lists
#' @rdname tmapMapbox
tmapMapboxAuxPrepare = function(a, bs, id, o) {
	UseMethod("tmapMapboxAuxPrepare")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAuxPrepare.default = function(a, bs, id, o) {
	""
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAuxPlot = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	UseMethod("tmapMapboxAuxPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAuxPlot.default = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	NULL
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreAuxPrepare = function(a, bs, id, o) {
	UseMethod("tmapMaplibreAuxPrepare")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreAuxPrepare.default = function(a, bs, id, o) {
	""
}



#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreAuxPlot = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	UseMethod("tmapMaplibreAuxPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreAuxPlot.default = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	NULL
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAuxPrepare.tm_aux_basemap = function(a, bs, id, o) {
	mapgl_tiles_prep(a, bs, id, o, e = .TMAP_MAPBOX, mode = "mapbox")
}

#' @export
#' @rdname tmapMapbox
tmapMaplibreAuxPrepare.tm_aux_basemap = function(a, bs, id, o) {
	mapgl_tiles_prep(a, bs, id, o, e = .TMAP_MAPLIBRE, mode = "maplibre")
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAuxPrepare.tm_aux_tiles = function(a, bs, id, o) {
	mapgl_tiles_prep(a, bs, id, o, e = .TMAP_MAPBOX, mode = "mapbox")
}

#' @export
#' @rdname tmapMapbox
tmapMaplibreAuxPrepare.tm_aux_tiles = function(a, bs, id, o) {
	mapgl_tiles_prep(a, bs, id, o, e = .TMAP_MAPLIBRE, mode = "maplibre")
}




mapgl_tiles_prep = function(a, bs, id, o, e, mode) {
	serv = a$server
	# Allow raw style URLs to pass through; only validate named providers.
	is_url = is.character(serv) && grepl("^(https?|mapbox|maptiler)://", serv)
	if (!is_url && !(serv %in% tmap::tmap_providers(mode))) {
		eq = utils::getFromNamespace("basemap_equivalent", "tmap")(serv, mode)
		if (!is.na(eq)) {
			utils::getFromNamespace("message_basemaps_equivalent", "tmap")(serv, mode, eq)
			serv = eq
		} else {
			fallback = o$basemap.server[1]
			utils::getFromNamespace("message_basemaps_invalid_provider", "tmap")(serv, mode, fallback)
			serv = fallback
		}
	}
	e$style = serv
	# Optional per-basemap API key from tm_basemap(api = ...) (the same `api`
	# argument used for Stadia/Thunderforest in plot mode). NULL unless supplied,
	# in which case it is used for esri/maptiler instead of the env var.
	e$api_key = a$api
	serv
}


#' @export
#' @rdname tmapMapbox
tmapMapboxAuxPlot.tm_aux_basemap = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	NULL
}

#' @export
#' @rdname tmapMapbox
tmapMaplibreAuxPlot.tm_aux_basemap = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	NULL
}


#' @export
#' @rdname tmapMapbox
tmapMapboxAuxPlot.tm_aux_tiles = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	NULL
}

#' @export
#' @rdname tmapMapbox
tmapMaplibreAuxPlot.tm_aux_tiles = function(a, bi, bbx, facet_row, facet_col, facet_page, id, pane, group, o) {
	NULL
}




#' @param bi bi
#' @export
#' @keywords internal
#' @name tmapMapboxGridXLab
#' @rdname tmapMapbox
tmapMapboxGridXLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	NULL
}

#' @export
#' @keywords internal
#' @name tmapMapboxGridYLab
#' @rdname tmapMapbox
tmapMapboxGridYLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	NULL
}




#' @export
#' @name tmapMaplibreGridXLab
#' @rdname tmapMapbox
tmapMaplibreGridXLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	NULL
}

#' @export
#' @name tmapMaplibreGridYLab
#' @rdname tmapMapbox
tmapMaplibreGridYLab = function(bi, bbx, facet_row, facet_col, facet_page, o) {
	NULL
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	UseMethod("tmapMapboxDataPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.default = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NULL
}


#' @param shpTM, dt, pdt, popup.format, hdt, idt, gp args
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_polygons = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	mapgl_polygons(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ..., mode = "mapbox")
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_polygons_3d = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	mapgl_polygons_3d(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ..., mode = "mapbox")
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_lines = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	mapgl_lines(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ..., mode = "mapbox")
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_symbols = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	mapgl_symbols(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ..., mode = "mapbox")
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_circles = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	mapgl_circles(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ..., mode = "mapbox")
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_raster = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	mapgl_raster(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ..., mode = "mapbox")
}



#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_fill = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_borders = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_dots = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxDataPlot.tm_data_bubbles = function(a, shpTM, dt, pdt, popup.format, hdt, idt, gp, bbx, facet_row, facet_col, facet_page, id, pane, group, glid, o, ...) {
	NextMethod()
}

#' Mapbox mode options
#'
#' Mapbox mode options. These options are specific to the mapbox mode.
#'
#' @param pitch The pitch angle
#' @param bearing The bearing (rotation angle) of the map, in degrees
#' @param control.position The position of the layer control box
#' @param control.collapse Should the layer control box be collapsed?
#' @param zoom The zoom level of the map
#' @example examples/mapbox.R
#' @return a [tmap::tmap-element]
#' @export
tm_mapbox = function(pitch, bearing, control.position, control.collapse, zoom) {
	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called_from = "tm_mapbox"
	do.call(tmap::tm_options, args)
}

#' Maplibre mode options
#'
#' Maplibre mode options. These options are specific to the maplibre mode.
#'
#' @param pitch The pitch angle
#' @param bearing The bearing (rotation angle) of the map, in degrees
#' @param control.position The position of the layer control box
#' @param control.collapse Should the layer control box be collapsed?
#' @param zoom The zoom level of the map
#' @example examples/maplibre.R
#' @return a [tmap::tmap-element]
#' @export
tm_maplibre = function(pitch, bearing, control.position, control.collapse, zoom) {
	args = lapply(as.list(rlang::call_match(dots_expand = TRUE)[-1]), eval, envir = parent.frame())
	args$called_from = "tm_maplibre"
	do.call(tmap::tm_options, args)
}

#' @param return.asp return.asp
#' @param vp vp
#' @param dg base map widget (internal; \code{NULL} in proxy mode, where the
#'   proxy object \code{m} is used as the drawing canvas instead)
#' @param m the proxy object from `maplibre_proxy()`/`mapboxgl_proxy()`, supplied in proxy mode
#' @export
#' @keywords internal
#' @name tmapMapboxInit
#' @rdname tmapMapbox
tmapMapboxInit = function(o, return.asp = FALSE, vp, prx, dg = NULL, m = NULL, ...) {
	mapgl_init(o = o, return.asp = return.asp, vp = vp, prx = prx,
			   dg = dg, m = m, e = .TMAP_MAPBOX, ...)
}
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreInit = function(o, return.asp = FALSE, vp, prx, dg = NULL, m = NULL, ...) {
	mapgl_init(o = o, return.asp = return.asp, vp = vp, prx = prx,
			   dg = dg, m = m, e = .TMAP_MAPLIBRE, ...)
}
mapgl_init = function(o, return.asp, vp, prx, dg = NULL, m = NULL, e, ...) {
	if (return.asp) return(1)
	per_page = rep(o$ncols * o$nrows, o$npages)
	k = o$ncols * o$nrows * o$npages
	if (o$n < k) {
		per_page[o$npages] = per_page[o$npages] - (k - o$n)
	}
	proxy = isTRUE(.TMAP$proxy)
	if (proxy) {
		mode = if (identical(e, .TMAP_MAPBOX)) "mapbox" else "maplibre"

		# Legends + controls are rebuilt wholesale on every render, so clear the
		# previous ones first (otherwise they stack). clear-all = no id argument.
		m = m |>
			mapgl::clear_legend() |>
			mapgl::clear_controls()

		# clear_controls() also removes the navigation / globe controls. On a full
		# render those are created in mapgl_shape() (navigation then globe, leaving
		# the globe icon on top), but mapgl_shape() early-returns in proxy mode, so
		# re-add them here or they vanish after the first update.
		#
		# In the proxy re-add path the controls stack in the OPPOSITE visual order
		# to the initial widget render, so to keep the globe on top (as at start)
		# we add globe first, then navigation.
		if (mode == "maplibre") m = mapgl::add_globe_control(m)
		m = mapgl::add_navigation_control(m, visualize_pitch = TRUE)
	}

	# layer removal is targeted: only the zindexes named by tm_remove_layer()
	if (proxy && length(prx)) {
		pane_name = utils::getFromNamespace("pane_name", "tmap")
		zres   = vapply(prx, function(p) p$zindex, FUN.VALUE = numeric(1))
		rm_ids = mapgl_layer_ids(e, pane_name(zres))
		if (length(rm_ids)) {
			mapgl::clear_layer(m, rm_ids)
			e$layer_zindex[pane_name(zres)] = NULL
		}
	}

	# proxy mode: draw onto the live map via `m`; otherwise use base widget `dg`
	base = if (proxy) m else dg
	ms = lapply(per_page, function(p) lapply(seq_len(p), function(i) base))
	e$ms     = ms
	e$nrow   = o$nrows
	e$ncol   = o$ncols
	e$leg_id = 1

	# don't wipe the registries on a proxy update
	if (!proxy) {
		e$grps         = list()
		e$layer_zindex = list()
	}
	NULL
}
#' @param q q
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxAux = function(o, q) {
	NULL
}
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreAux = function(o, q) {
	NULL
}

#' @param label label
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxWrap = function(label, facet_row, facet_col, facet_page, o) {
	NULL
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxXtab = function(label, facet_row, facet_col, facet_page, o) {
	NULL
}


#' @param label label
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreWrap = function(label, facet_row, facet_col, facet_page, o) {
	NULL
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreXtab = function(label, facet_row, facet_col, facet_page, o) {
	NULL
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPrepare.tm_title = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompHeight.tm_title = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompWidth.tm_title = function(comp, o) {
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPlot.tm_title = function(comp, m, o) {
	m
}


############ scalebar


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPrepare.tm_scalebar = function(comp, o) {
	mapgl_comp_prepare(comp, o)
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompHeight.tm_scalebar = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompWidth.tm_scalebar = function(comp, o) {
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPlot.tm_scalebar = function(comp, m, o) {
	mapgl_scalebar_plot(comp, m, o)

}

############ credits



#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPrepare.tm_snow = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompHeight.tm_snow = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompWidth.tm_snow = function(comp, o) {
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPlot.tm_snow = function(comp, m, o) {
	m |> mapgl::set_snow()
}




#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPrepare.tm_title = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompHeight.tm_title = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompWidth.tm_title = function(comp, o) {
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPlot.tm_title = function(comp, m, o) {
	m
}




#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPrepare.tm_scalebar = function(comp, o) {
	mapgl_comp_prepare(comp, o)
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompHeight.tm_scalebar = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompWidth.tm_scalebar = function(comp, o) {
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPlot.tm_scalebar = function(comp, m, o) {
	mapgl_scalebar_plot(comp, m, o)
}


mapgl_scalebar_plot = function(comp, m, o) {
	m |> mapgl::add_scale_control(position = comp$legpos)
}





#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPrepare.tm_credits = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompHeight.tm_credits = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompWidth.tm_credits = function(comp, o) {
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPlot.tm_credits = function(comp, m, o) {
	m
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPrepare.tm_mouse_coordinates = function(comp, o) {
	comp$show = TRUE
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompHeight.tm_mouse_coordinates = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompWidth.tm_mouse_coordinates = function(comp, o) {
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPlot.tm_mouse_coordinates = function(comp, m, o) {
	m
}








########## minimap

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPrepare.tm_minimap = function(comp, o) {
	mapgl_comp_prepare(comp, o)
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPrepare.tm_minimap = function(comp, o) {
	mapgl_comp_prepare(comp, o)
}




#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompHeight.tm_minimap = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompWidth.tm_minimap = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompHeight.tm_minimap = function(comp, o) {
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompWidth.tm_minimap = function(comp, o) {
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPlot.tm_minimap = function(comp, m, o) {
	mapgl_minimap_plot(comp, m, o)
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPlot.tm_minimap = function(comp, m, o) {
	mapgl_minimap_plot(comp, m, o)
}


mapgl_comp_prepare = function(comp, o) {
	comp$show = TRUE
	comp$legpos = mapgl_pos(comp$position)
	comp
}

mapgl_minimap_plot = function(comp, m, o) {
	m |> mapgl::add_globe_minimap(position = comp$legpos)
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapValuesCheck_height = tmap::tmapValuesCheck_size
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapValuesIsDiv_height = tmap::tmapValuesIsDiv_size
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapValuesRange_height = tmap::tmapValuesRange_size
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapValuesVV_height = tmap::tmapValuesVV_size
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapValuesSubmit_height = tmap::tmapValuesSubmit_size
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapValuesScale_height = tmap::tmapValuesScale_size
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapValuesColorize_height = tmap::tmapValuesColorize_size
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapValuesCVV_height = tmap::tmapValuesCVV_size

#' Internal tmap methods
#'
#' Internal tmap methods
#'
#' @param comp the shape object
#' @param o the list of options
#' @param m mapbox oer maplibre object
#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPrepare = function(comp, o) {
	UseMethod("tmapMapboxCompPrepare")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompHeight = function(comp, o) {
	UseMethod("tmapMapboxCompHeight")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompWidth = function(comp, o) {
	UseMethod("tmapMapboxCompWidth")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPlot = function(comp, m, o) {
	UseMethod("tmapMapboxCompPlot")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPrepare.default = function(comp, o) {
	cls = class(comp)[1]
	id = paste("mapbox_mode", cls, sep = "_")
	cli::cli_inform("{.field [mapbox mode]} Map component {.fun {cls}} not supported in {.str mapbox} mode.",
					.frequency_id = id,
					.frequency = "once")
	comp$show = FALSE
	comp
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMapboxCompPrepare.tm_chart_none = function(comp, o) {
	comp$show = FALSE
	comp
}





#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPrepare = function(comp, o) {
	UseMethod("tmapMaplibreCompPrepare")
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompHeight = function(comp, o) {
	UseMethod("tmapMaplibreCompHeight")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompWidth = function(comp, o) {
	UseMethod("tmapMaplibreCompWidth")
}

#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPlot = function(comp, m, o) {
	UseMethod("tmapMaplibreCompPlot")
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPrepare.default = function(comp, o) {
	cls = class(comp)[1]
	id = paste("maplibre_mode", cls, sep = "_")
	cli::cli_inform("{.field [maplibre mode]} Map component {.fun {cls}} not supported in {.str maplibre} mode.",
					.frequency_id = id,
					.frequency = "once")
	comp$show = FALSE
	comp
}


#' @export
#' @keywords internal
#' @rdname tmapMapbox
tmapMaplibreCompPrepare.tm_chart_none = function(comp, o) {
	comp$show = FALSE
	comp
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_aux_layers.R
\name{tm_rain}
\alias{tm_rain}
\title{Map layer: let it rain!}
\usage{
tm_rain(
  density = 0.5,
  intensity = 1,
  color = "#a8adbc",
  opacity = 0.7,
  center_thinning = 0.57,
  direction = c(0, 80),
  droplet_size = c(2.6, 18.2),
  distortion_strength = 0.7,
  vignette = 1,
  vignette_color = "#464646"
)
}
\arguments{
\item{density}{A number between 0 and 1 controlling the rain particles density. Default is 0.5.}

\item{intensity}{A number between 0 and 1 controlling the rain particles movement speed. Default is 1.}

\item{color}{A string specifying the color of the rain droplets. Default is "#a8adbc".}

\item{opacity}{A number between 0 and 1 controlling the rain particles opacity. Default is 0.7.}

\item{center_thinning}{A number between 0 and 1 controlling the thinning factor of rain particles from center. Default is 0.57.}

\item{direction}{A numeric vector of length 2 defining the azimuth and polar angles of the rain direction. Default is c(0, 80).}

\item{droplet_size}{A numeric vector of length 2 controlling the rain droplet size (x - normal to direction, y - along direction). Default is c(2.6, 18.2).}

\item{distortion_strength}{A number between 0 and 1 controlling the rain particles screen-space distortion strength. Default is 0.7.}

\item{vignette}{A number between 0 and 1 controlling the screen-space vignette rain tinting effect intensity. Default is 1.0.}

\item{vignette_color}{A string specifying the rain vignette screen-space corners tint color. Default is "#464646".}
}
\value{
tmap layer
}
\description{
Map layer that generates rain. Only available in "mapbox" mode.
}
\details{
Arguments and their default values have been taken from [mapbox::set_rain()] (package version 0.2.2)
}
\seealso{
[tm_show()]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_variable_height.R, R/mapgl_aux.R,
%   R/mapgl_comp.R, R/mapgl_comp_charts.R, R/mapgl_draw.R, R/mapgl_geocoder.R,
%   R/mapgl_init.R, R/mapgl_legend.R, R/mapgl_providers.R, R/mapgl_run.R,
%   R/mapgl_shape.R, R/mapgl_wrap_xtab.R, R/tm_aux_layers.R, R/tm_fullscreen.R,
%   R/tmapMapboxDataPlot.R, R/tmapMaplibreDataPlot.R
\name{tmapValuesCheck_height}
\alias{tmapValuesCheck_height}
\alias{tmapValuesIsDiv_height}
\alias{tmapValuesRange_height}
\alias{tmapValuesVV_height}
\alias{tmapValuesSubmit_height}
\alias{tmapValuesScale_height}
\alias{tmapValuesColorize_height}
\alias{tmapValuesCVV_height}
\alias{tmapMapboxAuxPrepare}
\alias{tmapMapboxAuxPrepare.default}
\alias{tmapMapboxAuxPlot}
\alias{tmapMapboxAuxPlot.default}
\alias{tmapMaplibreAuxPrepare}
\alias{tmapMaplibreAuxPrepare.default}
\alias{tmapMaplibreAuxPlot}
\alias{tmapMaplibreAuxPlot.default}
\alias{tmapMapboxAuxPrepare.tm_aux_basemap}
\alias{tmapMaplibreAuxPrepare.tm_aux_basemap}
\alias{tmapMapboxAuxPrepare.tm_aux_tiles}
\alias{tmapMaplibreAuxPrepare.tm_aux_tiles}
\alias{tmapMapboxAuxPlot.tm_aux_basemap}
\alias{tmapMaplibreAuxPlot.tm_aux_basemap}
\alias{tmapMapboxAuxPlot.tm_aux_tiles}
\alias{tmapMaplibreAuxPlot.tm_aux_tiles}
\alias{tmapMapboxGridXLab}
\alias{tmapMapboxGridYLab}
\alias{tmapMaplibreGridXLab}
\alias{tmapMaplibreGridYLab}
\alias{tmapMapboxCompPrepare.tm_title}
\alias{tmapMapboxCompHeight.tm_title}
\alias{tmapMapboxCompWidth.tm_title}
\alias{tmapMapboxCompPlot.tm_title}
\alias{tmapMapboxCompPrepare.tm_scalebar}
\alias{tmapMapboxCompHeight.tm_scalebar}
\alias{tmapMapboxCompWidth.tm_scalebar}
\alias{tmapMapboxCompPlot.tm_scalebar}
\alias{tmapMapboxCompPrepare.tm_snow}
\alias{tmapMapboxCompHeight.tm_snow}
\alias{tmapMapboxCompWidth.tm_snow}
\alias{tmapMapboxCompPlot.tm_snow}
\alias{tmapMaplibreCompPrepare.tm_title}
\alias{tmapMaplibreCompHeight.tm_title}
\alias{tmapMaplibreCompWidth.tm_title}
\alias{tmapMaplibreCompPlot.tm_title}
\alias{tmapMaplibreCompPrepare.tm_scalebar}
\alias{tmapMaplibreCompHeight.tm_scalebar}
\alias{tmapMaplibreCompWidth.tm_scalebar}
\alias{tmapMaplibreCompPlot.tm_scalebar}
\alias{tmapMaplibreCompPrepare.tm_credits}
\alias{tmapMaplibreCompHeight.tm_credits}
\alias{tmapMaplibreCompWidth.tm_credits}
\alias{tmapMaplibreCompPlot.tm_credits}
\alias{tmapMaplibreCompPrepare.tm_mouse_coordinates}
\alias{tmapMaplibreCompHeight.tm_mouse_coordinates}
\alias{tmapMaplibreCompWidth.tm_mouse_coordinates}
\alias{tmapMaplibreCompPlot.tm_mouse_coordinates}
\alias{tmapMapboxCompPrepare.tm_minimap}
\alias{tmapMaplibreCompPrepare.tm_minimap}
\alias{tmapMapboxCompHeight.tm_minimap}
\alias{tmapMapboxCompWidth.tm_minimap}
\alias{tmapMaplibreCompHeight.tm_minimap}
\alias{tmapMaplibreCompWidth.tm_minimap}
\alias{tmapMapboxCompPlot.tm_minimap}
\alias{tmapMaplibreCompPlot.tm_minimap}
\alias{tmapMapboxCompPrepare}
\alias{tmapMapboxCompHeight}
\alias{tmapMapboxCompWidth}
\alias{tmapMapboxCompPlot}
\alias{tmapMapboxCompPrepare.default}
\alias{tmapMapboxCompPrepare.tm_chart_none}
\alias{tmapMaplibreCompPrepare}
\alias{tmapMaplibreCompHeight}
\alias{tmapMaplibreCompWidth}
\alias{tmapMaplibreCompPlot}
\alias{tmapMaplibreCompPrepare.default}
\alias{tmapMaplibreCompPrepare.tm_chart_none}
\alias{tmapMapboxCompPrepare.tm_draw}
\alias{tmapMapboxCompHeight.tm_draw}
\alias{tmapMapboxCompWidth.tm_draw}
\alias{tmapMapboxCompPlot.tm_draw}
\alias{tmapMaplibreCompPrepare.tm_draw}
\alias{tmapMaplibreCompHeight.tm_draw}
\alias{tmapMaplibreCompWidth.tm_draw}
\alias{tmapMaplibreCompPlot.tm_draw}
\alias{tmapMapboxCompPrepare.tm_geocoder}
\alias{tmapMapboxCompHeight.tm_geocoder}
\alias{tmapMapboxCompWidth.tm_geocoder}
\alias{tmapMapboxCompPlot.tm_geocoder}
\alias{tmapMaplibreCompPrepare.tm_geocoder}
\alias{tmapMaplibreCompHeight.tm_geocoder}
\alias{tmapMaplibreCompWidth.tm_geocoder}
\alias{tmapMaplibreCompPlot.tm_geocoder}
\alias{tmapMapboxInit}
\alias{tmapMaplibreInit}
\alias{tmapMapboxAux}
\alias{tmapMaplibreAux}
\alias{tmapMapboxCompPlot.tm_legend_portrait}
\alias{tmapMapboxCompPlot.tm_legend_landscape}
\alias{tmapMaplibreCompPlot.tm_legend_portrait}
\alias{tmapMaplibreCompPlot.tm_legend_landscape}
\alias{tmapMapboxComp}
\alias{tmapMapboxLegend}
\alias{tmapMaplibreComp}
\alias{tmapMapboxProviders}
\alias{tmapMaplibreProviders}
\alias{tmapMapboxRun}
\alias{tmapMaplibreRun}
\alias{tmapMapboxArrange}
\alias{tmapMaplibreArrange}
\alias{tmapMapboxShape}
\alias{tmapMaplibreShape}
\alias{tmapMapboxOverlay}
\alias{tmapMaplibreOverlay}
\alias{tmapMapboxWrap}
\alias{tmapMapboxXtab}
\alias{tmapMaplibreWrap}
\alias{tmapMaplibreXtab}
\alias{tmapMapboxAuxPrepare.tm_aux_snow}
\alias{tmapMapboxAuxPlot.tm_aux_snow}
\alias{tmapMapboxAuxPrepare.tm_aux_rain}
\alias{tmapMapboxAuxPlot.tm_aux_rain}
\alias{tmapMapboxCompPrepare.tm_fullscreen}
\alias{tmapMapboxCompHeight.tm_fullscreen}
\alias{tmapMapboxCompWidth.tm_fullscreen}
\alias{tmapMapboxCompPlot.tm_fullscreen}
\alias{tmapMaplibreCompPrepare.tm_fullscreen}
\alias{tmapMaplibreCompHeight.tm_fullscreen}
\alias{tmapMaplibreCompWidth.tm_fullscreen}
\alias{tmapMaplibreCompPlot.tm_fullscreen}
\alias{tmapMapboxDataPlot}
\alias{tmapMapboxDataPlot.default}
\alias{tmapMapboxDataPlot.tm_data_polygons}
\alias{tmapMapboxDataPlot.tm_data_polygons_3d}
\alias{tmapMapboxDataPlot.tm_data_lines}
\alias{tmapMapboxDataPlot.tm_data_symbols}
\alias{tmapMapboxDataPlot.tm_data_circles}
\alias{tmapMapboxDataPlot.tm_data_raster}
\alias{tmapMapboxDataPlot.tm_data_fill}
\alias{tmapMapboxDataPlot.tm_data_borders}
\alias{tmapMapboxDataPlot.tm_data_dots}
\alias{tmapMapboxDataPlot.tm_data_bubbles}
\alias{tmapMaplibreDataPlot}
\alias{tmapMaplibreDataPlot.default}
\alias{tmapMaplibreDataPlot.tm_data_polygons}
\alias{tmapMaplibreDataPlot.tm_data_polygons_3d}
\alias{tmapMaplibreDataPlot.tm_data_lines}
\alias{tmapMaplibreDataPlot.tm_data_symbols}
\alias{tmapMaplibreDataPlot.tm_data_circles}
\alias{tmapMaplibreDataPlot.tm_data_raster}
\alias{tmapMaplibreDataPlot.tm_data_fill}
\alias{tmapMaplibreDataPlot.tm_data_borders}
\alias{tmapMaplibreDataPlot.tm_data_dots}
\alias{tmapMaplibreDataPlot.tm_data_bubbles}
\title{Internal tmap methods}
\usage{
tmapValuesCheck_height(x, is_var = TRUE)

tmapValuesIsDiv_height(x)

tmapValuesRange_height(x, n, isdiv)

tmapValuesVV_height(
  x,
  value.na,
  isdiv,
  n,
  dvalues,
  are_breaks,
  midpoint,
  range,
  scale,
  rep,
  o
)

tmapValuesSubmit_height(x, args)

tmapValuesScale_height(x, scale)

tmapValuesColorize_height(x, pc)

tmapValuesCVV_height(x, value.na, n, range, scale, rep, o)

tmapMapboxAuxPrepare(a, bs, id, o)

\method{tmapMapboxAuxPrepare}{default}(a, bs, id, o)

tmapMapboxAuxPlot(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapMapboxAuxPlot}{default}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

tmapMaplibreAuxPrepare(a, bs, id, o)

\method{tmapMaplibreAuxPrepare}{default}(a, bs, id, o)

tmapMaplibreAuxPlot(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapMaplibreAuxPlot}{default}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapMapboxAuxPrepare}{tm_aux_basemap}(a, bs, id, o)

\method{tmapMaplibreAuxPrepare}{tm_aux_basemap}(a, bs, id, o)

\method{tmapMapboxAuxPrepare}{tm_aux_tiles}(a, bs, id, o)

\method{tmapMaplibreAuxPrepare}{tm_aux_tiles}(a, bs, id, o)

\method{tmapMapboxAuxPlot}{tm_aux_basemap}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapMaplibreAuxPlot}{tm_aux_basemap}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapMapboxAuxPlot}{tm_aux_tiles}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapMaplibreAuxPlot}{tm_aux_tiles}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

tmapMapboxGridXLab(bi, bbx, facet_row, facet_col, facet_page, o)

tmapMapboxGridYLab(bi, bbx, facet_row, facet_col, facet_page, o)

tmapMaplibreGridXLab(bi, bbx, facet_row, facet_col, facet_page, o)

tmapMaplibreGridYLab(bi, bbx, facet_row, facet_col, facet_page, o)

\method{tmapMapboxCompPrepare}{tm_title}(comp, o)

\method{tmapMapboxCompHeight}{tm_title}(comp, o)

\method{tmapMapboxCompWidth}{tm_title}(comp, o)

\method{tmapMapboxCompPlot}{tm_title}(comp, m, o)

\method{tmapMapboxCompPrepare}{tm_scalebar}(comp, o)

\method{tmapMapboxCompHeight}{tm_scalebar}(comp, o)

\method{tmapMapboxCompWidth}{tm_scalebar}(comp, o)

\method{tmapMapboxCompPlot}{tm_scalebar}(comp, m, o)

\method{tmapMapboxCompPrepare}{tm_snow}(comp, o)

\method{tmapMapboxCompHeight}{tm_snow}(comp, o)

\method{tmapMapboxCompWidth}{tm_snow}(comp, o)

\method{tmapMapboxCompPlot}{tm_snow}(comp, m, o)

\method{tmapMaplibreCompPrepare}{tm_title}(comp, o)

\method{tmapMaplibreCompHeight}{tm_title}(comp, o)

\method{tmapMaplibreCompWidth}{tm_title}(comp, o)

\method{tmapMaplibreCompPlot}{tm_title}(comp, m, o)

\method{tmapMaplibreCompPrepare}{tm_scalebar}(comp, o)

\method{tmapMaplibreCompHeight}{tm_scalebar}(comp, o)

\method{tmapMaplibreCompWidth}{tm_scalebar}(comp, o)

\method{tmapMaplibreCompPlot}{tm_scalebar}(comp, m, o)

\method{tmapMaplibreCompPrepare}{tm_credits}(comp, o)

\method{tmapMaplibreCompHeight}{tm_credits}(comp, o)

\method{tmapMaplibreCompWidth}{tm_credits}(comp, o)

\method{tmapMaplibreCompPlot}{tm_credits}(comp, m, o)

\method{tmapMaplibreCompPrepare}{tm_mouse_coordinates}(comp, o)

\method{tmapMaplibreCompHeight}{tm_mouse_coordinates}(comp, o)

\method{tmapMaplibreCompWidth}{tm_mouse_coordinates}(comp, o)

\method{tmapMaplibreCompPlot}{tm_mouse_coordinates}(comp, m, o)

\method{tmapMapboxCompPrepare}{tm_minimap}(comp, o)

\method{tmapMaplibreCompPrepare}{tm_minimap}(comp, o)

\method{tmapMapboxCompHeight}{tm_minimap}(comp, o)

\method{tmapMapboxCompWidth}{tm_minimap}(comp, o)

\method{tmapMaplibreCompHeight}{tm_minimap}(comp, o)

\method{tmapMaplibreCompWidth}{tm_minimap}(comp, o)

\method{tmapMapboxCompPlot}{tm_minimap}(comp, m, o)

\method{tmapMaplibreCompPlot}{tm_minimap}(comp, m, o)

tmapMapboxCompPrepare(comp, o)

tmapMapboxCompHeight(comp, o)

tmapMapboxCompWidth(comp, o)

tmapMapboxCompPlot(comp, m, o)

\method{tmapMapboxCompPrepare}{default}(comp, o)

\method{tmapMapboxCompPrepare}{tm_chart_none}(comp, o)

tmapMaplibreCompPrepare(comp, o)

tmapMaplibreCompHeight(comp, o)

tmapMaplibreCompWidth(comp, o)

tmapMaplibreCompPlot(comp, m, o)

\method{tmapMaplibreCompPrepare}{default}(comp, o)

\method{tmapMaplibreCompPrepare}{tm_chart_none}(comp, o)

\method{tmapMapboxCompPrepare}{tm_draw}(comp, o)

\method{tmapMapboxCompHeight}{tm_draw}(comp, o)

\method{tmapMapboxCompWidth}{tm_draw}(comp, o)

\method{tmapMapboxCompPlot}{tm_draw}(comp, m, o)

\method{tmapMaplibreCompPrepare}{tm_draw}(comp, o)

\method{tmapMaplibreCompHeight}{tm_draw}(comp, o)

\method{tmapMaplibreCompWidth}{tm_draw}(comp, o)

\method{tmapMaplibreCompPlot}{tm_draw}(comp, m, o)

\method{tmapMapboxCompPrepare}{tm_geocoder}(comp, o)

\method{tmapMapboxCompHeight}{tm_geocoder}(comp, o)

\method{tmapMapboxCompWidth}{tm_geocoder}(comp, o)

\method{tmapMapboxCompPlot}{tm_geocoder}(comp, m, o)

\method{tmapMaplibreCompPrepare}{tm_geocoder}(comp, o)

\method{tmapMaplibreCompHeight}{tm_geocoder}(comp, o)

\method{tmapMaplibreCompWidth}{tm_geocoder}(comp, o)

\method{tmapMaplibreCompPlot}{tm_geocoder}(comp, m, o)

tmapMapboxInit(o, return.asp = FALSE, vp, prx, dg = NULL, m = NULL, ...)

tmapMaplibreInit(o, return.asp = FALSE, vp, prx, dg = NULL, m = NULL, ...)

tmapMapboxAux(o, q)

tmapMaplibreAux(o, q)

\method{tmapMapboxCompPlot}{tm_legend_portrait}(comp, m, o)

\method{tmapMapboxCompPlot}{tm_legend_landscape}(comp, m, o)

\method{tmapMaplibreCompPlot}{tm_legend_portrait}(comp, m, o)

\method{tmapMaplibreCompPlot}{tm_legend_landscape}(comp, m, o)

tmapMapboxComp(
  comp,
  o,
  facet_row = NULL,
  facet_col = NULL,
  facet_page,
  class,
  stack,
  stack_auto,
  pos.h,
  pos.v,
  bbox
)

tmapMaplibreComp(
  comp,
  o,
  facet_row = NULL,
  facet_col = NULL,
  facet_page,
  class,
  stack,
  stack_auto,
  pos.h,
  pos.v,
  bbox
)

tmapMapboxProviders(credits)

tmapMaplibreProviders(credits)

tmapMapboxRun(o, q, show, knit, knit_opts, args)

tmapMaplibreRun(o, q, show, knit, knit_opts, args)

tmapMapboxArrange(tms, nx, ncol, nrow, opts, knit, show, args, options)

tmapMaplibreArrange(tms, nx, ncol, nrow, opts, knit, show, args, options)

tmapMapboxShape(bbx, facet_row, facet_col, facet_page, o)

tmapMaplibreShape(bbx, facet_row, facet_col, facet_page, o)

tmapMapboxOverlay(bbx, facet_row, facet_col, facet_page, o)

tmapMaplibreOverlay(bbx, facet_row, facet_col, facet_page, o)

tmapMapboxWrap(label, facet_row, facet_col, facet_page, o)

tmapMapboxXtab(label, facet_row, facet_col, facet_page, o)

tmapMaplibreWrap(label, facet_row, facet_col, facet_page, o)

tmapMaplibreXtab(label, facet_row, facet_col, facet_page, o)

\method{tmapMapboxAuxPrepare}{tm_aux_snow}(a, bs, id, o)

\method{tmapMapboxAuxPlot}{tm_aux_snow}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapMapboxAuxPrepare}{tm_aux_rain}(a, bs, id, o)

\method{tmapMapboxAuxPlot}{tm_aux_rain}(
  a,
  bi,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  o
)

\method{tmapMapboxCompPrepare}{tm_fullscreen}(comp, o)

\method{tmapMapboxCompHeight}{tm_fullscreen}(comp, o)

\method{tmapMapboxCompWidth}{tm_fullscreen}(comp, o)

\method{tmapMapboxCompPlot}{tm_fullscreen}(comp, m, o)

\method{tmapMaplibreCompPrepare}{tm_fullscreen}(comp, o)

\method{tmapMaplibreCompHeight}{tm_fullscreen}(comp, o)

\method{tmapMaplibreCompWidth}{tm_fullscreen}(comp, o)

\method{tmapMaplibreCompPlot}{tm_fullscreen}(comp, m, o)

tmapMapboxDataPlot(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{default}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_polygons}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_polygons_3d}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_lines}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_symbols}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_circles}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_raster}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_fill}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_borders}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_dots}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMapboxDataPlot}{tm_data_bubbles}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

tmapMaplibreDataPlot(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{default}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_polygons}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_polygons_3d}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_lines}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_symbols}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_circles}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_raster}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_fill}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_borders}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_dots}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)

\method{tmapMaplibreDataPlot}{tm_data_bubbles}(
  a,
  shpTM,
  dt,
  pdt,
  popup.format,
  hdt,
  idt,
  gp,
  bbx,
  facet_row,
  facet_col,
  facet_page,
  id,
  pane,
  group,
  glid,
  o,
  ...
)
}
\arguments{
\item{o}{the list of options}

\item{args}{args}

\item{bs}{bs}

\item{id}{id}

\item{bi}{bi}

\item{bbx}{bbx}

\item{facet_row, facet_col, facet_page}{row column and page id}

\item{pane}{pane}

\item{group}{group}

\item{comp}{the shape object}

\item{m}{the proxy object from `maplibre_proxy()`/`mapboxgl_proxy()`, supplied in proxy mode}

\item{return.asp}{return.asp}

\item{vp}{vp}

\item{dg}{base map widget (internal; \code{NULL} in proxy mode, where the
proxy object \code{m} is used as the drawing canvas instead)}

\item{q}{q}

\item{class}{class}

\item{stack}{stack}

\item{stack_auto}{stack_auto}

\item{pos.h}{pos.h}

\item{pos.v}{pos.v}

\item{bbox}{bbox}

\item{credits}{credits}

\item{show}{show}

\item{knit}{knit}

\item{knit_opts}{knit options}

\item{tms}{list of tmap objects}

\item{nx}{number of facets}

\item{nrow}{number of grid rows}

\item{opts}{arrange options (widths, heights, height, sync, ...)}

\item{label}{label}

\item{shpTM, }{dt, pdt, popup.format, hdt, idt, gp args}
}
\value{
internal tmap lists
}
\description{
Internal tmap methods
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg.R
\docType{package}
\name{tmap.mapgl-package}
\alias{tmap.mapgl}
\alias{tmap.mapgl-package}
\title{Extensions to 'tmap' with Two New Modes: 'mapbox' and 'maplibre'}
\description{
The 'tmap' package provides two plotting modes for static and interactive
thematic maps. This package extends 'tmap' with two additional modes based
on 'mapbox' <https://mapbox.com> and 'maplibre' <https://maplibre.org>.
These modes feature interactive vector tiles, globe views, and other modern
web-mapping capabilities, while using the same 'tmap' syntax for creating
maps (Tennekes 2018, <doi:10.32614/RJ-2018-027>).
}
\examples{
\dontrun{
library(tmap)
library(tmap.mapgl)
# getting API: https://walker-data.com/mapgl/articles/getting-started.html
# check API envir var: Sys.getenv("MAPBOX_PUBLIC_TOKEN")

tmap_mode("mapbox")
tm_shape(World) +
  tm_polygons("HPI", fill.scale = tm_scale_intervals(values = "brewer.rd_yl_gn"))

tm_shape(NLD_dist) +
  tm_polygons("employment_rate",
    fill.scale = tm_scale_intervals(values = "scico.roma"),
    lwd = 0.1) +
tm_shape(NLD_muni) +
  tm_polygons(fill = NULL, lwd = 1) +
tm_mapbox(pitch = 60) +
tm_basemap("mapbox.dark")
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-tmap/tmap.mapgl}
  \item \url{https://r-tmap.github.io/tmap.mapgl/}
  \item Report bugs at \url{https://github.com/r-tmap/tmap.mapgl/issues}
}

}
\author{
Martijn Tennekes \email{mtennekes@gmail.com}
}
\concept{GIS}
\concept{bubble map}
\concept{choropleth}
\concept{statistical maps}
\concept{thematic maps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_mapbox.R
\name{tm_maplibre}
\alias{tm_maplibre}
\title{Maplibre mode options}
\usage{
tm_maplibre(pitch, control.position, control.collapse, zoom)
}
\arguments{
\item{pitch}{The pitch angle}

\item{control.position}{The position of the layer control box}

\item{control.collapse}{Should the layer control box be collapsed?}

\item{zoom}{The zoom level of the map}
}
\value{
a [tmap::tmap-element]
}
\description{
Maplibre mode options. These options are specific to the maplibre mode.
}
\examples{
\donttest{
library(tmap)
library(tmap.mapgl)
tmap_mode("maplibre")
tm_shape(World) +
  tm_polygons("HPI", fill.scale = tm_scale_intervals(values = "brewer.rd_yl_gn"))

tm_shape(NLD_dist) +
  tm_polygons("employment_rate",
    fill.scale = tm_scale_intervals(values = "scico.roma"),
    lwd = 0.1) +
  tm_shape(NLD_muni) +
	tm_polygons(fill = NULL, lwd = 1) +
tm_mapbox(pitch = 60) +
tm_basemap("carto.dark_matter")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_mapbox.R
\name{tm_mapbox}
\alias{tm_mapbox}
\title{Mapbox mode options}
\usage{
tm_mapbox(pitch, control.position, control.collapse, zoom)
}
\arguments{
\item{pitch}{The pitch angle}

\item{control.position}{The position of the layer control box}

\item{control.collapse}{Should the layer control box be collapsed?}

\item{zoom}{The zoom level of the map}
}
\value{
a [tmap::tmap-element]
}
\description{
Mapbox mode options. These options are specific to the mapbox mode.
}
\examples{
\dontrun{
library(tmap)
library(tmap.mapgl)
# getting API: https://walker-data.com/mapgl/articles/getting-started.html
# check API envir var: Sys.getenv("MAPBOX_PUBLIC_TOKEN")

tmap_mode("mapbox")
tm_shape(World) +
  tm_polygons("HPI", fill.scale = tm_scale_intervals(values = "brewer.rd_yl_gn"))

tm_shape(NLD_dist) +
  tm_polygons("employment_rate",
    fill.scale = tm_scale_intervals(values = "scico.roma"),
    lwd = 0.1) +
tm_shape(NLD_muni) +
  tm_polygons(fill = NULL, lwd = 1) +
tm_mapbox(pitch = 60) +
tm_basemap("mapbox.dark")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_layers_polygons_3d.R
\name{tm_polygons_3d}
\alias{tm_polygons_3d}
\alias{opt_tm_polygons_3d}
\title{Map layer: polygons in 3d (experimental)}
\usage{
tm_polygons_3d(
  height = tmap::tm_const(),
  height.scale = tmap::tm_scale(),
  height.legend = tmap::tm_legend_hide(),
  height.chart = tmap::tm_chart_none(),
  height.free = NA,
  fill = tmap::tm_const(),
  fill.scale = tmap::tm_scale(),
  fill.legend = tmap::tm_legend(),
  fill.chart = tmap::tm_chart_none(),
  fill.free = NA,
  col = tmap::tm_const(),
  col.scale = tmap::tm_scale(),
  col.legend = tmap::tm_legend(),
  col.chart = tmap::tm_chart_none(),
  col.free = NA,
  lwd = tmap::tm_const(),
  lwd.scale = tmap::tm_scale(),
  lwd.legend = tmap::tm_legend(),
  lwd.chart = tmap::tm_chart_none(),
  lwd.free = NA,
  lty = tmap::tm_const(),
  lty.scale = tmap::tm_scale(),
  lty.legend = tmap::tm_legend(),
  lty.chart = tmap::tm_chart_none(),
  lty.free = NA,
  fill_alpha = tmap::tm_const(),
  fill_alpha.scale = tmap::tm_scale(),
  fill_alpha.legend = tmap::tm_legend(),
  fill_alpha.chart = tmap::tm_chart_none(),
  fill_alpha.free = NA,
  col_alpha = tmap::tm_const(),
  col_alpha.scale = tmap::tm_scale(),
  col_alpha.legend = tmap::tm_legend(),
  col_alpha.chart = tmap::tm_chart_none(),
  col_alpha.free = NA,
  linejoin = "round",
  lineend = "round",
  plot.order = tmap::tm_plot_order("lwd", reverse = TRUE, na.order = "bottom"),
  zindex = NA,
  group = NA,
  group.control = "check",
  popup.vars = NA,
  popup.format = list(),
  hover = NA,
  id = "",
  options = opt_tm_polygons_3d()
)

opt_tm_polygons_3d(
  polygons.only = "ifany",
  height.max = "10\%",
  height.min = "0.1\%",
  cast.shadows = NULL,
  ambient.occlusion.intensity = NULL,
  ambient.occlusion.radius = NULL,
  vertical.gradient = NULL,
  emissive.strength = NULL
)
}
\arguments{
\item{height, height.scale, height.legend, height.chart, height.free}{`r .doc_vv("height")`}

\item{fill, fill.scale, fill.legend, fill.chart, fill.free}{`r .doc_vv("fill")`}

\item{col, col.scale, col.legend, col.chart, col.free}{`r .doc_vv("col")`}

\item{lwd, lwd.scale, lwd.legend, lwd.chart, lwd.free}{`r .doc_vv("lwd")`}

\item{lty, lty.scale, lty.legend, lty.chart, lty.free}{`r .doc_vv("lty")`}

\item{fill_alpha, fill_alpha.scale, fill_alpha.chart, fill_alpha.legend, fill_alpha.free}{`r .doc_vv("fill_alpha")`}

\item{col_alpha, col_alpha.scale, col_alpha.legend, col_alpha.chart, col_alpha.free}{`r .doc_vv("col_alpha")`}

\item{linejoin, lineend}{Line join and line end. See [gpar()][grid::gpar()] for details.}

\item{plot.order}{Specification in which order the spatial features are drawn.
See [tm_plot_order()] for details.}

\item{zindex}{Map layers are drawn on top of each other. The `zindex` numbers
(one for each map layer) determines the stacking order.
By default the map layers are drawn in the order they are called.}

\item{group}{Name of the group to which this layer belongs. This is only
relevant in view mode, where layer groups can be switched (see `group.control`)}

\item{group.control}{In view mode, the group control determines how layer groups
can be switched on and off. Options: `"radio"` for radio buttons
(meaning only one group can be shown), `"check"` for check boxes
(so multiple groups can be shown), and `"none"` for no control
(the group cannot be (de)selected).}

\item{popup.vars}{names of data variables that are shown in the popups
in `"view"` mode. Set popup.vars to `TRUE` to show all variables in the
shape object. Set popup.vars to `FALSE` to disable popups. Set `popup.vars`
to a character vector of variable names to those those variables in the popups.
The default (`NA`) depends on whether visual variables (e.g.`fill`) are used.
If so, only those are shown. If not all variables in the shape object are shown.}

\item{popup.format}{list of formatting options for the popup values.
See the argument `legend.format` for options. Only applicable for
numeric data variables. If one list of formatting options is provided,
it is applied to all numeric variables of `popup.vars`. Also, a (named)
list of lists can be provided. In that case, each list of formatting options
is applied to the named variable.}

\item{hover}{name of the data variable that specifies the hover labels (view mode only). Set to `FALSE` to disable hover labels. By default `FALSE`, unless `id` is specified. In that case, it is set to `id`,}

\item{id}{name of the data variable that specifies the indices of the spatial
features. Only used for `"view"` mode.}

\item{options}{options passed on to the corresponding `opt_<layer_function>` function}

\item{polygons.only}{should only polygon geometries of the shape object (defined in [tm_shape()]) be plotted? By default `"ifany"`, which means `TRUE` in case a geometry collection is specified.}

\item{height.max, height.min}{Maximum and minimum height. The data variable values assigned to `height` are are scaled between these two values. Each can be an absolute number (in meters) or relative to the bounding box area, more specifically, the percentage of the square root of the bounding box area. The default values are `"10\%"` for `height.max` and `0.1\%` for `height.min`. The latter is not set to 0 because of artefacts with maplibre in globe view.}

\item{cast.shadows}{Should the extruded polygons cast shadows? **`"mapbox"` mode only** (silently ignored with a message in `"maplibre"` mode, whose fill-extrusion implementation does not support shadows). Passed to `fill_extrusion_cast_shadows`. The default `NULL` keeps the engine default.}

\item{ambient.occlusion.intensity, ambient.occlusion.radius}{Ambient-occlusion intensity (0-1) and radius for the extruded polygons, adding soft contact shading in corners and crevices. **`"mapbox"` mode only** (ignored with a message in `"maplibre"` mode). Passed to `fill_extrusion_ambient_occlusion_intensity` / `fill_extrusion_ambient_occlusion_radius`. The default `NULL` keeps the engine default.}

\item{vertical.gradient}{Should a vertical gradient be applied to the sides of the extrusions? When `TRUE`, the side walls are shaded from darker at the base to lighter toward the top, which gives a sense of depth. Supported in **both** `"maplibre"` and `"mapbox"` modes. Passed to `fill_extrusion_vertical_gradient`. The default `NULL` keeps the engine default, which is `TRUE` - so the visible effect is mainly when set to `FALSE` (flat, evenly-coloured side walls).}

\item{emissive.strength}{Emissive strength of the extruded polygons, controlling how much they appear self-lit under the map's 3D lighting. **`"mapbox"` mode only** (ignored with a message in `"maplibre"` mode). Passed to `fill_extrusion_emissive_strength`. The default `NULL` keeps the engine default.}
}
\value{
a [tmap::tmap-element], supposed to be stacked after [tmap::tm_shape()] using the `+` operator. The `opt_<layer_function>` function returns a list that should be passed on to the `options` argument.
}
\description{
Map layer that draws 3d (extruded) polygons. Supported visual variables are: `height` (the ), `fill` (the fill color),
`col` (the border color), `lwd` (line width), `lty` (line type),
`fill_alpha` (fill color alpha transparency) and `col_alpha` (border color alpha transparency).
}
\details{
The visual variable arguments (e.g. `col`) can be specified with either a data
variable name (e.g., a spatial vector attribute or a raster layer of the object
specified in [tm_shape()]), or with a visual value (for `col`, a color is expected).
See \href{https://r-tmap.github.io/tmap/articles/basics_vv}{vignette about visual variables}.

Multiple values can be specified: in that case facets are created.
These facets can be combined with other faceting data variables, specified with [tm_facets()].
See \href{https://r-tmap.github.io/tmap/articles/basics_facets}{vignette about facets}.

- The `*.scale` arguments determine the used scale to map the data values to
visual variable values. These can be specified with one of the available
`tm_scale_*()` functions. The default is specified by the tmap option ([tm_options()]) `scales.var`.
See \href{https://r-tmap.github.io/tmap/articles/basics_scales}{vignette about scales}.

- The `*.legend` arguments determine the used legend, specified with [tm_legend()].
The default legend and its settings are determined by the tmap options ([tm_options()]) `legend.` .
See \href{https://r-tmap.github.io/tmap/articles/basics_legends}{vignette about legends}.

- The `*.chart` arguments specify additional charts, specified with `tm_chart_`, e.g. [tm_chart_histogram()].
See \href{https://r-tmap.github.io/tmap/articles/basics_charts}{vignette about charts}.

- The `*.free` arguments determine whether scales are applied freely across facets, or shared.
A logical value is required. They can also be specified with a vector of three
logical values; these determine whether scales are applied freely per facet dimension.
This is only useful when facets are applied (see [tm_facets()]).
There are maximally three facet dimensions: rows, columns, and pages. This only
applies for a facet grid ([tm_facets_grid()]). For instance, `col.free = c(TRUE, FALSE, FALSE)`
means that for the visual variable `col`, each row of facets will have its own
scale, and therefore its own legend. For facet wraps and stacks
([tm_facets_wrap()] and [tm_facets_stack()]) there is only one facet dimension,
so the `*.free` argument requires only one logical value.
}
\examples{
\donttest{
library(tmap)
library(tmap.mapgl)
tmap_mode("maplibre")
NLD_dist$pop_dens = NLD_dist$population / NLD_dist$area
tm_shape(NLD_dist) +
  tm_polygons_3d(height = "pop_dens",
    fill = "edu_appl_sci",
    fill.scale = tm_scale_intervals(style = "kmeans", values = "-pu_gn"),
    fill.legend = tm_legend("Univeristy degree")) +
tm_maplibre(pitch = 45) +
tm_crs(3857)
}
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/examples_choro_World}{Choropleth example (1)} and \href{https://r-tmap.github.io/tmap/articles/examples_choro_NLD}{choropleth example (2)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapgl_export.R
\name{tmap_mapbox}
\alias{tmap_mapbox}
\alias{tmap_maplibre}
\title{Export tmap to mapbox and maplibre}
\usage{
tmap_mapbox(x, show = FALSE, ...)

tmap_maplibre(x, show = FALSE, ...)
}
\arguments{
\item{x}{a tmap object.}

\item{show}{show the map?}

\item{...}{passed on to [`tmap`][tmap::print.tmap()]}
}
\value{
a [`mapgl`][mapgl::mapboxgl()] object (`"mapbox" mode`) or a [`mapgl`][mapgl::maplibregl()] object (`"maplibre"` mode). In case small multiples are shown, a list is returned.
}
\description{
* `tmap_mapbox()` returns a [`mapgl`][mapgl::mapboxgl()] object (`"mapbox" mode`)
* `tmap_maplibre()` a [`mapgl`][mapgl::maplibregl()] object (`"maplibre"` mode).
}
\examples{
library(tmap)
library(tmap.mapgl)
map = tm_shape(World) + tm_polygons()
tmap_maplibre(map, show = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapgl_draw.R
\name{tm_draw}
\alias{tm_draw}
\title{Map component: draw toolbox}
\usage{
tm_draw(download_button, show_measurements, stack, position, group_id, z)
}
\arguments{
\item{download_button}{show download button? By default `TRUE`.}

\item{show_measurements}{show measurements?  By default `TRUE`. The measurement unit is set in [`tmap`][tmap::tm_shape()]}

\item{stack}{stack with other map components, either `"vertical"` or `"horizontal"`.}

\item{position}{The position specification of the component: an object created with `tm_pos_in()` or `tm_pos_out()`. Or, as a shortcut, a vector of two values, specifying the x and y coordinates. The first is `"left"`, `"center"` or `"right"` (or upper case, meaning tighter to the map frame), the second `"top"`, `"center"` or `"bottom"`. Numeric values are also supported, where 0, 0 means left bottom and 1, 1 right top. See also \href{https://r-tmap.github.io/tmap/articles/adv_positions}{vignette about positioning}. In case multiple components should be combined (stacked), use `group_id` and specify `component` in [tm_comp_group()].}

\item{group_id}{Component group id name. All components (e.g. legends, titles, etc) with the same `group_id` will be grouped. The specifications of how they are placed (e.g. stacking, margins etc.) are determined in [tm_comp_group()] where its argument `id` should correspond to `group_id`.}

\item{z}{z index, e.g. the place of the component relative to the other componets}
}
\value{
a [tmap::tmap-element]
}
\description{
Map component that adds a draw toolbox
}
\seealso{
\href{https://r-tmap.github.io/tmap/articles/basics_components}{Vignette about components}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tm_aux_layers.R
\name{tm_snow}
\alias{tm_snow}
\title{Map layer: let it snow!}
\usage{
tm_snow(
  density = 0.85,
  intensity = 1,
  color = "#ffffff",
  opacity = 1,
  center_thinning = 0.4,
  direction = c(0, 50),
  flake_size = 0.71,
  vignette = 0.3,
  vignette_color = "#ffffff"
)
}
\arguments{
\item{density}{A number between 0 and 1 controlling the snow particles density. Default is 0.85.}

\item{intensity}{A number between 0 and 1 controlling the snow particles movement speed. Default is 1.0.}

\item{color}{A string specifying the color of the snow particles. Default is "#ffffff".}

\item{opacity}{A number between 0 and 1 controlling the snow particles opacity. Default is 1.0.}

\item{center_thinning}{A number between 0 and 1 controlling the thinning factor of snow particles from center. Default is 0.4.}

\item{direction}{A numeric vector of length 2 defining the azimuth and polar angles of the snow direction. Default is c(0, 50).}

\item{flake_size}{A number between 0 and 5 controlling the snow flake particle size. Default is 0.71.}

\item{vignette}{A number between 0 and 1 controlling the snow vignette screen-space effect. Default is 0.3.}

\item{vignette_color}{A string specifying the snow vignette screen-space corners tint color. Default is "#ffffff".}
}
\value{
tmap layer
}
\description{
Map layer that generates snow.
}
\details{
Arguments and their default values have been taken from [mapbox::set_snow()] (package version 0.2.2)
}
\seealso{
[tm_rain()]
}
